"""
Example:
    def init(): return
    def run(i): return
    b = Benchmark(number=100, concurrency=100)
    b.set_init_callback(init)
    b.set_run_callback(run)
    b.run()

The init callback is called after the child is spawned.

The run callback is called in a loop until the number of runs is exceeded.
The iterator count is passed to the run callback, which is a unique number
over all children.
"""

from multiprocessing import Process
from multiprocessing import Value, Lock
import sys
import time

class Benchmark:
    def __init__(self, number, concurrency):
        self.number = number
        self.concurrency = concurrency
        self.number_per_child = int(number / concurrency)
        self.children = {}
        self.values = []
        self.start = Value("i", 0, lock=Lock())
        self.done = False
        self.init_callback = None

    def set_init_callback(self, callback):
        self.init_callback = callback

    def set_run_callback(self, callback):
        self.run_callback = callback

    def run(self):
        self.spawn_children()
        self.wait_for_children()
        self.check_stats()

    def wait_for_children(self):
        while self.check_ready_clients() is False:
            time.sleep(0.1)
        self.start.value = 1

    def check_ready_clients(self):
        for value in self.values:
            if value.value == -1:
                return False
        return True

    def spawn_children(self):
        for i in range(self.concurrency):
            start = self.number_per_child * i + 1 # no 0
            end = start + self.number_per_child
            # -1 = child not initialized
            # the child set the value to 0 if the child is ready
            value = Value("i", -1, lock=Lock())
            self.values.append(value)
            process = Process(target=self.run_child, args=(value, start, end))
            process.start()
            self.children[process.pid] = process

    def reap_children(self):
        for pid in list(self.children):
            c = self.children[pid]
            if not c.is_alive():
                c.join()
                del self.children[pid]

        if len(self.children) == 0:
            self.done = True

    def check_stats(self):
        start_time = time.time()
        pformat = "%-19s %12s %12s"

        print(pformat % ("Time", "Total", "Exec/s"))

        while self.done is False:
            time.sleep(5)
            total = 0

            for value in self.values:
                total += value.value

            run_time = time.time() - start_time
            timestamp = time.strftime("%Y-%m-%d %H:%M:%S")
            runs_per_sec = int(total / run_time)
            print(pformat % (timestamp, total, runs_per_sec))
            self.reap_children()

        print("Total time: %ss" % int(time.time() - start_time))

    def run_child(self, value, start, end):
        if self.init_callback:
            self.init_callback()

        value.value = 0
        while self.start.value == 0:
            time.sleep(0.1)

        while start < end:
            self.run_callback(start)
            start += 1
            value.value += 1
