import json
import os
import re
from bloonix.exception import BloonixError

def parse_file(filename):
    return __Config().parse_file(filename)

def parse_string(string):
    return __Config().parse_string(string)

class __Config:
    def __init__(self):
        self._open_subs = 0
        self.line_num_by_file = {}

    def finish(self):
        if self._open_subs > 1:
            raise BloonixError(
                "ERR: %s un-closed sections found " % str(self._open_subs)
            )
        elif self._open_subs > 0:
            raise BloonixError("ERR: 1 un-closed section found")

    def parse_file(self, filename):
        self.include = True
        self.filename = filename
        self.path = filename
        self.path = re.sub(r"[^/\\]+$", "", self.path)
        self.path = re.sub(r"[/\\]$", "", self.path)
        config = {}
        self.include_config(self.filename, config)
        self.finish()
        return config

    def parse_string(self, string):
        self.include = False
        config = {}
        lines = re.split(r"\n", string)
        self.line_num_by_file["inline"] = 0
        self.parse_config(lines, "inline", config)
        self.finish()
        return config

    def include_config(self, filename, config):
        if os.path.isdir(filename):
            self.include_dir(filename, config)
            return

        f = None
        try:
            f = open(filename, "r", encoding="utf-8")
            lines = f.readlines()
            self.line_num_by_file[filename] = 0
            self.parse_config(lines, filename, config)
        finally:
            if f: f.close()

    def include_dir(self, dirname, config):
        for filename in os.listdir(dirname):
            full_filename = "/".join([dirname, filename])
            if (os.path.isfile(full_filename)
                    and re.search(r".+\.conf\Z", filename)):
                self.include_config(full_filename, config)

    def parse_config(self, lines, filename, config):
        key = None
        value = None
        is_multiline = False
        multilines = []

        while len(lines):
            line = lines.pop(0)
            self.increase_line_num(filename)
            line = self.prepare_line(line)

            # Comments and whitespaces was removed.
            # Empty lines will be ignored.
            if len(line) == 0:
                continue

            # A section begins with a keyword
            if re.search(r"^\s*\{", line):
                self.do_raise(filename, "keyword missing for opened section")

            if re.search(r"^([^\s]+)\s*\{", line) or re.search(r"^\s*\}", line):
                if len(multilines) > 0:
                    self.do_raise(
                        filename,
                        "open/close a section if the last value " +
                        "were marked as multiline is not valid"
                    )

            # A section will be close with "}"
            if re.search(r"^\s*\}", line):
                # A section can only be close if a section were opened
                if self._open_subs == 0:
                    self.do_raise(filename, "un-opened section closed")
                self._open_subs -= 1
                return

            # If a line begins with "keyword {" then it's a sub-section.
            m = re.search(r"^([^\s]+)\s*\{", line)
            if m:
                key = m.group(1)
                value = {}
                self._open_subs += 1
                self.add_key_value(key, value, config)
                self.parse_config(lines, filename, value)
                continue

            # A key value pair. The value can be an empty string.
            if len(multilines) > 0:
                value = line
            else:
                m = re.search(r"^([^\s]+)\s*(.*)", line)
                if m:
                    key = m.group(1)
                    value = m.group(2)
                else:
                    self.do_raise(filename, "unable to parse line '" + line + "'")

            # Look if the end of the line is marked as multiline.
            if re.search(r"\s*\\\Z", value):
                is_multiline = True
                value = re.sub(r"\s*\\\Z", "", value)

            # Remove the quotes of quoted values.
            m = re.search(r"^'(.*)'\Z", value)
            if m:
                value = m.group(1)
            else:
                m = re.search(r'^"(.*)"\Z', value)
                if m:
                    value = m.group(1)
                else:
                    value = re.sub(r"^\s+", "", value)
                    value = re.sub(r"\s+\Z", "", value)

            # If the line is marked as multiline, then just
            # push the value into a temporary array.
            if is_multiline is True:
                multilines.append(value)
                is_multiline = False
                continue

            # If the last parsed parameter was a multine parameter,
            # then the array @multiline contains the values.
            if len(multilines) > 0:
                multilines.append(value)
                value = "".join(multilines)
                multilines = []

            # Add the key value pair to the config.
            self.add_key_value(key, value, config)
            key = None
            value = None

        if len(multilines) > 0:
            self.do_raise(
                filename,
                "multiline parameter does not end correctly"
            )

    def prepare_line(self, line):
        # Prepare the line and cut newline, comments
        # and whitespaces from the begin and the end
        # of each line.
        line = re.sub(r"[\r\n]+\Z", "", line)
        line = re.sub(r"\s+#.+", "", line)
        line = re.sub(r"^\s*#.*", "", line)
        line = re.sub(r"\\#", "#", line)
        line = re.sub(r"^\s*", "", line)
        line = re.sub(r"\s*\Z", "", line)
        return line

    def add_key_value(self, key, value, config):
        if key == "include" and self.include is True:
            # Match root with "/" or "C:\"
            if not re.search(r"^(/|[a-zA-Z]:[/\\])", value):
                if self.path != "":
                    value = "/".join([self.path, value])
            self.include_config(value, config)
            return

        # Force a list if the key ends with []
        # Example: key[] value
        if re.search(r"\[\]\Z", key):
            key = re.sub(r"\[\]\Z", "", key)
            if key not in config:
                config[key] = []
        if re.search(r"\[json\]\Z", key):
            value = json.loads(value)
            key = re.sub(r"\[json\]\Z", "", key)
            if key not in config:
                config[key] = []

        if key not in config:
            config[key] = value
        elif isinstance(config[key], list):
            config[key].append(value)
        else:
            config[key] = [config[key], value]

    def increase_line_num(self, filename):
        self.line_num_by_file[filename] += 1

    def get_current_line_num(self, filename):
        return str(self.line_num_by_file[filename])

    def do_raise(self, filename, msg):
        raise BloonixError(
            "ERR: " + msg + " " +
            "in file '%s' " % filename +
            "on line %s" % self.get_current_line_num(filename)
        )
