import os
import re
import sys

uwsgi_template = """
[uwsgi]
plugin = http,python3
socket = :{port}
processes = 20
master = true
chdir = {target}
wsgi-file = {target}/app/heaven.py
pyargv = {target}/etc/main.conf
need-app = true
disable-logging = true
uid = bloonix
gid = bloonix
env = PYTHONIOENCODING=UTF-8
"""

main_template = """
app-name bloonix_webapp
user bloonix
group bloonix
app {{
}}
logger {{
    file {{
        filename {target}/log/webapp.log
        level notice
    }}
}}
model {{
}}
plugin {{
}}
"""

systemd_template = """
[Unit]
Description={name}
After=syslog.target

[Service]
Type=notify
ExecStart=/usr/bin/uwsgi --ini {target}/etc/uwsgi.ini
KillSignal=2
WorkingDirectory={target}
StandardError=syslog
NotifyAccess=all
TimeoutStartSec=30
TimeoutStopSec=30

[Install]
WantedBy=multi-user.target
"""

nginx_template = """
server {{
    listen 80 default_server;
    listen [::]:80 default_server;
    server_name _;
    access_log /var/log/nginx/{name}_access.log;
    error_log /var/log/nginx/{name}_error.log;
    error_page 404 /404.html;
    root {target}/public;
    location /favicon.ico {{
        #expires 365d;
        add_header Cache-Control public;
        root {target}/public/assets;
    }}
    location /assets {{
        #expires 365d;
        add_header Cache-Control public;
    }}
    location / {{
        expires off;
        include uwsgi_params;
        uwsgi_pass 127.0.0.1:{port};
    }}
}}
"""

heaven_template = """
from bloonix.heaven import heaven
application = heaven.run
"""

html_header_template = """
<!DOCTYPE html>
<html>
<head>
<title>Hello World</title>
<meta charset="UTF-8"/>
</head>
<body>
"""

html_footer_template = """
</body>
</html>
"""

html_content_template = """
{% include ___template___ %}
"""

html_wrapper_template = """
{% include "html-header.html" %}
{% include "html-content.html" %}
{% include "html-footer.html" %}
"""

html_index_template = """
Hello World
"""

html_404_template = """
404 Not Found
"""

webapp_init_template = """
from bloonix.heaven import H
H.model.autoload()
H.view.autoload()
H.controller.autoload()
H.plugin.autoload()

from bloonix.validator import validate
# Validate app {} section from main.conf
H.config = validate(H.config, {
    "parameter": {
        "type": str,
        "default": "value",
    }
})
"""

controller_init_template = """
VERSION = "1.0"
import time
from bloonix.heaven import H

class Controller:
    def startup(self):
        H.route("/", "index")

    def auto(self, opts):
        # H.model.db.dbi.connect()
        H.stash.data["meta"] = {"version": VERSION}
        return True

    def default(self, opts):
        H.stash.template = "404.html"
        H.res.status = "404 Not Found"
        H.render("jinja")

    def error(self, opts):
        H.log.error("500 Internal Server Error")
        H.res.status = "500 Internal Server Error"

    def index(self, opts):
        H.stash.template = "index.html"
        H.render("jinja")

    def end(self, opts):
        # H.model.db.dbi.disconnect()
        runtime = time.time() - H.time
        H.req.path_info()
        H.log.notice("finished request %s (took %ss)" % (H.req.path_info(), runtime))
        return True
"""

bin_init_systemd_template = """
#!/bin/bash
ln -s /srv/{name}/etc/webapp.service \\
    /etc/systemd/system/{service}.service
systemctl daemon-reload
systemctl enable {service}
"""

build_template = """
#!/bin/bash

RESTART=$1
VERSION="1.0"
REVISION=$(date +'%y%m%d%H%M%S')
ROOT_CONTROLLER="lib/bloonix_webapp/controller/root/__init__.py"
sed -i "1s/.*/VERSION = \\"$VERSION-$REVISION\\"/" $ROOT_CONTROLLER

if [[ $RESTART = "restart" ]] ; then
    systemctl restart {name} || systemctl status {name}
    if [[ $? -ne 0 ]] ; then
        systemctl status {name}
    fi
fi
"""

def create_path(path):
    print("create path %s" % path)
    os.system("mkdir -p %s" % path)

def create_file(file, content, mode=None):
    content = re.sub(r"^\n*", "", content)
    print("create file %s" % file)
    f = open(file, "w")
    f.write(content)
    f.close()
    if mode != None:
        execute("chmod %s %s/bin/init-systemd" % (mode, target))

def execute(cmd):
    print("\nexec %s" % cmd)
    os.system(cmd)
    print()

def create_webapp(name, port):
    print(name)
    if not re.search(r"^[a-z0-9.]+\Z", name):
        raise Exception("allowed characters in app name: a-z 0-9 .")

    target = "/srv/%s" % name
    service = name

    if os.path.exists(target):
        raise Exception("path %s already exists" % target)

    execute("apt-get update")
    execute("apt-get install -y uwsgi-plugin-python3 python3-jinja2")
    execute("getent group bloonix >/dev/null || /usr/sbin/groupadd bloonix")
    execute(
        "getent passwd bloonix >/dev/null || /usr/sbin/useradd "
        "bloonix -g bloonix -d /var/lib/bloonix -s /bin/bash -r")

    create_path("%s/" % target)
    create_path("%s/bin" % target)
    create_path("%s/etc" % target)
    create_path("%s/app" % target)
    create_path("%s/log" % target)
    create_path("%s/templates" % target)
    create_path("%s/public" % target)
    create_path("%s/lib/bloonix_webapp/model" % target)
    create_path("%s/lib/bloonix_webapp/view" % target)
    create_path("%s/lib/bloonix_webapp/controller/root" % target)
    create_path("%s/lib/bloonix_webapp/plugin" % target)

    create_file(
        "%s/build" % target,
        build_template.format(name=name),
        "0750")

    create_file(
        "%s/etc/uwsgi.ini" % target,
        uwsgi_template.format(target=target, port=port))

    create_file(
        "%s/etc/main.conf" % target,
        main_template.format(target=target))

    create_file(
        "%s/etc/webapp.service" % target,
        systemd_template.format(target=target, name=name))

    create_file(
        "%s/etc/nginx.conf" % target,
        nginx_template.format(target=target, name=name, port=port))

    create_file(
        "%s/app/heaven.py" % target,
        heaven_template)

    create_file(
        "%s/templates/html-header.html" % target,
        html_header_template)

    create_file(
        "%s/templates/html-content.html" % target,
        html_content_template)

    create_file(
        "%s/templates/html-footer.html" % target,
        html_footer_template)

    create_file(
        "%s/templates/html-wrapper.html" % target,
        html_wrapper_template)

    create_file(
        "%s/templates/index.html" % target,
         html_index_template)

    create_file(
        "%s/templates/404.html" % target,
         html_404_template)

    create_file(
        "%s/lib/bloonix_webapp/__init__.py" % target,
        webapp_init_template)

    create_file(
        "%s/lib/bloonix_webapp/controller/root/__init__.py" % target,
        controller_init_template)

    execute("chown bloonix:bloonix %s/log" % target)
    execute("chown bloonix:bloonix %s/log" % target)

    create_file(
        "%s/bin/init-systemd" % target,
        bin_init_systemd_template.format(name=name, service=service),
        "0750")

    print("\n\n##### NOTICE #####\n")
    print("Create the systemd service with the following commands:\n")
    print("  {target}/bin/init-systemd".format(target=target))
    print("\n")

if __name__ == "__main__":
    argv = list(sys.argv)
    argv.pop(0)
    name = None
    port = 9000
    force = None
    while len(argv):
        opt = argv.pop(0)
        if opt == "--name":
            name = argv.pop(0)
        elif opt == "--port":
            port = argv.pop(0)
        elif opt == "--force":
            force = True
        else:
            print("ERR: invalid option: %s" % opt)
            sys.exit(1)
    if name == None or force == None:
        print("Usage: python3 -m bloonix.create_webapp [ options ]")
        print("Options:")
        print("--name <name>")
        print("  The name is used as directory /srv/$name and as service name.")
        print("--port <port>")
        print("  Default: 9000")
        print("--force")
        print("  Force the installation")
        sys.exit(1)
    create_webapp(name, port)
