import os
import re
import time
import traceback
from bloonix.logger import logger
from bloonix.dbi.validate import validate_config

"""
Please use always connect and disconnect():

    while True:
        dbi.connect()
        do_your_stuff()
        dbi.disconnect()

Control the connection handling using the parameter
persistent=yes|no. The default is 'no' for MySQL/MariaDB
and 'yes' for PostgreSQL.

Why?

MySQL/MariaDB use the SO_LINGER socket flag, which causes
the local port to enter the TIME_WAIT state after a connection
is closed. In high-traffic applications, this can quickly lead
to local ports being exhausted. For MySQL/MariaDB persistent=yes
should be set.

PostgreSQL, on the other hand, releases the local port immediately
after closing a connection. Therefore, it is generally recommended
to explicitly call connect() and disconnect() for PostgreSQL.
The setting persistent=no is especially recommended for PostgreSQL
to avoid unnecessary resource consumption.
"""

class dbiBase:
    def __init__(self, **kwargs):
        self.config = validate_config(kwargs)
        self.max_conn_age = self.config["max_connection_age"]
        self.max_conn_uses = self.config["max_connection_uses"]
        self.max_uses_counter = None
        self.max_age_expire = None

        if self.config["logger"]:
            self.log = logger(self.config["logger_name"])
            self.log.config(self.config["logger"])
        else:
            self.log = logger("bloonix")

        self.pid = os.getpid()
        self.errstr = None
        self.dbh = None
        self._autocommit_default = True

    def autocommit_default(self, value):
        self._autocommit_default = value

    def execute(self, stmt, bind=None):
        # After fork()
        if self.dbh and self.pid != os.getpid():
            self.log.debug(
                "process id changed from %s to %s"
                % (self.pid, os.getpid()))
            self.pid = os.getpid()
            self.disconnect(force=True)
            self.connect()
        #if self.dbh is None:
        #    self._connect()
        self.log.debug("sql exec:", stmt, bind)
        return self._execute(stmt, bind)

    def ping(self):
        try:
            self.log.debug("ping database")
            cursor = self.cursor()
            cursor.execute("select 1;")
            cursor.fetchall()
            return True
        except Exception as err:
            self.log.debug("ping to database failed:", err)
            return None

    def connect(self):
        if self.config["persistent"]:
            if self.max_uses_counter is None:
                self.max_uses_counter = 0
                self.max_age_expire = time.time() + self.max_conn_age

            self.max_uses_counter += 1

            if (self.max_uses_counter >= self.max_conn_uses
                    or self.max_age_expire <= time.time()):
                self.max_uses_counter = 0
                self.max_age_expire = time.time() + self.max_conn_age
                self.disconnect()

        if self.dbh and self.ping():
            return

        self._connect()

    def disconnect(self, force=None):
        if self.dbh is None or (self.config["persistent"] and not force):
            return

        dbh = self.dbh
        self.dbh = None
        self.log.info("disconnect from database")

        try: dbh.close()
        except Exception: self.log.exception()

    def fetchall(self, stmt, bind=None):
        cursor = self.execute(stmt, bind)
        return cursor.fetchall()

    def fetchone(self, stmt, bind=None):
        cursor = self.execute(stmt, bind)
        return cursor.fetchone()

    def count(self, stmt, bind=None):
        count = self.fetchone(stmt, bind)
        return count["count"]

    def commit(self):
        self.dbh.commit()

    def rollback(self):
        self.dbh.rollback()

    def run_transaction(self, lock_table, callback, *args, **kwargs):
        """
        Args:
            lock_table={
                table_name_1: lock_type,
                table_name_2: lock_type,
                table_name_3: lock_type
            }
            callback=function
            *args     passed to callback
            **kwargs  passed to callback
        Returns the return value of the callback.
        """
        ret = None

        if self.dbh is None:
            self.connect()

        try:
            self.dbh.autocommit = False
            self.lock(lock_table)
            ret = callback(*args, **kwargs)
            self.commit()
            self.unlock(lock_table)
        except Exception:
            try:
                self.rollback()
            except Exception:
                self.log.exception(message="rollback failed")

            try:
                self.unlock(lock_table)
            except Exception:
                self.log.exception(message="unlock failed")

            self.disconnect(force=True)
            raise Exception(message="transaction failed")
        finally:
            self.dbh.autocommit = self._autocommit_default

        return ret

    def _validate_table_lock_definition(self, table):
        for table_name in table:
            lock_type = table[table_name]
            if (not isinstance(table_name, str)
                    or not re.search(r"^[a-zA-Z0-9_]+\Z", table_name)
                    or not re.search(r"^(read|write)\Z", lock_type)):
                raise Exception("invalid lock definition")

    def add_range_partition(self, table, name, from_value, to_value, index_on=None):
        """
        index_on=[
            "foo", 
            "bar",
            ["foo", "bar"]
        ]
        """
        if (not re.search(r"^[a-zA-Z_0-9]+\Z", table)
                or not re.search(r"^[a-zA-Z_0-9]+\Z", name)
                or not isinstance(from_value, int)
                or not isinstance(to_value, int)):
            raise Exception("invalid partition arguments")
        if index_on:
            for indexes in index_on:
                if isinstance(indexes, str):
                    indexes = [indexes]
                for index in indexes:
                    if not re.search(r"^[a-zA-Z_0-9]+\Z", index):
                        raise Exception("invalid index_on arguments")
        self._add_range_partition(table, name, from_value, to_value, index_on)

    def find_partitions(self, table, regex):
        if not re.search(r"^[a-zA-Z_0-9]+\Z", table):
            raise Exception("invalid table name")
        return self._find_partitions(table, regex)

    def delete_partition(self, *args):
        for name in args:
            if not re.search(r"^[a-zA-Z_0-9]+\Z", name):
                raise Exception("invalid table/partition name")
        return self._delete_partition(*args)
