from bloonix.dbi import DBI
from bloonix.findmod import load_path
from bloonix.logger import logger
from bloonix.sql import SQL
import importlib.util
import re

"""
The directory structure looks like:

    lib/webapp/model/db/__init__.py
    lib/webapp/model/db/user.py
    lib/webapp/model/db/songlist.py

The module __init__.py should looks like

    from bloonix.dbi.classloader import ClassLoader
    class Schema(ClassLoader):
        def __init__(self, db_config):
            super().__init__(__path__[0], db_config)

With the autoload() method all other python modules in the same
directory will be loaded. The file names should be identically
to the table names. As example if you have a table name called
"user" or "songlist" then the python modules should be named
identically "user.py" and "songlist.py". The modules can
be accessed via

    model = ClassLoader("/path/to/my/modules", config)
    model.user
    model.songlist

After the table module is loaded some attributes will be set.
Please note that no table module should have this names:

    model.user.dbi       - Accessor to bloonix.dbi
    model.user.log       - Accessor to bloonix.logger
    model.user.schema    - Accessor to the class loader itself.
                           model.user and model.schema.user is the same,
                           so it's possible to access other table modules.
    model.user.sql       - Accessor to bloonix.sql

Note that config, log, sql and dbi are reserved:

    model.config
    model.log
    model.sql
    model.dbi

is an direct accessor to the database and shouldn't be uses as module.
"""

class ClassLoader:
    # Path is needed to load all table modules.
    # This is normaly lib/$project_name/model/$schema
    def __init__(self, path, config):
        self.config = config
        self.log = logger("bloonix-database")
        self.sql = SQL(config["driver"])
        self.dbi = DBI.setup(**config)

        for module_name in load_path(path):
            # connect() does not exist currently, but maybe
            # it will be implemented one time.
            if module_name == "connect":
                raise Exception(
                    "module name %s (%s) is reserved for "
                    "internal usage and not allowed"
                    % (module_name, module))

            import_module = "%s/%s.py" % (path, module_name)
            self.log.notice("import module", import_module)
            spec = importlib.util.spec_from_file_location("module.name", import_module)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            # Try to find the class name of the table module.
            #       user = User
            #       user_session = UserSession
            # If the class name is not present then we try
            # to try to load class name "Table".
            class_name = module_name.title()
            class_name = re.sub("_", "", class_name)
            #if not hasattr(module, class_name):
            #    class_name = "Table"

            try:
                self.log.notice("load model class", module_name, class_name)
                attr = getattr(module, class_name)
            except Exception as err:
                try:
                    self.log.notice("load model class", module_name, "Table")
                    attr = getattr(module, "Table")
                except Exception:
                    raise Exception(
                        "unable to load class '%s' or class "
                        "'Table' from module '%s': %s"
                        % (class_name, import_module, err))
            instance = attr()
            setattr(self, module_name, instance)

            for key in ["log", "dbi", "sql", "schema", "table"]:
                if hasattr(instance, key):
                    raise Exception(
                        "method name %s in %s is reserved"
                        % (key, module))

            setattr(instance, "dbi", self.dbi)
            setattr(instance, "log", self.log)
            setattr(instance, "schema", self)
            setattr(instance, "sql", self.sql)

            # Filename user.py
            # self.table = "user"
            if not hasattr(instance, "table"):
                setattr(instance, "table", module_name)
