import copy
import re
from bloonix.genattr import GenAttr

class CRUD:
    def get(self, value, column="id", columns="*"):
        """ get a row by id: get(id) """
        stmt, bind = self.sql.get(self.table, value, column, columns)
        return self.dbi.fetchone(stmt, bind)

    def find(self, *args, **kwargs):
        """
        get a row by column and value, combined with and
            find(column_name1, value1, column_name2, value2)
        """
        stmt, bind = self.sql.find(self.table, *args, **kwargs)
        return self.dbi.fetchone(stmt, bind)

    def findall(self, *args, **kwargs):
        stmt, bind = self.sql.find(self.table, *args, **kwargs)
        return self.dbi.fetchall(stmt, bind)

    def fetchall(self, **kwargs):
        if "table" not in kwargs or isinstance(kwargs["table"], str):
            kwargs["table"] = [
                kwargs.get("table", self.table),
                kwargs.get("columns", "*")
            ]
        stmt, bind = self.sql.select(**kwargs)
        return self.dbi.fetchall(stmt, bind)

    def fetchone(self, **kwargs):
        if "table" not in kwargs or isinstance(kwargs["table"], str):
            kwargs["table"] = [
                kwargs.get("table", self.table),
                kwargs.get("columns", "*")
            ]
        stmt, bind = self.sql.select(**kwargs)
        return self.dbi.fetchone(stmt, bind)

    def count(self, **kwargs):
        """ count(count="id") """
        if "table" not in kwargs:
            kwargs["table"] = self.table
        stmt, bind = self.sql.select(**kwargs)
        return self.dbi.count(stmt, bind)

    # Insert data into tables without auto increment tables
    def create(self, data=None, **kwargs):
        if data is None:
            data = kwargs
        stmt, bind = self.sql.insert(self.table, data)
        return self.dbi.do(stmt, bind)

    # Insert data into tables with a auto increment columm
    # and get the last inserted row. That is only possible
    # if a single row is inserted.
    def create_and_get(self, data=None):
        stmt, bind = self.sql.insert(self.table, data)
        last_insert_id = self.dbi.do(stmt, bind, get_last_insert_id=True)
        return self.find("id", last_insert_id)

    def update_by_id(self, *args, **kwargs):
        stmt, bind = self.sql.update_by_id(self.table, *args, **kwargs)
        return self.dbi.do(stmt, bind)

    def update_by_col(self, *args, **kwargs):
        stmt, bind = self.sql.update_by_col(self.table, *args, **kwargs)
        return self.dbi.do(stmt, bind)

    def update(self, data, condition):
        stmt, bind = self.sql.update(self.table, data, condition)
        return self.dbi.do(stmt, bind)

    def delete_by_id(self, value):
        stmt, bind = self.sql.delete_by_id(self.table, value)
        return self.dbi.do(stmt, bind)

    def delete_by_col(self, *args):
        cond = []
        args = list(args)

        if not len(args):
            raise Exception("no column and values in argument list")

        while len(args):
            column = args.pop(0)
            value = args.pop(0)
            if len(cond):
                cond.append("and")
            else:
                cond.append("where")
            cond.append([column, value])

        stmt, bind = self.sql.delete(self.table, cond)
        return self.dbi.do(stmt, bind)

    def delete(self, condition):
        stmt, bind = self.sql.delete(self.table, condition)
        return self.dbi.do(stmt, bind)

    """
    The query method has the same arguments like
    sql.select() with 3 extra arguments:

        parse_int=[ columns whose values must be int ]
        exclude=[ columns to be excluded from sort and query ]
        query={
            "limit": 10,
            "offset": 0,
            "order": ["hostname", "asc"],
            "query": [
                "hostname", "like", "loadbalancer",
                "or",
                "id", ">", 6000
            ]
        }

    "limit" and "offset" will be passed to sql.select().
    "sort" will be converted into "order" and passed to sql.select().
    "query" will be parsed and appended to the pre-defined condition
    with the operator "and" and within round brackets: 'and (..query..)'
    """
    def query(self, query, **kwargs):
        ret = GenAttr(data=False, total=False, success=False)
        select = self.parse_query(query, **kwargs)
        if not select:
            return ret
        stmt, bind = self.sql.select(**select[0])
        ret.data = self.dbi.fetchall(stmt, bind)
        stmt, bind = self.sql.select(**select[1])
        ret.total = self.dbi.count(stmt, bind)
        ret.success = True
        return ret

    def parse_query(self, query, **kwargs):
        if ("table" in kwargs
                and isinstance(kwargs["table"], str)
                and "columns" in kwargs):
            kwargs["table"] = [kwargs["table"], kwargs["columns"]]
            del kwargs["columns"]

        columns = {}
        if "count" in kwargs:
            count = kwargs["count"]
            del kwargs["count"]
        else:
            count = "id"

        tables = list(kwargs["table"])
        while len(tables):
            table = tables.pop(0)
            cols = tables.pop(0)
            if not isinstance(cols, list):
                cols = [cols]
            for col in cols:
                # coalesce(max(name), 1)
                if isinstance(col, dict):
                    columns[col.get("as", col["value"]["column"])] = "%s.%s" % (
                        table, col["value"]["column"])
                else:
                    m = re.search(r"(.+)\s+(as|AS)\s+(.+)", col)
                    if m:
                        columns[m.group(3)] = "%s.%s" % (table, m.group(1))
                    elif col == "*":
                        raise Exception("'*' is not allowed in query")
                    else:
                        columns[col] = "%s.%s" % (table, col)

        if "parse_int" in kwargs:
            parse_int = kwargs["parse_int"]
            del kwargs["parse_int"]
            if not isinstance(parse_int, list):
                self.log.error("parse query: 'parse_int' is not a list")
                return None
        else:
            parse_int = []

        if "skip" in kwargs:
            skip = kwargs["skip"]
            del kwargs["skip"]
            if not isinstance(skip, list):
                self.log.error("parse query: 'skip' is not a list")
                return None
        else:
            skip = []

        if ("query" in query and isinstance(query["query"], list)
                and len(query["query"]) > 0):
            condition = []
            parts = list(query["query"])

            maps = {
                "==": "=",
                "!=": "!=",
                "<": "<",
                "<=": "<=",
                ">": ">",
                ">=": ">=",
                "=~": "ilike",
                "!~": "not ilike",
                "eq": "=",
                "ne": "!=",
                "lt": "<",
                "le": ">=",
                "gt": ">",
                "ge": ">=",
                "like": "like",
                "ilike": "ilike"}

            while len(parts) > 2:
                column = parts.pop(0)
                op = parts.pop(0)
                value = str(parts.pop(0))

                if column not in columns or column in skip:
                    self.log.error(
                        "parse query: invalid column '%s' %s"
                        % (column, columns))
                    return None

                if column in parse_int and not re.search(r"^\d+(?:\.\d+){0,1}\Z", value):
                    self.log.error(
                        "parse query: value of column %s is not of type int/float"
                        % column)
                    return None

                if maps.get(op, None) == None:
                    self.log.error(
                        "parse query: invalid op '%s' %s"
                        % (op, maps))
                    return None

                op = maps[op]

                if (op == "like"
                        or op == "ilike"
                        or op == "not like"
                        or op == "not ilike"):
                    value, a = re.subn(r"^\^", "", value)
                    value, b = re.subn(r"\$\Z", "", value)
                    if a and not b:
                        value = "%s%%" % value
                    elif b and not a:
                        value = "%%%s" % value
                    elif not a and not b:
                        value = "%%%s%%" % value

                condition.append([columns[column], op, value])

                if len(parts) > 3:
                    ao = parts.pop(0)

                    if ao not in ["&&", "||", "and", "or"]:
                        self.log.debug(
                            "parse query: invalid condition <%s>"
                            % ao)
                        return None

                    if ao == "&&":
                        condition.append("and")
                    elif ao == "||":
                        condition.append("or")
                    else:
                        condition.append(ao)

            if "condition" not in kwargs:
                kwargs["condition"] = ["where"]
            else:
                kwargs["condition"].append("and")

            kwargs["condition"].append("(")
            kwargs["condition"] += condition
            kwargs["condition"].append(")")

        if ("sort" in query
                and isinstance(query["sort"], list)
                and len(query["sort"]) % 2 == 0):
            order = []
            while len(query["sort"]):
                sort_by = query["sort"].pop(0)
                sort_type = query["sort"].pop(0)
                if sort_by in columns and (sort_type == "asc" or sort_type == "desc"):
                    order.append(sort_type)
                    order.append(sort_by)
            if len(order):
                kwargs["order"] = order

        if kwargs.get("unlimited", None) == None:
            kwargs["limit"] = query["limit"]
            kwargs["offset"] = query["offset"]

        select = kwargs
        select_count = copy.deepcopy(kwargs)

        if "join" in select:
            j = select["join"]
            select["join"] = []
            select_count["join"] = []
            while(len(j)):
                join_type = j.pop(0)
                join = j.pop(0)
                if join_type[0] == "-": # -inner, -left, -right, -outer
                    join_type = join_type[1:]
                    select["join"].append(join_type)
                    select["join"].append(join)
                else:
                    select["join"].append(join_type)
                    select["join"].append(join)
                    select_count["join"].append(join_type)
                    select_count["join"].append(join)

        for key in ["order", "group", "limit", "offset"]:
            if key in select_count:
                del select_count[key]

        select_count["table"] = self.table
        select_count["count"] = count

        if isinstance(count, str):
            m = re.search(r"^([a-zA-Z_0-9]+)\.([a-zA-Z_0-9]+)\Z", count)
            if m:
                select_count["table"] = m.group(1)
                select_count["count"] = m.group(2)

        return select, select_count

    # name = column name used as form:label
    # value = column name used as form:value
    def form_data(self, label, value):
        data = []
        rows = self.fetchall(
            columns=[label, value],
            order=["asc", label])
        for row in rows:
            data.append({"label": row[label], "value": row[value]})
        return data

    # form_data_ext(
    #     table=[
    #         "user", [
    #             "id as value",
    #             "name as label"
    #         ]
    #     ],
    #     order=["asc", "label"]
    # )
    def form_data_ext(self, **kwargs):
        data = []
        rows = self.fetchall(**kwargs)
        for row in rows:
            data.append({"label": row["label"], "value": row["value"]})
        return data


class CRUDV(CRUD):
    """
    Init Validator at first:
        self.validator = Validator()
    """
    def get_options(self):
        return self.validator.get_options()

    def get_defaults(self):
        return self.validator.get_defaults()

    def validate(self, data, ignore_missing=None):
        return self.validator.validate(data, ignore_missing=ignore_missing)
