import re
import mariadb
from mariadb.constants import FIELD_TYPE
from bloonix.dbi.base import dbiBase
from bloonix.json import json
from bloonix.exception import BloonixDatabaseError, BloonixDuplicationError

def _json_convert(s):
    return json.loads(s)

class DBI(dbiBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.config["persistent"] is None:
            self.config["persistent"] = 1

    def _connect(self):
        self.log.info("connect to database")
        self.dbh = mariadb.connect(
            user=self.config["user"],
            password=self.config["password"],
            host=self.config["host"],
            port=self.config["port"],
            database=self.config["database"],
            converter={FIELD_TYPE.JSON: _json_convert})
        self.dbh.autocommit = self._autocommit_default

    def _execute(self, stmt, bind):
        try:
            cursor = self.cursor()
            cursor.execute(stmt, bind)
            return cursor
        except mariadb.IntegrityError as e:
            if e.errno == 1062:
                raise BloonixDuplicationError(e)
            raise BloonixDatabaseError("errno=%s" % e.errno, e)

    def do(self, stmt, bind=None, get_last_insert_id=None):
        cursor = self.execute(stmt, bind)
        if get_last_insert_id is True:
            return cursor.lastrowid
        return True

    def cursor(self):
        return self.dbh.cursor(dictionary=True, buffered=True)

    def lock(self, table):
        self._validate_table_lock_definition(table)
        tables = []
        for t in table:
            tables.append("`%s` %s" % (t, table[t]))
        self.do("lock table %s" % ", ".join(tables))

    def unlock(self, table=None):
        self.do("unlock tables")

    def _has_so_linger(self):
        return True

    def _add_range_partition(self, table, name, from_value, to_value, index_on):
        self.do(
            'alter table `%s` add partition ('
            'partition `%s` values less than (%s))'
            % (table, name, to_value))

    def _find_partitions(self, table, regex):
        data = []
        rows = self.fetchone("show create table `%s`" % table)
        for key, value in rows.items():
            for line in re.split("\n", value):
                m = re.search(r"PARTITION\s+([^\s]+)", line)
                if m:
                    name = re.sub(r"`", "", m.group(1))
                    if re.search(regex, name):
                        data.append(name)
        return data

    def _delete_partition(self, table, name):
        self.do('alter table `%s` drop partition `%s`' % (table, name))
