"""
This module is specially created to calculate from-to times
for table paritions. For example, if a PostgreSQL partition is
called table_p202101, the time string 202101 can be used to
calculate the period for which the partition stores data.
"""

import re
import time
from bloonix import utils

def date_to_unixtime(partition):
    """
    Range partitions are always beween
        >= time <

    Partition format:
        table_name_p%Y%m%d

    Example calls:
        partition_date_to_unixtime("table_p2021")
        partition_date_to_unixtime("table_p202101")
        partition_date_to_unixtime("table_p20210101")

    Returns min and max time of the current table partition
        2021      -  min 2021-01-01 00:00:00  max 2022-01-01 00:00:00
        202101    -  min 2021-01-01 00:00:00  max 2021-02-01 00:00:00
        20210101  -  min 2021-01-01 00:00:00  max 2021-01-02 00:00:00
    """
    m = re.search(r"(?:^|_p)(\d+)\Z", partition)
    if not m:
        raise Exception("invalid table partition format: %s" % partition)
    date = m.group(1)
    f, t = get_from_to_year_month_day(date)
    f = time.strptime(f, "%Y-%m-%d %H:%M:%S")
    f = time.mktime(f)
    t = time.strptime(t, "%Y-%m-%d %H:%M:%S")
    t = time.mktime(t)
    t += 1
    return int(f), int(t)

def get_from_to_year_month_day(date):
    m = re.search(r"^(\d{4})(\d\d)(\d\d)\Z", date)
    if m:
        year = m.group(1)
        month = m.group(2)
        day = m.group(3)
        return (
            "%s-%s-%s 00:00:00" % (year, month, day),
            "%s-%s-%s 23:59:59" % (year, month, day))

    m = re.search(r"^(\d{4})(\d\d)\Z", date)
    if m:
        year = m.group(1)
        month = m.group(2)
        day = utils.get_last_day_of_month(int(year), int(month))
        return (
            "%s-%s-01 00:00:00" % (year, month),
            "%s-%s-%s 23:59:59" % (year, month, day))

    m = re.search(r"^(\d{4})\Z", date)
    if m:
        year = m.group(1)
        return (
            "%s-01-01 00:00:00" % year,
            "%s-12-31 23:59:59" % year)

def days_from_now(num):
    t = time.localtime( int(time.time()) + (86400 * num) )
    from_time = utils.unixtime("%s-%02d-%02d 00:00:00" % (t.tm_year, t.tm_mon, t.tm_mday))
    to_time = utils.unixtime("%s-%02d-%02d 23:59:59" % (t.tm_year, t.tm_mon, t.tm_mday))
    return {
        "from": from_time,
        "to": to_time + 1,
        "year": t.tm_year,
        "month": str(t.tm_mon).zfill(2),
        "day": str(t.tm_mday).zfill(2)
    }

def days_back_to(num):
    t = time.localtime( int(time.time()) - (86400 * num) )
    from_time = utils.unixtime("%s-%02d-%02d 00:00:00" % (t.tm_year, t.tm_mon, t.tm_mday))
    to_time = utils.unixtime("%s-%02d-%02d 23:59:59" % (t.tm_year, t.tm_mon, t.tm_mday))
    return {
        "from": from_time,
        "to": to_time + 1,
        "year": t.tm_year,
        "month": str(t.tm_mon).zfill(2),
        "day": str(t.tm_mday).zfill(2)
    }

def months_from_now(num):
    t = time.localtime()
    y = t.tm_year
    m = t.tm_mon
    for i in range(num):
        m += 1
        if m == 13:
            m = 1
            y += 1
    td = str(utils.get_last_day_of_month(y, m)).zfill(2)
    m = str(m).zfill(2)
    y = str(y)
    from_time, to_time = date_to_unixtime("%s%s" % (y + m))
    return {
        "from": from_time,
        "to": to_time,
        "year": y,
        "month": m,
        "from_day": "01",
        "to_day": td
    }

def months_back_to(num):
    t = time.localtime()
    y = t.tm_year
    m = t.tm_mon
    for i in range(num):
        m -= 1
        if m == 0:
            m = 12
            y -= 1
    td = str(utils.get_last_day_of_month(y, m)).zfill(2)
    m = str(m).zfill(2)
    y = str(y)
    from_time, to_time = date_to_unixtime("%s%s" % (y + m))
    return {
        "from": from_time,
        "to": to_time,
        "year": y,
        "month": m,
        "from_day": "01",
        "to_day": td
    }
