import re
import psycopg2
import psycopg2.extras
from bloonix.dbi.base import dbiBase
from bloonix.exception import BloonixDatabaseError, BloonixDuplicationError

class DBI(dbiBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.config["persistent"] is None:
            self.config["persistent"] = 0

    def _connect(self):
        self.log.info("connect to database")
        self.dbh = psycopg2.connect(
            "host='%s' port='%s' dbname='%s' user='%s' password='%s'" % (
                self.config["host"],
                self.config["port"],
                self.config["database"],
                self.config["user"],
                self.config["password"]
            ))
        self.dbh.autocommit = self._autocommit_default

    def _execute(self, stmt, bind):
        try:
            cursor = self.cursor()
            cursor.execute(stmt, bind)
            return cursor
        except psycopg2.Error as e:
            if e.pgcode == "23505":
                raise BloonixDuplicationError(e)
            raise BloonixDatabaseError("pgcode=%s" % e.pgcode, e)

    def do(self, stmt, bind=None, get_last_insert_id=None):
        if get_last_insert_id is True:
            stmt += " returning id"
        cursor = self.execute(stmt, bind)
        if get_last_insert_id is True:
            row = cursor.fetchone()
            return row["id"]
        return True

    def cursor(self):
        return self.dbh.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

    def lock(self, table):
        self._validate_table_lock_definition(table)
        for table_name in table:
            if table[table_name] == "write":
                lock_type = "in exclusive mode"
            elif table[table_name] == "share":
                lock_type = "in share mode"
            self.do('lock table "%s" %s' % (table_name, lock_type))

    def unlock(self, table=None):
        """
        There is no UNLOCK TABLE command; locks are
        always released at transaction end.
        """
        return

    def _has_so_linger(self):
        return False

    def _add_range_partition(self, table, name, from_value, to_value, index_on):
        self.do(
            'create table "%s" partition of %s '
            'for values from (%s) to (%s)'
            % (name, table, from_value, to_value))
        #for indexes in index_on:
        #    if isinstance(indexes, str):
        #        indexes = [indexes]
        #    index_name = "%s_%s_index" % (name, "_".join(indexes))
        #    columns = []
        #    for column in indexes:
        #        columns.append('"%s"' % column)
        #    self.do(
        #        'create index "%s" on "%s" (%s)'
        #        % (index_name, name, ",".join(columns)))

    def _find_partitions(self, table, regex):
        data = []
        rows = self.fetchall(
            "SELECT relname FROM pg_class "
            "WHERE relkind='r' AND relname !~ '^(pg_|sql_)'")
        for row in rows:
            if re.search(regex, row["relname"]):
                data.append(row["relname"])
        return data

    def _delete_partition(self, table, name):
        self.do('drop table "%s"' % name)
