from multiprocessing import Process, Pipe
import os
import signal
import sys
import time
from bloonix.logger import logger
from bloonix import utils

class Dispatcher:
    def __init__(self, worker=10, worker_timeout=300,
            exit_on_sig_term=True, friendly_sleep_max=1):
        self.num_worker = worker
        self.worker_timeout = worker_timeout
        self.exit_on_sig_term = exit_on_sig_term
        self.worker = {}
        self.ready_workers = []
        self.jobs = []
        self.jobs_in_progress = {}
        self.done = False
        self.cpu_friendly_sleep_counter = 0
        self.friendly_sleep_max = friendly_sleep_max
        self.reload_in_progress = False
        self.log = logger("bloonix")
        self.callback = Callback()

    def stop(self):
        self.done = True

    def on(self, action, callback):
        if action not in ["init", "ready", "process", "finish", "reload"]:
            raise Exception("invalid action %s" % action)
        setattr(self.callback, action, callback)

    def run(self):
        self.init_parent_signal_handler()
        while self.done is False:
            for name in [
                    "manage_workers",
                    "manage_worker_requests",
                    "manage_jobs"]:
                if self.done is False:
                    method = getattr(self, name)
                    try:
                        method()
                    except Exception:
                        self.log.exception()
                        time.sleep(1)
        self.stop_worker()

    def add_job(self, job):
        if job is not None:
            self.jobs.append(job)

    def manage_workers(self):
        for pid in list(self.worker):
            process = self.worker[pid]["process"]
            self.log.debug(
                "dispatcher: check if worker is alive",
                pid, self.worker[pid]["start"])

            if not process.is_alive():
                process.join()
                del self.worker[pid]["parent_conn"]
                del self.worker[pid]["child_conn"]
                del self.worker[pid]
                if pid in self.jobs_in_progress:
                    job = self.jobs_in_progress[pid]
                    del self.jobs_in_progress[pid]
                    self.log.warning(
                        "dispatcher: child %s died with running job" % pid, job)
                    if pid in self.ready_workers:
                        del self.ready_workers[self.ready_workers.index(pid)]
                    self.callback.finish(False, job)
            elif (self.worker[pid]["start"] > 0
                    and self.worker[pid]["start"]
                        + self.worker_timeout < time.time()):
                try:
                    os.kill(pid, signal.SIGKILL)
                except Exception:
                    pass
                self.log.warning("dispatcher: child %s runs on timeout, killed" % pid)

        if len(self.worker) < self.num_worker:
            self.spawn_worker(self.num_worker - len(self.worker))

    def spawn_worker(self, count):
        self.log.notice("dispatcher: spawn %s workers" % count)

        for i in range(count):
            parent_conn, child_conn = Pipe()
            process = Process(target=self.run_worker, args=(child_conn,))
            process.start()
            pid = process.pid
            self.worker[pid] = {
                "process": process,
                "parent_conn": parent_conn,
                "child_conn": child_conn,
                "start": 0
            }
            self.ready_workers.append(pid)
            self.log.info("dispatcher: spawned child %s" % pid)

    def manage_worker_requests(self):
        for pid in self.worker:
            self.log.debug("dispatcher: check worker", pid)
            parent_conn = self.worker[pid]["parent_conn"]
            if parent_conn.poll():
                self.log.debug("dispatcher: start recv from pipe")
                job = parent_conn.recv()
                self.log.debug("dispatcher: data received")
                if job:
                    self.log.debug("dispatcher: recv data", pid, job)
                    self.worker[pid]["start"] = 0
                    self.log.debug("dispatcher: reset start value for worker", pid)
                    self.ready_workers.append(pid)
                    self.callback.finish(True, job)

    def manage_jobs(self):
        if len(self.jobs) < len(self.ready_workers):
            jobs = self.callback.ready(len(self.ready_workers) - len(self.jobs))

            if isinstance(jobs, list):
                for job in jobs:
                    self.add_job(job)
            elif jobs is not None:
                self.jobs.append(jobs)

        if len(self.ready_workers) == 0 or len(self.jobs) == 0:
            time.sleep(self.cpu_friendly_sleep_counter)
            if self.cpu_friendly_sleep_counter < self.friendly_sleep_max:
                self.cpu_friendly_sleep_counter += 0.05
            return

        self.cpu_friendly_sleep_counter = 0

        while (self.done is False
                and len(self.ready_workers) > 0
                and len(self.jobs) > 0):
            job = self.jobs.pop(0)
            pid = self.ready_workers.pop(0)
            self.worker[pid]["start"] = time.time()
            self.worker[pid]["parent_conn"].send(job)
            self.log.debug(
                "dispatcher: send ready job to worker",
                pid, job, self.worker[pid]["start"])
            self.jobs_in_progress[pid] = job

    def stop_worker(self):
        # at first send SIGINT to all worker
        self.kill_worker(time.time() + 15, signal.SIGTERM)

        # if a worker doesn't exit after 15 seconds,
        # SIGKILL is sent to all worker
        self.kill_worker(0, signal.SIGKILL)

    def kill_worker(self, timeout, kill_signal):
        pids = list(self.worker)

        if len(pids) == 0:
            return

        pids = map(str, pids)
        pids = ", ".join(pids)
        self.log.info("dispatcher: send signal %s to pids:" % kill_signal, pids)

        for pid in self.worker:
            try:
                os.kill(pid, kill_signal)
            except Exception:
                pass

        while len(self.worker) and (timeout == 0 or timeout > time.time()):
            pids = map(str, list(self.worker))
            pids = ", ".join(pids)
            for pid in list(self.worker):
                worker = self.worker[pid]["process"]
                if not worker.is_alive():
                    worker.join()
                    del self.worker[pid]
            time.sleep(0.1)

    def sig_hup_handler(self, signum, frame):
        if self.reload_in_progress is True:
            self.log.info("dispatcher: signal HUP received - reload already in progress")
            return

        self.reload_in_progress = True
        self.log.info("dispatcher: signal HUP received - reloading")

        try:
            self.callback.reload()
        except Exception as err:
            self.log.error("dispatcher: reloading failed:", err)

        self.reload_in_progress = False

    def sig_term_handler(self, signum, frame):
        self.log.info("dispatcher: signal TERM received - quitting")
        self.done = True

    def sig_int_handler(self, signum, frame):
        self.log.info("dispatcher: signal INT received - quitting")
        self.done = True

    def sig_hup_handler_child(self, signum, frame):
        self.log.info("dispatcher: signal HUP received - ignoring")

    def sig_term_handler_child(self, signum, frame):
        self.log.info("dispatcher: signal TERM received - quitting")
        sys.exit(0)

    def sig_int_handler_child(self, signum, frame):
        self.log.info("dispatcher: signal INT received - quitting")
        sys.exit(0)

    def init_parent_signal_handler(self):
        signal.signal(signal.SIGHUP, self.sig_hup_handler)
        signal.signal(signal.SIGINT, self.sig_int_handler)
        signal.signal(signal.SIGTERM, self.sig_term_handler)

    def init_child_signal_handler(self):
        signal.signal(signal.SIGHUP, self.sig_hup_handler_child)
        if self.exit_on_sig_term:
            signal.signal(signal.SIGINT, self.sig_int_handler_child)
            signal.signal(signal.SIGTERM, self.sig_term_handler_child)

    def run_worker(self, child_conn):
        try:
            self.init_child_signal_handler()
            self.callback.init()
            while self.done is False:
                if child_conn.poll(1):
                    self.log.debug("dispatcher: start recv from pipe")
                    job = child_conn.recv()
                    self.log.debug("dispatcher: data received")
                    if job:
                        ret = self.callback.process(job)
                        if ret is None:
                            ret = job
                        child_conn.send(ret)
            self.log.info("dispatcher: good bye")
        except Exception:
            self.log.exception()


class Callback:
    def init(self):
        return

    def ready(self, num):
        return []

    def process(self, job):
        return job

    def finish(self, success, job):
        return

    def reload(self):
        return
