from multiprocessing import Process, Queue
from queue import Empty, Full
import os
import signal
import sys
import time
from bloonix.logger import logger

# ALTERNATIVE MODULE
# If Pipe() doesnt work on Win.
class Dispatcher:
    def __init__(self, worker=10, worker_timeout=300, exit_on_sig_term=True, friendly_sleep_max=1):
        self.num_worker = worker
        self.worker_timeout = worker_timeout
        self.exit_on_sig_term = exit_on_sig_term
        self.worker = {}
        self.ready_workers = []
        self.jobs = []
        self.jobs_in_progress = {}
        self.done = False
        self.cpu_friendly_sleep_counter = 0
        self.friendly_sleep_max = friendly_sleep_max
        self.reload_in_progress = False
        self.log = logger("bloonix")
        self.queue = Transport()
        self.callback = Callback()

    def on(self, action, callback):
        if action not in ["init", "ready", "process", "finish", "reload"]:
            raise Exception("invalid action %s" % action)
        setattr(self.callback, action, callback)

    def run(self):
        self.init_parent_signal_handler()
        while self.done is False:
            for name in [
                    "manage_workers",
                    "manage_worker_requests",
                    "manage_jobs"]:
                if self.done is False:
                    method = getattr(self, name)
                    method()
        self.stop_worker()

    def add_job(self, job):
        self.jobs.append(job)

    def manage_workers(self):
        for pid in list(self.worker):
            process = self.worker[pid]["process"]

            if not process.is_alive():
                process.join()
                self.queue.destroy_queue(self.worker[pid]["queue"])
                del self.worker[pid]
                if pid in self.jobs_in_progress:
                    job = self.jobs_in_progress[pid]
                    del self.jobs_in_progress[pid]
                    self.log.warning("child %s died with running job" % pid)
                    if pid in self.ready_workers:
                        del self.ready_workers[self.ready_workers.index(pid)]
                    self.callback.finish(False, job)
            elif (self.worker[pid]["start"] > 0
                    and self.worker[pid]["start"] < time.time()
                        - self.worker_timeout):
                try:
                    os.kill(pid, signal.SIGKILL)
                except Exception:
                    pass
                self.log.warning("child %s runs on timeout, killed" % pid)

        if len(self.worker) < self.num_worker:
            self.spawn_worker(self.num_worker - len(self.worker))

    def spawn_worker(self, count):
        self.log.notice("spawn %s workers" % count)

        for i in range(count):
            q = self.queue.create_queue()
            p = Process(target=self.run_worker, args=(q,))
            p.start()
            self.worker[p.pid] = {
                "process": p,
                "queue": q,
                "start": 0
            }
            self.ready_workers.append(p.pid)
            self.log.info("spawned child %s" % p.pid)

    def manage_worker_requests(self):
        while self.done is False:
            req = self.queue.recv_from_worker()

            if (not req or not isinstance(req, dict) or
                    "status" not in req or "pid" not in req or
                    req["pid"] not in self.worker):
                break

            pid = req["pid"]
            self.worker[pid]["start"] = 0
            self.ready_workers.append(pid)

            if req["status"] == "finish":
                self.callback.finish(True, req["job"])

    def manage_jobs(self):
        if len(self.jobs) < len(self.ready_workers):
            jobs = self.callback.ready(
                len(self.ready_workers) - len(self.jobs))

            if isinstance(jobs, list):
                self.jobs += jobs
            else:
                self.jobs.append(jobs)

        if len(self.ready_workers) == 0 or len(self.jobs) == 0:
            time.sleep(self.cpu_friendly_sleep_counter)
            if self.cpu_friendly_sleep_counter < self.friendly_sleep_max:
                self.cpu_friendly_sleep_counter += 0.05
            return

        self.cpu_friendly_sleep_counter = 0

        while (self.done is False
                and len(self.ready_workers) > 0
                and len(self.jobs) > 0):
            job = self.jobs.pop(0)
            pid = self.ready_workers.pop(0)
            self.worker[pid]["start"] = time.time()
            self.queue.send_to_worker(self.worker[pid]["queue"], job)
            self.jobs_in_progress[pid] = job

    def stop_worker(self):
        # at first send SIGINT to all worker
        self.kill_worker(time.time() + 15, signal.SIGTERM)

        # if a worker doesn't exit after 15 seconds,
        # SIGKILL is sent to all worker
        self.kill_worker(0, signal.SIGKILL)

    def kill_worker(self, timeout, kill_signal):
        pids = list(self.worker)

        if len(pids) == 0:
            return

        pids = map(str, pids)
        pids = ", ".join(pids)
        self.log.info("send signal %s to pids:" % kill_signal, pids)

        for pid in self.worker:
            try:
                os.kill(pid, kill_signal)
            except Exception:
                pass

        while len(self.worker) and (timeout == 0 or timeout > time.time()):
            pids = map(str, list(self.worker))
            pids = ", ".join(pids)
            for pid in list(self.worker):
                worker = self.worker[pid]["process"]
                if not worker.is_alive():
                    worker.join()
                    del self.worker[pid]
            time.sleep(0.1)

    def sig_hup_handler(self, signum, frame):
        if self.reload_in_progress is True:
            self.log.info("signal HUP received - reload already in progress")
            return

        self.reload_in_progress = True
        self.log.info("signal HUP received - reloading")

        try:
            self.callback.reload()
        except Exception as err:
            self.log.error("reloading failed:", err)

        self.reload_in_progress = False

    def sig_term_handler(self, signum, frame):
        self.log.info("signal TERM received - quitting")
        self.done = True

    def sig_int_handler(self, signum, frame):
        self.log.info("signal INT received - quitting")
        self.done = True

    def sig_hup_handler_child(self, signum, frame):
        self.log.info("signal HUP received - ignoring")

    def sig_term_handler_child(self, signum, frame):
        self.log.info("signal TERM received - quitting")
        sys.exit(0)

    def sig_int_handler_child(self, signum, frame):
        self.log.info("signal INT received - quitting")
        sys.exit(0)

    def init_parent_signal_handler(self):
        signal.signal(signal.SIGHUP, self.sig_hup_handler)
        signal.signal(signal.SIGINT, self.sig_int_handler)
        signal.signal(signal.SIGTERM, self.sig_term_handler)

    def init_child_signal_handler(self):
        signal.signal(signal.SIGHUP, self.sig_hup_handler_child)
        if self.exit_on_sig_term:
            signal.signal(signal.SIGINT, self.sig_int_handler_child)
            signal.signal(signal.SIGTERM, self.sig_term_handler_child)

    def run_worker(self, queue):
        self.init_child_signal_handler()
        cpu_friendly_sleep_counter = 0
        self.callback.init()
        while self.done is False:
            job = self.queue.recv_from_parent(queue)
            if not job:
                if cpu_friendly_sleep_counter < 1:
                    cpu_friendly_sleep_counter += 0.05
                time.sleep(cpu_friendly_sleep_counter)
                continue
            cpu_friendly_sleep_counter = 0
            self.queue.send_to_parent({
                "status": "finish",
                "job": self.callback.process(job),
                "pid": os.getpid()
            })
        self.log.info("good bye")


class Callback:
    def init(self):
        return

    def ready(self, num):
        return []

    def process(self, job):
        return job

    def finish(self, success, job):
        return

    def reload(self):
        return


class Transport:
    def __init__(self, maxsize=0):
        self.maxsize = maxsize
        self.parent_queue = self.create_queue()

    def create_queue(self):
        return Queue(maxsize=self.maxsize)

    def destroy_queue(self, queue):
        del queue

    def send_to_parent(self, data):
        return self.send(self.parent_queue, data)

    def send_to_worker(self, queue, data):
        return self.send(queue, data)

    def recv_from_parent(self, queue):
        return self.recv(queue)

    def recv_from_worker(self):
        return self.recv(self.parent_queue)

    def send(self, queue, data):
        # maxsize=0 so it should never happends
        # that the queue is full
        try:
            queue.put(data)
        except Full:
            return False

        return True

    def recv(self, queue):
        try:
            return queue.get(block=False)
        except Empty:
            pass
