import re
import sys
from bloonix.genattr import GenAttr

class Getopt:
    def __init__(self, title=None, desc=None, examples=None,
            add_help=True, auto_help=True, version=None):
        self.title = title
        self.desc = desc
        self.examples = examples
        self.version = version
        self.add_help = add_help
        self.auto_help = auto_help
        self.options_by_opt = {}
        self.options = []

    def add_option(self, *args, attr=None, default=None, desc=None,
            optional=None, multiple=None, value=None, regex=None, parse_int=None):
        if attr is None:
            for arg in args:
                if re.search("^--", arg):
                    attr = re.sub("^--", "", arg)
                    attr = re.sub("-", "_", attr)
                    break

        kwargs = {
            "args": args,
            "attr": attr,
            "default": default,
            "desc": desc,
            "value": value,
            "optional": optional,
            "multiple": multiple,
            "regex": regex,
            "parse_int": parse_int,
        }

        self.options.append(kwargs)

        for arg in args:
            self.options_by_opt[arg] = kwargs

    def parse_options(self, argv=None, genattr=None):
        args = {}

        if argv is None:
            argv = list(sys.argv)

        self.progname = re.sub(".+?/", "", argv.pop(0))

        while len(argv):
            arg = argv.pop(0)
            value = None

            if self.auto_help and (arg == "-h" or arg == "--help"):
                self.print_help()

            if self.version and (arg == "-v" or arg == "--version"):
                self.print_version()

            if arg not in self.options_by_opt:
                print("invalid option '%s'" % arg)
                sys.exit(1)

            option = self.options_by_opt[arg]
            attr = option["attr"]

            if option["value"] is None:
                args[attr] = True
                continue

            if len(argv) == 0:
                print("missing value for option '%s'" % arg)
                sys.exit(1)

            if option["multiple"] is True:
                if attr not in args:
                    args[attr] = []
                args[attr].append(argv.pop(0))
            else:
                args[attr] = argv.pop(0)

        for option in self.options:
            attr = option["attr"]
            if attr in args:
                value = args[attr]
                if option["regex"] and not re.search(option["regex"], args[attr]):
                    print("invalid value for option '%s'" % option["args"])
                    sys.exit(1)
                if option["parse_int"]:
                    if not re.search(r"^\d+\Z", value):
                        print("invalid value for option '%s'" % option["args"])
                        sys.exit(1)
                    args[attr] = int(value)
                continue

            if "default" in option and option["default"] is not None:
                args[attr] = option["default"]
                continue

            if option["optional"] is True:
                args[attr] = None
                continue

            print("missing mandatory option '%s'" % " or ".join(option["args"]))
            sys.exit(1)

        if genattr is True:
            return GenAttr(args)

        return args

    def print_help(self):
        if self.title is not None:
            print(self.title)
            print()

        print("\nUsage: %s [ OPTIONS ]\n" % self.progname)

        if self.desc is not None:
            print(self.desc)
            print()

        print("Options:\n")

        for opt in self.options:
            print(", ".join(opt["args"]), end="")

            if opt["value"]:
                print(" [%s]" % opt["value"])
            else:
                print()

            if opt["desc"]:
                print("    %s" % opt["desc"])

            if opt["default"] or opt["default"] is True:
                print("    Default: %s" % opt["default"])

            print()

        if self.add_help:
            print("-h, --help")
            print("    Print the help.")

        if self.examples is not None:
            print("\nExamples:\n")
            print(self.examples)
            print()

        print()
        sys.exit(0)

    def print_version(self):
        print("%s v%s" % (self.progname, self.version))
        sys.exit(0)
