import importlib
import os
import re
import sys
import time

from bloonix.json import json
from bloonix.config import parse_file
from bloonix.logger import logger
import bloonix.procutils as procutils
import bloonix.rx as rx
import bloonix.utils as utils
from bloonix.validator import validate

from bloonix.heaven.model import Model
from bloonix.heaven.view import View
from bloonix.heaven.controller import Controller
from bloonix.heaven.plugin import Plugin
from bloonix.heaven.routing import Routing
from bloonix.heaven.request import Request
from bloonix.heaven.response import Response
from bloonix.heaven.cache import Cache
from bloonix.heaven.stash import Stash
from bloonix.language import Language

logger("bloonix").set_pattern("%Y", "-") # user id

class Heaven:
    def init(self):
        self._load_config()
        self._load_env()
        self._load_logger()
        self._set_app_path()

        self.version = "n/a"
        self.time = int(time.time())
        self.routing = Routing(self)
        self.model = Model(self)
        self.view = View(self)
        self.controller = Controller(self)
        self.plugin = Plugin(self)

        # Short
        self.route = self.routing.route
        self.render = self.view.render

        # auto load lang if the path exist
        lang_path = "%s/%s" % (self.base_path, "lang")
        if os.path.isdir(lang_path):
            if "app-base-lang" in self._config:
                self.lang = Language(
                    path=lang_path,
                    default_lang=self._config["app-base-lang"])
            else:
                self.lang = Language(path=lang_path, default_lang="en")

        # Load app
        self._load_application()
        self._validate_routes()

    def run(self, env, start_response):
        """
        This method will be called in a loop by uwsgi for each http request.
        The runtime attributes are cleared before each run to ensure a fresh
        environment and at the end to free memory.
        Note: Reponse() is the only part which is not excepted.
        """
        self._reset_runtime();
        self.res = Response(start_response)

        try:
            self.time = int(time.time())
            self.req = Request(env)
            self.stash = Stash() # data exchange with the view
            self.cache = {} # temporary cache for all things
            self.view.reset()
            self._process_request()
        except Exception:
            self.log.exception()
            self.error()

        ret = self.res.process()
        self._reset_runtime();
        return ret

    def _reset_runtime(self):
        self.maintenance = None
        self.user = None
        self.session_id = None
        self.session = None
        self.time = None
        self.action = None
        self.process = None
        self.req = None
        self.res = None
        self.stash = None
        self.cache = None
        self.log.set_pattern("%Y", "-")

    def _load_config(self):
        if len(sys.argv) != 2:
            print("Usage: heaven.py [ config-file ]")
            sys.exit(0)

        self._config_file = sys.argv[1]
        self._config = parse_file(self._config_file)

        validate(self._config, {
            "app-name": {
                "type": str,
            },
            "user": {
                "type": str,
                "optional": True
            },
            "group": {
                "type": str,
                "optional": True
            },
            "logger": {
                "type": dict,
                "default": {},
                "allow_empty": True
            },
            "app": { # application wide configuration
                "type": dict,
                "default": {},
                "allow_empty": True
            },
            # mvc configuration
            "model": {
                "type": dict,
                "default": {},
                "allow_empty": True
            },
            "view": {
                "type": dict,
                "default": {},
                "allow_empty": True
            },
            "controller": {
                "type": dict,
                "default": {},
                "allow_empty": True
            },
            "plugin": {
                "type": dict,
                "default": {},
                "allow_empty": True
            },
        })

        self.app_name = self._config["app-name"]
        self.config = self._config["app"]

    def _load_env(self):
        if "group" in self._config:
            procutils.switch_group(self._config["group"])
        if "user" in self._config:
            procutils.switch_user(self._config["user"])

    def _load_logger(self):
        self.log = logger("bloonix")
        if "logger" in self._config:
            self.log.config(self._config["logger"])

    def _set_app_path(self):
        self.base_path = os.getcwd()
        self.log.debug("base_path", self.base_path)
        self.lib_path = "%s/%s" % (self.base_path, "lib")
        self.template_path = "%s/%s" % (self.base_path, "templates")
        self.log.debug("lib_path", self.lib_path)
        self.app_path = "%s/%s" % (self.lib_path, self.app_name)
        self.log.debug("app_path", self.app_path)
        sys.path.insert(0, self.lib_path)

    def _load_application(self):
        self.log.notice("load base application", self.app_name)
        importlib.import_module(self.app_name)
        #self.log.notice(self.routing._routes, dump=True)

    def _validate_routes(self):
        for module in self.routing.actions():
            self.log.notice("validate routes of module", module)
            for action in self.routing.actions(module):
                self.log.notice("validate route of module", module, "action", action)
                controller = self.controller.get(module)
                if (not hasattr(controller["instance"], action) or
                        not callable(getattr(controller["instance"], action))):
                    raise Exception("%s.%s() is not callable" % (module, action))

    def _process_request(self):
        self.log.notice("start processing request", self.req.path_info())
        module, action, opts, args = self.routing.parse(self.req.path_info())
        if args == None:
            self.controller.process(module, action, opts)
        else:
            self.controller.process(module, action, opts, args)
        self.view.process()

    def add_config(self, config):
        self._extra_config = config

    def error(self, msg="error"):
        self.res.status = "500 Internal Server Error"
        self.res.content_type = "text/html; charset=utf-8"
        self.render("plain")
        self.res.content = (
            "\n".join([
                '<!DOCTYPE html>',
                '<html lang="en">',
                '<head>',
                '    <title>Internal error</title>',
                '    <meta http-equiv="content-type" content="text/html; charset=UTF-8">',
                '    <style>',
                '    body { font-family: "arial"; }',
                '    </style>',
                '</head>',
                '<body style="background-color: #ffffff;">',
                '<div style="width: 500px; margin: 10% auto 0 auto;">',
                '<p style="margin: 0; padding: 20px 0 0 0;">Error: %s</p>' % msg,
                '<pre style="margin: 0; padding: 20px 0 0 0;">',
                '(en) Please come back later',
                '(fr) SVP veuillez revenir plus tard',
                '(es) Por favor, vuelva más tarde',
                '(de) Bitte versuchen sie es spaeter nocheinmal',
                "(at) Konnten's bitt'schoen spaeter nochmal reinschauen",
                '(no) Vennligst prov igjen senere',
                '(dk) Venligst prov igen senere',
                '(pl) Prosze sprobowac pozniej',
                '(pt) Por favor volte mais tarde',
                '(ru) Попробуйте еще раз позже',
                '(ua) Спробуйте ще раз пізніше',
                '(cn) 請稍後再來',
                '</pre>',
                '</div>',
                '</body>',
                '</html>'
            ])
        )
