from bloonix.logger import logger
from bloonix.findmod import load_path
import importlib
import inspect
import os
import re

"""
self.route_id_key is set to "id" in each controller if not exists.
This is the key in a route which is set to match the unique id of
the object in the route, as example /customer/:id

This attribute is also used in plugin.crud.
"""

class Controller:
    def __init__(self, heaven):
        self.log = logger("bloonix")
        self.controller = {}
        self.auto_path = {} # cache auto path
        self.root_controller = {}
        self.heaven = heaven
        self.root = "%s.controller.root" % self.heaven.app_name

    def autoload(self, path=None):
        paths = load_path("%s/%s" % (self.heaven.app_path, "controller"), deep=True)
        for p in paths:
            self.load(p)

    def load(self, *args):
        for arg in args:
            self._load(arg)

    def _load(self, name):
        name = re.sub(r"/", ".", name)
        module_name = "%s.controller.%s" % (self.heaven.app_name, name)
        self.log.notice("load controller", module_name)

        module = importlib.import_module(module_name)

        if hasattr(module, "Controller"):
            attr = getattr(module, "Controller")
            if name in self.heaven._config["controller"]:
                instance = attr(self.heaven._config["controller"][name])
            else:
                instance = attr()
        else:
            instance = module

        if not hasattr(instance, "route_id_key"):
            instance.route_id_key = "id"
        if hasattr(instance, "startup"):
            instance.startup()

        self.controller[module_name] = {"instance": instance, "shortname": name}

        # default and error only in root controller
        for action in ["auto", "default", "error", "begin", "end"]:
            if hasattr(instance, action):
                if (module_name == self.root
                        or action in ["auto", "begin", "end"]):
                    self.controller[module_name][action] = getattr(
                        instance, action)
            else:
                self.controller[module_name][action] = None

        if module_name == self.root:
            self.root_controller = self.controller[module_name]

    # Call a instance from another controller
    #   instance("hosts.list")
    def instance(self, name):
        module_name = "%s.controller.%s" % (self.heaven.app_name, name)
        return self.controller[module_name]["instance"]

    def get(self, controller):
        return self.controller[controller]

    def shortname(self, controller):
        return self.controller[controller]["shortname"]

    def has(self, controller):
        if controller in self.controller:
            return True
        return False

    def process(self, module, action, opts):
        if not module or not action:
            module = self.root
            action = "default"

        if self.root_controller["error"] is not None:
            try:
                self.__process(module, action, opts)
            except Exception:
                self.log.exception()
                self.root_controller["error"](opts)
        else:
            self.__process(module, action, opts)

    def __process(self, module, action, opts):
        self.heaven.process = "%s.%s" % (self.shortname(module), action)
        self.heaven.action = action
        self.log.notice("exec module", module, "action", action, "opts", opts)
        controller = self.controller[module]
        instance = controller["instance"]

        if not hasattr(instance, action):
            m = "action %s of controller %s does not exists" % (action, module)
            self.log.error(m)
            raise Exception(m)

        if self.process_auto_path(module, opts):
            if controller["begin"]:
                self.log.debug("call", module, "-> begin")
                controller["begin"](opts)
            self.log.debug("call", module, "->", action)
            action = getattr(instance, action)
            action(opts)
            if controller["end"]:
                self.log.debug("call", module, "-> end")
                controller["end"](opts)

            if self.root_controller["end"]:
                self.log.debug("call", self.root, "-> end")
                self.root_controller["end"](opts)

    def process_auto_path(self, module, opts):
        auto_path = self.get_auto_path(module)
        for m in reversed(auto_path):
            self.log.debug("call", m, "-> auto")
            controller = self.controller[m]
            auto = getattr(controller["instance"], "auto")
            if not auto(opts):
                self.log.debug("auto() of controller", m, "returned false")
                return False

        return True

    def get_auto_path(self, module):
        if module in self.auto_path:
            return self.auto_path[module]

        parts = module.split(".")
        self.auto_path[module] = auto_path = []

        while len(parts):
            m = ".".join(parts)
            if m in self.controller and self.controller[m]["auto"]:
                auto_path.append(m)
            parts.pop()

        if (self.root not in auto_path
                and self.root_controller
                and self.root_controller["auto"]):
            auto_path.append(self.root)

        return auto_path
