"""
This module simplified the handling of CRUD (create, read, update, delete) operations.

    Function:

    create()  ->  /users/create           create a new user
    show()    ->  /users/:id              show user data
    update()  ->  /users/:id/update       update user data
    delete()  ->  /users/:id/delete       delete user
    options() ->  /users/:id/options      form options for the user
              ->  /users/options          form options to create a new user

It is necessary to set the object to the table schema during startup().

    self.db = H.model.db.table_name

If :id exists in the route and the Auto class is used, then the value is
selected from the table and stored with the name of the table in the
heaven-cache:

    H.cache[self.db.table] = self.db.get(opts[self.route_id_key])

The attribute self.route_id_key is used as match key (:id) in the route.
You can override this key in __init__ or startup() in the controller.

If the List class is used, then the database query can be configured
using the list_opts attribute. The table name is set from self.db.table.
The default configuration is set as follows.

    self.list_opts = {
        "count": "id", 
        "columns": "*",
        "order": ["asc", "id"]
    }


"""

from bloonix.heaven import H
from bloonix.heaven.validator import Validator
from bloonix.exception import BloonixDuplicationError


class RaiseOnDupError:
    def raise_on_dup_error(self):
        if self.db.has_unique_column:
            return H.plugin.error.set("err-610", failed={
                self.db.has_unique_column["column"]: {
                    "status": self.db.has_unique_column["err"]
                }
            })
        return H.plugin.error.set("err-602")


class Auto:
    def auto(self, opts):
        if self.route_id_key in opts:
            value = self.db.get(opts[self.route_id_key])
            if not value:
                return H.plugin.error.set("err-600")
            H.cache[self.db.table] = value
        return True


class Validate:
    def validate(self, opts, ignore_missing=None, validator=None):
        """
        This class only works if bloonix.heaven.validator
        is used instead bloonix.validator.
        """
        if not H.plugin.token.check():
            return
        if validator is None:
            validator = self.db.validator
        if not isinstance(validator, Validator):
            raise Error("not a bloonix.heaven.validator object")
        res = validator.validate(ignore_missing=ignore_missing) 
        if res.has_failed:
            return H.plugin.error.set("err-610", failed=res.failed)
        return res.data

# If a table has just one unique column, then it's possible
# to intercept duplication errors with BloonixDuplicationError
# from bloonix.exception. For that the unique column must be
# set within the schema class like follows:
#   self.db.has_unique_column = {
#       column: "name",
#       err: "err-code"
#   }
class Create(Validate, RaiseOnDupError):
    def create(self, opts, validator=None, data_callback=None):
        data = self.validate(opts, validator=validator)
        if not data:
            return
        if data_callback != None:
            data_callback(opts, data)
        try:
            data = self.db.create_and_get(data)
        except BloonixDuplicationError:
            return self.raise_on_dup_error()
        except Exception:
            H.log.exception()
            return H.plugin.error.set("err-608")
        if hasattr(H.plugin, "audit"):
            H.plugin.audit.create(
                target=self.db.table,
                new_data=data)
        H.stash.data = data
        return True


class Update(Validate, RaiseOnDupError):
    def update(self, opts, validator=None, data_callback=None):
        data = self.validate(opts, ignore_missing=True, validator=validator)
        if not data:
            return
        if data_callback != None:
            data_callback(opts, data)
        try:
            self.db.update_by_id(opts[self.route_id_key], data)
        except BloonixDuplicationError:
            return self.raise_on_dup_error()
        except Exception:
            return H.plugin.error.set("err-607")
        if hasattr(H.plugin, "audit"):
            H.plugin.audit.update(
                target=self.db.table,
                old_data=H.cache.get(self.db.table),
                new_data=data)
        return True


class Delete:
    def delete(self, opts):
        if not H.plugin.token.check():
            return
        try:
            ret = self.db.delete_by_id(opts[self.route_id_key])
        except Exception:
            return H.plugin.error.set("err-614")
        if hasattr(H.plugin, "audit"):
            H.plugin.audit.delete(
                target=self.db.table,
                old_data=H.cache.get(self.db.table))
        return ret


class Show:
    def show(self, opts):
        if self.db.table in H.cache:
            H.stash.data = H.cache.get(self.db.table)
        else:
            H.stash.data = self.db.get(opts[self.route_id_key])


class List:
    def list(self, opts):
        """
        Default setup:
           def startup(self):
               self.list_opts = {
                   "count": "id",
                   "columns": "*",
                   "order": ["asc", "id"]
               }
        """
        ret = self.db.query(
            query=H.req.get_query_data(),
            count=self.list_opts.get("count", "id"),
            table=[
                self.list_opts.get("table", self.db.table),
                self.list_opts.get("columns", "*"),
            ],
            order=self.list_opts.get("order", ["asc", "id"])
        )
        if ret.success == False:
            return H.plugin.error.set("err-606")
        H.stash.data = ret.data
        H.stash.total = ret.total

    def add_list_route(self, route):
        module_name = self.__class__.__module__
        H.route(route, "list")


class Options:
    def options(self, opts, validator=None):
        if validator is None:
            validator = self.db.validator
        H.stash.data["options"] = validator.get_options()
        if self.route_id_key in opts:
            H.stash.data["values"] = H.cache.get(self.db.table)
        else:
            H.stash.data["values"] = validator.get_defaults()


class CRUD(Auto, Create, Update, Delete, Show, Options, List):
    def add_crud_routes(self, route):
        module_name = self.__class__.__module__
        H.route(route, "%s.list" % module_name)
        H.route(
            "%s/create" % route,
            "%s.create" % module_name)
        H.route(
            "%s/:%s/update" % (route, self.route_id_key),
            "%s.update"  % module_name)
        H.route(
            "%s/:%s/delete" % (route, self.route_id_key),
            "%s.delete" % module_name)
        H.route(
            "%s/:%s" % (route, self.route_id_key),
            "%s.show" % module_name)
        H.route(
            "%s/:%s/options" % (route, self.route_id_key),
            "%s.options" % module_name)
        H.route(
            "%s/options" % route,
            "%s.options" % module_name)


class Crud:
    pass
