import re
import urllib.parse

from bloonix.logger import logger
from bloonix.json import json

class Request:
    def __init__(self, env):
        self.ENV = env
        self.log = logger("bloonix")
        self.postdata = None
        self.jsondata = None
        self.maxlen = 0
        self.data = {}
        if re.search(r"application/json", self.content_type()):
            self.is_json = True
        else:
            self.is_json = False
        self._process_post_data()
        self._process_form_urlencoded(self.query_string())

    def cookie(self,
            name=None, value=None, path="/",
            expires=None, domain=None):
        if value is None:
            cookie = self.http_cookie()
            if cookie is not None:
                for pair in cookie.split(";"):
                    m = re.search(r"^\s*(.+)\s*=\s*(.+)\s*\Z", pair)
                    if m:
                        key = m.group(1)
                        value = m.group(2)
                        if key == name:
                            return value
        return None

    def _process_post_data(self):
        content_length = self.content_length()
        if content_length == 0:
            return
        if self.maxlen > 0 and content_length > self.maxlen:
            self.log.warning(
                "post data to long with %s bytes"
                % content_length)
            return
        self.postdata = self.ENV["wsgi.input"].read(content_length)
        if len(self.postdata) != content_length:
            self.postdata = None
            self.log.warning(
                "body length is less then content length:",
                len(body), "<", str(content_length))
        if self._decode_postdata():
            if re.search(
                    r"^application/x-www-form-urlencoded",
                    self.content_type()):
                self._process_form_urlencoded(self.postdata)
            elif self.is_json:
                try:
                    self.jsondata = json.loads(self.postdata)
                    if type(self.jsondata) is dict:
                        self.data = self.jsondata
                except Exception as e:
                    self.log.warning(
                        "unable to load json string from postdata:",
                        str(e))

    def _decode_postdata(self):
        try:
            self.postdata = self.postdata.decode("utf-8")
            return True
        except Exception as e:
            self.log.warning(
                "unable to decode postdata "
                "from bytes to utf-8 string: %s" % str(e))
            return False

    # Very useful if the application provides a query api.
    def get_query_data(self, offset=0, limit=50, max_limit=1000, unlimited=None):
        data = {}

        if unlimited == None:
            data["limit"] = self.data.get("limit", limit)
            data["offset"] = self.data.get("offset", offset)

            for key in ["offset", "limit"]:
                if (isinstance(data[key], str) and re.search(r"^-?\d+\Z", data[key])):
                    data[key] = int(data[key])

            if not isinstance(data["limit"], int) or data["limit"] < 0:
                data["limit"] = limit

            if data["limit"] > max_limit:
                data["limit"] = max_limit

            if not isinstance(data["offset"], int) or data["offset"] < 0:
                data["offset"] = offset

        if "query" in self.data and isinstance(self.data["query"], list):
            data["query"] = self.data["query"]

        if "sort" in self.data and isinstance(self.data["sort"], list):
            data["sort"] = self.data["sort"]

        return data

    def exist(self, param):
        if param in self.data:
            return True
        return False

    def params(self, params):
        data = {}
        for param in params:
            if param in self.data:
                data[param] = self.data[param]
        return data

    def param(self, param):
        return self.data.get(param, None)

    def _process_form_urlencoded(self, string):
        params = self.data
        pairs = re.split(r"[&;]", string)

        for pair in pairs:
            if "=" in pair:
                param, value = pair.split("=", 1)
            elif len(pair):
                param = pair
                value = ""
            else:
                continue

            param = urllib.parse.unquote_plus(param)
            value = urllib.parse.unquote_plus(value)
            self._add_param_value(param, value)

    def _add_param_value(self, param, value):
        data = self.data
        if param in data:
            if type(data[param]) is not list:
                data[param] = [data[param]]
            data[param].append(value)
        else:
            data[param] = value

    def path_info(self):
        path = self.env("PATH_INFO")
        if path is not None:
            return re.sub(r"/+\Z", "", path)
        return path

    def server_name(self):
        return self.env("SERVER_NAME")

    def script_name(self):
        return self.env("SCRIPT_NAME")

    def http_accept_encoding(self):
        return self.env("HTTP_ACCEPT_ENCODING")

    def server_admin(self):
        return self.env("SERVER_ADMIN")

    def http_connection(self):
        return self.env("HTTP_CONNECTION")

    def http_accept(self):
        return self.env("HTTP_ACCEPT")

    def request_method(self):
        return self.env("REQUEST_METHOD")

    def script_filename(self):
        return self.env("SCRIPT_FILENAME")

    def server_software(self):
        return self.env("SERVER_SOFTWARE")

    def http_accept_charset(self):
        return self.env("HTTP_ACCEPT_CHARSET")

    def http_cookie(self):
        return self.env("HTTP_COOKIE")

    def http_te(self):
        return self.env("HTTP_TE")

    def http_user_agent(self):
        return self.env("HTTP_USER_AGENT")

    def remote_addr(self):
        # X-Real-IP and X-Forwarded-For can be manipulated
        # by the client, so the real remote address must
        # be set in the webserver configuration.
        return self.env("REMOTE_ADDR")

    def remote_port(self):
        return self.env("REMOTE_PORT")

    def query_string(self):
        return self.env("QUERY_STRING")

    def server_signature(self):
        return self.env("SERVER_SIGNATURE")

    def server_port(self):
        return self.env("SERVER_PORT")

    def http_accept_language(self):
        return self.env("HTTP_ACCEPT_LANGUAGE")

    def server_protocol(self):
        return self.env("SERVER_PROTOCOL")

    def path(self):
        return self.env("PATH")

    def gateway_interface(self):
        return self.env("GATEWAY_INTERFACE")

    def request_uri(self):
        return self.env("REQUEST_URI")

    def server_addr(self):
        return self.env("SERVER_ADDR")

    def document_root(self):
        return self.env("DOCUMENT_ROOT")

    def http_host(self):
        return self.env("HTTP_HOST")

    def content_type(self):
        return self.env("CONTENT_TYPE")

    def env(self, var):
        return self.ENV.get(var, None)

    def envdump(self):
        data = {}
        for key in self.ENV:
            if isinstance(self.ENV[key], str):
                data[key] = self.ENV[key]
        return data

    def content_length(self):
        if "CONTENT_LENGTH" in self.ENV and len(self.ENV["CONTENT_LENGTH"]):
            return int(self.ENV["CONTENT_LENGTH"])
        return 0
