import certifi
import pycurl
import os
import re
import time
import base64
from bloonix.http.utils import (
    parse_url,
    unparse_url_quoted,
    quote_data,
    header_to_dict)

def request(*args, **kwargs):
    r = Curl()
    return r.request(*args, **kwargs)


class Curl:
    def __init__(self):
        self.c = pycurl.Curl()

    def _read_header(self, line):
        line = line.decode("iso-8859-1")
        if len(self._http_header) == 0:
            m = re.search(r"^HTTP/(1\.[01]|2)\s*(\d+)", line)
            if m:
                self._http_version = m.group(1)
                self._http_status = int(m.group(2))
                self._http_header += line
            return
        if ":" not in line:
            return
        self._http_header += line
        name, value = line.split(":", 1)
        name = name.strip().lower()
        value = value.strip()
        self._http_headers[name] = value
        if name == "content-type":
            self._content_type = value.lower()
            m = re.search(r"charset=(\S+)", self._content_type)
            if m:
                self._http_encoding = m.group(1)

    def _read_data(self, buf):
        if len(buf) + len(self._content) > self._max_read_len:
            self._max_read_len_exceeded = True
            return 0
        self._content += buf

    def _progress(self, *args):
        if time.time() > self._expire_time:
            # unfortunately the return value is ignored by cURL
            return 28 # CURLE_OPERATION_TIMEDOUT

    def _get_perf_data(self, c):
        time_total = c.getinfo(pycurl.TOTAL_TIME)
        time_namelookup = c.getinfo(pycurl.NAMELOOKUP_TIME)
        time_connect = (c.getinfo(pycurl.CONNECT_TIME)
            - c.getinfo(pycurl.NAMELOOKUP_TIME))
        if c.getinfo(pycurl.APPCONNECT_TIME) > 0:
            time_ssl_handshake = (c.getinfo(pycurl.APPCONNECT_TIME)
                - c.getinfo(pycurl.CONNECT_TIME))
        else:
             time_ssl_handshake = c.getinfo(pycurl.APPCONNECT_TIME)
        time_first_byte = (c.getinfo(pycurl.STARTTRANSFER_TIME)
            - c.getinfo(pycurl.PRETRANSFER_TIME))
        time_transfer = time_total - c.getinfo(pycurl.STARTTRANSFER_TIME)
        time_overhead = (
            time_total
            - time_transfer
            - time_first_byte
            - time_ssl_handshake
            - time_connect
            - time_namelookup)

        return {
            "time_total": float(format(time_total * 1000, ".3f")),
            "time_namelookup": float(format(time_namelookup * 1000, ".3f")),
            "time_connect": float(format(time_connect * 1000, ".3f")),
            "time_ssl_handshake": float(format(time_ssl_handshake * 1000, ".3f")),
            "time_first_byte": float(format(time_first_byte * 1000, ".3f")),
            "time_transfer": float(format(time_transfer * 1000, ".3f")),
            "time_overhead": float(format(time_overhead * 1000, ".3f"))}

    def request(
            self, url, ipaddr=None, data=None, encode_data=None,
            user_agent="-", auth_basic=None, header=None, insecure=None,
            ipv6=None, timeout=60, perf_vars=None, max_read_len=10485760,
            encoding="deflate,gzip", decode_fallback="utf-8", version=None):

        self._max_read_len = max_read_len
        self._max_read_len_exceeded = None
        self._http_version = None
        self._http_status = None
        self._http_header = ""
        self._http_headers = {}
        self._http_encoding = None
        self._content_type = None
        self._content = b""

        scheme, host, port, path, qs, frag = parse_url(url)

        # cURL does not quote parameter names in query_string or post data.
        url = unparse_url_quoted(scheme, host, port, path, qs)
        if data:
            if encode_data is True:
                # data as a dict: {"a": "1", "b": "2"}
                # data as a array: ["a", "1", "b", "2"]
                # data as a string: a=1\nb=2
                # Result: a=1&b=2
                if isinstance(data, str):
                    _data = data
                    data = []
                    for line in re.split(r"[\r\n]+", _data):
                        m = re.search(r"^([^\s]+)=(.*)\Z", line)
                        if m:
                            data.append(m.group(1))
                            data.append(m.group(2))
                if isinstance(data, (dict, list)):
                    data = quote_data(data)
                else:
                    raise Exception("unable to encode data - invalid data type")
            elif not isinstance(data, (str, bytes)):
                raise Exception(
                    "data must be a string or bytes - invalid data type '%s'"
                    % type(data))

        c = self.c
        c.setopt(c.NOSIGNAL, False)
        c.setopt(c.ENCODING, encoding)
        c.setopt(c.FOLLOWLOCATION, False)
        c.setopt(c.OPT_CERTINFO, True)
        c.setopt(c.URL, url)
        c.setopt(c.HEADERFUNCTION, self._read_header)
        c.setopt(c.WRITEFUNCTION, self._read_data)
        c.setopt(c.CAINFO, certifi.where())
        c.setopt(c.SSL_VERIFYPEER, 0 if insecure else 1)
        c.setopt(c.SSL_VERIFYHOST, 0 if insecure else 2)
        c.setopt(c.USERAGENT, user_agent or "-")
        # enable all supported built-in compressions
        c.setopt(c.ACCEPT_ENCODING, "")

        if version != None:
            if version == "1.0":
                c.setopt(pycurl.HTTP_VERSION, pycurl.CURL_HTTP_VERSION_1_0)
            elif version == "1.1":
                c.setopt(pycurl.HTTP_VERSION, pycurl.CURL_HTTP_VERSION_1_1)
            elif version == "2":
                c.setopt(pycurl.HTTP_VERSION, pycurl.CURL_HTTP_VERSION_2_0)
            elif version == "3":
                c.setopt(pycurl.HTTP_VERSION, pycurl.CURL_HTTP_VERSION_3_0)
            else:
                raise Exception("invalid http version")

        # Unfortunately, setting a global timeout in libcurl is not
        # possible and you can only set a timeout for the connection
        # and the transfer. For DNS problems the timeout in cURL is
        # fixed to 20 seconds. The solution is to use a progress
        # function which is called by cURL at least once per second.
        # This way you can simulate a global timeout and throw an
        # exception. 
        c.setopt(c.NOPROGRESS, 0) # required to use a progress function
        c.setopt(c.XFERINFOFUNCTION, self._progress) # called every second
        #c.setopt(c.CONNECTTIMEOUT, timeout)
        #c.setopt(c.TIMEOUT, timeout)

        #if method:
        #    c.setopt(c.CUSTOMREQUEST, method)

        if auth_basic:
            c.setopt(c.HTTPAUTH, pycurl.HTTPAUTH_BASIC)
            c.setopt(c.USERPWD, auth_basic)

        if ipv6:
            c.setopt(c.IPRESOLVE, c.IPRESOLVE_V6)

        if ipaddr:
            c.setopt(c.RESOLVE, ["%s:%s:%s" % (host, port, ipaddr)])

        header = header_to_dict(header)

        if data:
            if "content-type" not in header:
                header["content-type"] = "application/x-www-form-urlencoded"
            c.setopt(c.POSTFIELDS, data)

        if len(header):
            hdrs = ["%s: %s" % (k, v) for k, v in header.items()]
            c.setopt(c.HTTPHEADER, hdrs)

        self._expire_time = time.time() + timeout

        try:
            s = time.time()
            c.perform()
        except pycurl.error as err:
            if self._max_read_len_exceeded != True:
                if err.args[0] == 42:
                    return Response(has_failed=True, error="operation timed out")
                return Response(has_failed=True, error=str(err))
        except Exception as err:
            return Response(has_failed=True, error=err)

        decode_failed = None
        try:
            # If no charset is set, then we try at least
            # to decode the content to utf-8.
            # If this failes then the charset must be set
            # by the maintainer of the web application.
            if self._http_encoding:
                self._content = self._content.decode(self._http_encoding)
            elif decode_fallback:
                self._content = self._content.decode(decode_fallback)
        except Exception:
            self._content = str(base64.b64encode(self._content))
            decode_failed = True

        ret = Response(
            version=self._http_version,
            status=self._http_status,
            header=self._http_header,
            headers=self._http_headers,
            encoding=self._http_encoding,
            content=self._content,
            perf_data=self._get_perf_data(c),
            scheme=scheme,
            ipaddr=c.getinfo(pycurl.PRIMARY_IP),
            certinfo=c.getinfo(pycurl.INFO_CERTINFO),
            decode_failed=decode_failed,
            max_read_len_exceeded=self._max_read_len_exceeded)

        c.reset()
        return ret


class Response:
    def __init__(
            self, version=None, status=None, header=None, headers=None,
            encoding=None, content=None, has_failed=None, error=None,
            perf_data=None, scheme=None, ipaddr=None, certinfo=None,
            decode_failed=None, max_read_len_exceeded=None):
        self.version = version
        self.status = status
        self.header = header
        self.headers = headers
        self.encoding = encoding
        self.content = content
        self.has_failed = has_failed
        self.error = error
        self.perf_data = perf_data
        self.scheme = scheme
        self.ipaddr = ipaddr
        self.certinfo = certinfo
        self.decode_failed = decode_failed
        self.max_read_len_exceeded = max_read_len_exceeded
