import re

__url_base_re = re.compile(r"^(http|https)://([^?/]+)(.*)")
__host_ipv6_re = re.compile(r"^(\[.+?\]):(\d+)\Z")
__do_not_quote = (
    b"abcdefghijklmnopqrstuvwxyz"
    b"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    b"0123456789"
    b"-._~")

def quote(string, ignore=None):
    s = __do_not_quote
    if ignore:
        s += bytes(ignore, "utf-8")
    arr = []
    for b in bytes(str(string), "utf-8"):
        arr.append(chr(b) if b in s else "%{:02X}".format(b))
    return ''.join(arr)

def quote_url_path(string):
    s = quote(string, "/ ")
    return s.replace(" ", "+")

# This function would only works if the parameter and
# value doesn't contain the signs '=&;', for that reason
# the function shouldn't be used.
#def quote_string(string):
#    return quote(string, "=&;")

def quote_data(data):
    s = []
    if isinstance(data, dict):
        for k, v in data.items():
            if isinstance(v, list):
                for i in v:
                    s.append("%s=%s" % (quote(k), quote(i)))
            else:
                s.append("%s=%s" % (quote(k), quote(v)))
    elif isinstance(data, list):
        if len(data) % 2:
            raise Exception("invalid data array length")
        data = list(data) # dont maninpulate original
        while len(data):
            k = data.pop(0)
            v = data.pop(0)
            s.append("%s=%s" % (quote(k), quote(v)))
    return '&'.join(s)

def parse_url(url):
    port = None
    query_string = None
    fragment = None

    m = __url_base_re.search(url)
    if not m:
        raise Exception("unable to parse url")

    scheme = m.group(1)
    host = m.group(2)
    path = m.group(3)

    if ":" in host:
        m = __host_ipv6_re.search(host)
        if m:
            host = m.group(1)
            port = m.group(2)
        else:
            host, port = host.split(":", 1)
            port = port
        if not re.search(r"^\d+\Z", port):
            raise Exception("invalid port number: %s" % port)
        port = int(port)

    if port is None:
        if scheme == "http":
            port = 80
        else:
            port = 443

    if "#" in path:
        path, fragment = path.split("#", 1)

    if "?" in url:
        path, query_string = path.split("?", 1)

    path = path or "/"

    return scheme, host, port, path, query_string, fragment

def unparse_url_quoted(
        scheme, host, port=None, path=None,
        query_string=None, fragment=None):

    parts = [scheme, "://", host]
    if (port is not None and (
            (scheme == "http" and port != 80)
            or (scheme == "https" and port != 443))):
        parts.append(":%s" % port)

    if path is None or path == "/":
        parts.append("/")
    else:
        parts.append(quote_url_path(path))

    if query_string:
        parts.append("?%s" % query_string)
        #parts.append("?%s" % quote_string(query_string))

    if fragment:
        parts.append("#%s" % fragment)

    return ''.join(parts)

def header_to_dict(header):
    header = header or {}
    _header = {}
    if isinstance(header, list):
        for line in header:
            if ":" in line:
                k, v = line.split(":", 1)
                v = re.sub(r"^\s+", "", v)
                _header[k.lower()] = v
    elif isinstance(header, dict):
        for k, v in header.items():
            _header[k.lower()] = v
    return _header
