import os
import re
import sys
from bloonix.json import json
from bloonix.logger import logger

class Language:
    def __init__(self, path, default_lang="en"):
        self.path = re.sub(r"/\Z", "", path)
        self.default_lang = self.active_lang = default_lang
        self.log = logger("bloonix")
        self.text = {}
        self._parse_lang(self.default_lang)
        dirs = os.listdir(self.path)

        for d in dirs:
            if re.search("^[a-z]{2,3}$", d) and d != self.default_lang:
                self._parse_lang(d)

        self._check()
        self.set(self.default_lang)

    def _parse_lang(self, lang, base=False):
        if lang not in self.text:
            text = self.text[lang] = {}

        param = ""
        content = self._get_lang_content(lang=lang)

        def replace(m):
            return text[m.group(1)]

        for line in content:
            if re.search(r"^\s*$", line) or re.search(r"^#", line):
                continue

            line = re.sub(r"@@(var\.[^\s]+)@@", replace, line)

            match = re.search(r"^([^\s]+?):\s+(.*)$", line)
            if match:
                param = match.group(1)
                value = match.group(2)
                text[param] = value
                continue

            match = re.search(r"^\s+([^\s].*)$", line)
            if match:
                text[param] += " "
                text[param] += match.group(1)

    def _get_lang_content(self, lang):
        content = []
        default_lang_file = None
        lang_file = "%s/%s" % (self.path, lang)

        default_lang_files = [
            "%s/%s.base" % (self.path, lang),
            "/usr/lib/bloonix/etc/heaven/lang/%s.base" % lang,
            "/usr/local/lib/bloonix/etc/heaven/lang/%s.base" % lang,
        ]

        for f in default_lang_files:
            if os.path.isfile(f):
                default_lang_file = f
                break

        if default_lang_file is not None:
            f = open(default_lang_file, "r", encoding="utf-8")
            content += f.readlines()
            f.close()

        if os.path.isfile(lang_file):
            f = open(lang_file, "r", encoding="utf-8")
            content += f.readlines()
            f.close()

        return content

    def _check(self):
        default_lang = self.text[self.default_lang]

        for lang in self.text:
            if lang == self.default_lang:
                continue

            for key in self.text[lang]:
                if key not in default_lang:
                    print("ERR(%s): unknown key '%s'" % (lang, key))
                    sys.exit(1)

        for key in default_lang:
            for lang in self.text:
                if lang == self.default_lang:
                    continue

                if key not in self.text[lang]:
                    print("ERR(%s): missing key '%s'" % (lang, key))
                    sys.exit(1)

    def set(self, lang):
        self.active = self.text[lang]
        self.active_lang = lang

    def get(self, key, *args):
        value = self.active[key]

        if len(args):
            value = value % tuple(args)

        return value

    def get_text(self):
        return self.text
