import inspect
import os
import re
import socket
import stat
import sys
import time
import traceback

import json
#from bloonix.json import json  ## json.dumps({}, default=func) does not working???
from bloonix.logger.callable import CallableOutput
from bloonix.logger.file import FileOutput
from bloonix.logger.list import ListOutput
from bloonix.logger.mail import MailOutput
from bloonix.logger.screen import ScreenOutput
from bloonix.validator import Validator

import platform
if platform.system() != "Windows":
    import grp
    import pwd

class Logger:
    def __init__(self, config=None):
        self._progname = sys.argv[0]
        self._short_progname = re.sub(r".*[\\//](.+)", r"\1", self._progname)
        self._time = time.time()
        self._start_time = time.time()
        self._outputs_by_level = {}
        self._level_to_num = {
            "DEBUG": 7,
            "INFO": 6,
            "NOTICE": 5,
            "WARNING": 4,
            "ERROR": 3,
            "CRITICAL": 2,
            "ALERT": 1,
            "EMERGENCY": 0
        }
        # Patterns V-Z are reserved for the application.
        self._patterns = {
            "%L": self._p_level,
            "%T": self._p_time,
            "%P": self._p_pid,
            "%H": self._p_hostname,
            "%U": self._p_username,
            "%G": self._p_groupname,
            "%N": self._p_newline,
            "%C": self._p_caller,
            "%S": self._p_progname,
            "%s": self._p_short_progname,
            "%r": self._p_runtime,
            "%t": self._p_measurement,
            "%m": self._p_message,
            "%%": self._p_percent
        }
        self._logger_opts = [
            "type", "message_layout", "level", "timeformat",
            "die_on_errors", "prepare_multiline", "filter",
        ]
        if config:
            self.config(config)

    def add(self, **kwargs):
        """
            log.add(type="file", filename="test1.log", level="info")
            log.add(type="file", filename="test2.log", level="notice")
        """
        self.validate(kwargs)
        self._process_patterns(kwargs)
        self._register_output(kwargs)

    def config(self, config):
        """
        To create a simple screen logger for debug purposes
        just pass the log level to config():

            log.config("DEBUG")

        Possible formats:

            log.config([
                {"type":"file","filename":"test1.log"},
                {"type":"file","filename":"test2.log"},
            ])

            log.config({
                "file": {
                    "filename":"test1.log"
                }
            })

            log.config({
                "file": [
                    {"filename":"test1.log"},
                    {"filename":"test2.log"},
                ]
            })
        """
        if isinstance(config, str):
            config = {
                "screen": {
                    "level": config,
                    "timeformat": "%H:%M:%S",
                    "message_layout": "%T %L %P %m"
                }
            }

        if type(config) == list:
            for conf in config:
                self.add(**conf)
        else:
            for output_type in config:
                opts = config[output_type]
                if type(opts) == list:
                    for conf in opts:
                        conf["type"] = output_type
                        self.add(**conf)
                else:
                    opts["type"] = output_type
                    self.add(**opts)

    def validate(self, logger_opts):
        output_opts = self._split_options(logger_opts)
        v = self.validator()
        v.validate(logger_opts)
        level = logger_opts["level"].upper()
        logger_opts["level"] = self._level_to_num[level]

        if logger_opts["type"] == "file":
            logger_opts["log_obj"] = FileOutput(output_opts)
        elif logger_opts["type"] == "callable":
            logger_opts["log_obj"] = CallableOutput(output_opts)
        elif logger_opts["type"] == "list":
            logger_opts["log_obj"] = ListOutput(output_opts)
        elif logger_opts["type"] == "screen":
            logger_opts["log_obj"] = ScreenOutput(output_opts)
        elif logger_opts["type"] == "mail":
            logger_opts["log_obj"] = MailOutput(output_opts)

    def validator(self):
        v = Validator()
        v.set({
            "type": {
                "type": str,
                "regex": r"^(callable|file|list|screen|mail)\Z",
                "default": "screen"
            },
            "message_layout": {
                "type": str,
                "default": "%T %L %P %t %m"
            },
            "filter": {
                "type": str,
                "optional": True,
            },
            "level": {
                "type": str,
                "regex": r"^(critical|error|warning|notice|info|debug)\Z",
                "default": "info"
            },
            "timeformat": {
                "type": str,
                "default": "%b %d %Y %H:%M:%S"
            },
            "die_on_errors": {
                "type": bool,
                "default": True
            },
            "prepare_multiline": {
                "type": bool,
                "default": True
            }
        })
        return v

    def _split_options(self, logger_opts):
        output_opts = {}
        split_opts = []
        for opt in logger_opts:
            if opt not in self._logger_opts:
                split_opts.append(opt)
        for opt in split_opts:
            output_opts[opt] = logger_opts.pop(opt)
        return output_opts

    def _register_output(self, opts):
        for level in self._level_to_num:
            if self._level_to_num[level] <= opts["level"]:
                if level not in self._outputs_by_level:
                    self._outputs_by_level[level] = []
                self._outputs_by_level[level].append(opts)

    def _process_patterns(self, opts):
        opts["patterns"] = []

        regex = re.compile("(%[" + "".join(self._patterns.keys()) + "])")

        def replace_pattern(m):
            opts["patterns"].append(m.group(0))
            return "%s"

        opts["message_layout"] = re.sub(regex, replace_pattern, opts["message_layout"])

    def set_pattern(self, pattern, func):
        """
        Set your own pattern:

            import time
            def get_time():
                return time.time()
            log.set_pattern("%Y", get_time)

        Or just a static string:

            log.set_pattern("%X" "string")
        """
        self._patterns[pattern] = func

    def is_debug(self):
        return self.is_log(level="DEBUG")

    def is_info(self):
        return self.is_log(level="INFO")

    def is_notice(self):
        return self.is_log(level="NOTICE")

    def is_warning(self):
        return self.is_log(level="WARNING")

    def is_error(self):
        return self.is_log(level="ERROR")

    def is_critical(self):
        return self.is_log(level="CRITICAL")

    def is_alert(self):
        return self.is_log(level="ALERT")

    def is_emergency(self):
        return self.is_log(level="EMERGENCY")

    def is_log(self, level):
        if level in self._outputs_by_level:
            return True
        return False

    def exception(self, level="ERROR", message="Exception occurred",
            dump=False, add_trace=False):
        return self.log(
            "%s:\n%s" % (message, traceback.format_exc()),
            level=level,
            dump=False,
            add_trace=False)

    def debug(self, *message, dump=False, add_trace=False):
        return self.log(*message, level="DEBUG", dump=dump, add_trace=add_trace)

    def info(self, *message, dump=False, add_trace=False):
        return self.log(*message, level="INFO", dump=dump, add_trace=add_trace)

    def notice(self, *message, dump=False, add_trace=False):
        return self.log(*message, level="NOTICE", dump=dump, add_trace=add_trace)

    def warning(self, *message, dump=False, add_trace=False):
        return self.log(*message, level="WARNING", dump=dump, add_trace=add_trace)

    def error(self, *message, dump=False, add_trace=False):
        return self.log(*message, level="ERROR", dump=dump, add_trace=add_trace)

    def critical(self, *message, dump=False, add_trace=False):
        return self.log(*message, level="CRITICAL", dump=dump, add_trace=add_trace)

    def alert(self, *message, dump=False, add_trace=False):
        return self.log(*message, level="ALERT", dump=dump, add_trace=add_trace)

    def emergency(self, *message, dump=False, add_trace=False):
        return self.log(*message, level="EMERGENCY", dump=dump, add_trace=add_trace)

    #def trace(self, *message, level="DEBUG"):
    #    return self.log(*message, level=level, add_trace=True)

    #def dump(self, *message, level="DEBUG"):
    #    return self.log(*message, level=level, dump=True)

    def log(self, *args, level="DEBUG", add_trace=False, dump=False):
        level = level.upper()
        ret = True

        if level not in self._outputs_by_level:
            return True

        for output in self._outputs_by_level[level]:
            message = None

            if dump is True:
                args = [json.dumps(args, indent=2, default=self._convert)]

            message = self._create_message(output, level, args)

            if add_trace is True:
                message += self._add_trace()

            if output["prepare_multiline"] is True:
                message = re.sub(r"(\r\n|\n)", r"\1  ", message)

            if output["filter"] and not re.search(output["filter"], message):
                continue

            try:
                method = output["log_obj"]
                method.log(message, level)
            except:
                if output["die_on_errors"] is True:
                    raise
                ret = False

        return ret

    def _convert(self, o):
        return str(o)

    def _add_trace(self):
        stacktrace = [""]
        curframe = inspect.currentframe()
        i = 0

        for f in inspect.getouterframes(curframe):
            (frame, filename, line, fn, lines, index) = f
            stacktrace.append(
                "  CALL(%s): filename(%s) function(%s) line(%s)" % (i, filename, fn, line)
            )
            i += 1

        return "\n".join(stacktrace)

    def _create_message(self, output, level, message):
        message = " ".join(str(v) for v in message)
        patterns = self._generate_message_patterns(output, level, message)
        message = output["message_layout"] % patterns
        return message.rstrip()

    def _generate_message_patterns(self, output, level, message):
        m = []
        for p in output["patterns"]:
            func = self._patterns[p]
            if callable(func):
                m.append(func(output, level, message))
            else:
                m.append(str(func))
        return tuple(m)

    def _p_time(self, output, level, message):
        return time.strftime(output["timeformat"])

    def _p_level(self, output, level, message):
        return level

    def _p_message(self, output, level, message):
        return message

    def _p_pid(self, *args):
        return str(os.getpid())

    def _p_hostname(self, *args):
        return socket.gethostname()

    def _p_runtime(self, *args):
        runtime = time.time() - self._start_time
        return format(runtime, ".3f")

    def _p_measurement(self, *args):
        curtime = time.time()
        logtime = curtime - self._time
        self._time = curtime
        return format(logtime, ".6f")

    def _p_caller(self, *args):
        curframe = inspect.currentframe()
        (frame, filename, line, fn, lines, index) = inspect.getouterframes(curframe)[5]
        return "%s, line %s" % (filename, line)

    def _p_newline(self, *args):
        return "\n"

    def _p_username(self, *args):
        return pwd.getpwuid(os.geteuid()).pw_name

    def _p_groupname(self, *args):
        return grp.getgrgid(os.getegid()).gr_name

    def _p_progname(self, *args):
        return self._progname

    def _p_short_progname(self, *args):
        return self._short_progname

    def _p_percent(self, *args):
        return "%"
