import os
import platform
import re
import random
from bloonix.runcmd import RunCmd
from bloonix.logger import logger
from bloonix.json import json
from bloonix.socket.serialized import Client

log = logger("bloonix")

def execute_check(
        command, plugin_path, command_options=None, plugin_cache_dir=None,
        timeout=30, host_id=0, service_id=0, interpreter=None):

    command = re.sub(r"\n", "", command)
    command = re.sub(r"^\s+", "", command)
    command = re.sub(r"\s+\Z", "", command)
    command_regex = (
        r"^"
            r"(sudo-){0,1}check-[a-zA-Z_0-9]+"  # the command name like check-cpu
            r"(?:-[a-zA-Z_0-9]+)*"              # maybe a long name like check-linux-cpu
            r"(?:\.[a-zA-Z]+){0,1}"             # maybe a file extension
        r"\Z")

    if not re.search(command_regex, command):
        return {
            "status": "UNKNOWN",
            "message": "invalid command name %s" % command}

    if re.search(r"^sudo-check-", command):
        sudo_required = True
    else:
        sudo_required = None

    if isinstance(plugin_path, str):
        plugin_path = re.split(r"\s*,\s*", plugin_path)

    command_full_path = None
    for path in plugin_path:
        if os.path.isfile("%s/%s" % (path, command)):
            command_full_path = "%s/%s" % (path, command)
            break
    if command_full_path is None:
        return {
            "status": "UNKNOWN",
            "message": "plugin %s not found: %s" % (command, ",".join(plugin_path))}
    if timeout < 30:
        timeout = 30

    arguments = [
        "--internal-host-id",
        str(host_id),
        "--internal-service-id",
        str(service_id)
    ]

    if plugin_cache_dir is not None:
        arguments.append("--internal-plugin-cache-dir")
        arguments.append(plugin_cache_dir)

    if isinstance(command_options, dict):
        to_stdin = json.dumps(command_options)
        arguments.append("--stdin")
    else:
        to_stdin = None

    if sudo_required is True:
        arguments.insert(0, command_full_path)
        command_full_path = "/usr/bin/sudo"

    run_command = [command_full_path] + arguments
    if interpreter is not None:
        run_command.insert(0, interpreter)

    log.info("execute %s" % " ".join(run_command))
    log.info("timeout", timeout, "to_stdin", to_stdin)

    ret = RunCmd(
        command=run_command,
        to_stdin=to_stdin,
        timeout=timeout,
        kill_signal=9,
        exit_success_codes=[0,1,2,3,4,5,6]
    )

    if ret.stderr:
        log.warning("stderr %s: %s" % (command_full_path, ret.stderr))

    if ret.has_failed is True:
        if ret.has_timed_out:
            return {
                "status": "CRITICAL",
                "message": "check runs on a timeout after %ss" % timeout}

        return {"status": "UNKNOWN", "message": ret.error}

    try:
        log.info("output:", ret.stdout)
        return json.loads(ret.stdout)
    except Exception as err:
        log.warning(">>>", ret.stdout)
        return {
            "status": "UNKNOWN",
            "message": "unable to de-serialize json data: %s" % err}


"""  Example input:
    {
      "command": "check-http",
      "command_options": {
        "url": "https://www.bloonix.de/",
        "http_status_code": "200",
        "timeout": "15",
        "use_ipv6": false,
        "data_urlencode": [],
        "user_agent": "Bloonix Plugin Check-HTTP",
        "insecure": false,
        "certificate": "21,14",
        "http_header": [],
        "debug_header": false,
        "notice": [],
        "warning": [],
        "alert": [],
        "critical": [],
        "match_notice": [],
        "match_warning": [],
        "match_alert": [],
        "match_critical": []
      },
      "satellite_type": 1,
      "satellites": [11,12,13],
      "last_satellite": 11,
      "satellites": {
        11: {
          "id": 11,
          "hostname": "localhost",
          "port": 5464,
          "authkey": "bloonixbloonixbloonix..."
        },
        13: {
          "id": 13,
          "hostname": "localhost",
          "port": 5466,
          "authkey": "bloonixbloonixbloonix..."
        },
        12: {
          "id": 12,
          "hostname": "localhost",
          "port": 5465,
          "authkey": "bloonixbloonixbloonix..."
        }
      }
    }
"""
def execute_satellite_check(
        command, command_options, timeout,
        last_satellite, satellites, satellites_by_id,
        host_id=None, service_id=None):

    if last_satellite not in satellites:
        last_satellite = satellites[ random.randrange(len(satellites)) ]

    while True:
        _id = satellites.pop(0)
        satellites.append(_id)
        if _id == last_satellite:
            break

    i = 0
    client = None
    for _id in satellites:
        if _id not in satellites_by_id:
            continue
        satellite = satellites_by_id[_id]
        if i == 3:
            break
        i += 1
        c = Client(
            host=satellite["hostname"],
            port=satellite["port"],
            use_ssl=True,
            ssl_verify_mode="none",
            compression="zlib")
        try:
            c.connect(timeout=10)
            res = c.recv(timeout=10)
            if "status" in res and res["status"] != "ready":
                continue
            client = c
            break
        except Exception:
            log.exception(level="WARNING")
            continue

    if not client:
        return {
            "status": "UNKNOWN",
            "message": "unable to connect to any satellite"}

    res = None
    try:
        c.send({
            # the hostname of the satellite is passed
            # to make it possible to proxy the request
            "hostname": satellite["hostname"],
            "authkey": satellite["authkey"],
            "host_id": host_id,
            "service_id": service_id,
            "action": "exec",
            "data": {
                "command": command,
                "command_options": command_options,
                "timeout": timeout,
            }})
        # the check itself has 60s time to process,
        # we allow an overhead of 10s = 70s
        res = c.recv(timeout=70)
        """ response:
        {
            "status": ok | err
            "message": message is set if status is err
            "data": data is set if status is ok
        }
        """
    except Exception as err:
        return {
            "status": "UNKNOWN",
            "message": "satellite %s %s runs on an error: %s" % (
                satellite["id"], satellite["hostname"], err
            )
        }

    if (not isinstance(res, dict)
            or "status" not in res
            or not isinstance(res["status"], str)
            or res["status"] not in ["ok", "err"]):
        return {
            "status": "UNKNOWN",
            "message": "invalid data received from satellite %s %s" % (
                satellite["id"], satellite["hostname"]
            )
        }

    if res["status"] != "ok":
        return {
            "status": "UNKNOWN",
            "message": "satellite %s %s response status '%s' message: %s" % (
                satellite["id"],
                satellite["hostname"],
                res["status"],
                res.get("message", "-")
            )
        }

    if "data" not in res or not isinstance(res["data"], dict):
        return {
            "status": "UNKNOWN",
            "message": "no data received from satellite %s %s" % (
                satellite["id"], satellite["hostname"]
            )
        }

    ret = res["data"]
    ret["satellite_id"] = satellite["id"]
    ret["message"] = "%s (Satellite %s, %s)" % (
        ret["message"], satellite["name"], satellite["hostname"])
    return ret
