import os
import platform
import re
import socket
import sys
import time as _time
if platform.system() != "Windows":
    import fcntl
from bloonix.runcmd import RunCmd
from bloonix.validator import validate
from bloonix.json import json
from bloonix import utils
from bloonix import mstatus
import bloonix.rx

"""
TODO
    add_mtr

"""

class Check:
    def __init__(
            self, version, service, category, abstract,
            description, frequency="default",
            default_location="localhost",
            network_service=0, argv=None, info=None):
        """
        Plugin(
            version="0.1",
            service="Service",
            category="Webserver,Network,Apache,...",
            abstract="Short plugin info",
            description="Long plugin info.",
            network_service=0|1|2,
            default_location="localhost|remote")
        """
        # argv can be passed for debug purposes
        if argv is None:
            self.argv = list(sys.argv)
            self.argv.pop(0)
        else:
            self.argv = argv
        self.version = version
        self.service = service
        self.category = category
        self.abstract = abstract
        self.description = description
        # frequency interval.
        # default: 30, 60 ...
        # low: wtrm, check-updates and so on
        self.frequency = frequency
        # 0 = no, localhost only
        # 1 = yes, satellites, bloonix server or localhost
        # 2 = yes, bloonix server or localhost
        # 3 = yes, bloonix server only
        if network_service not in [0,1,2,3]:
            raise Exception("invalid network_service")
        self.network_service = network_service
        self.default_location = default_location
        self.info = None
        self.charts = []
        self.progname = sys.argv[0]
        self.shortname = re.sub(r".*[\\//](.+)", r"\1", self.progname)
        self.config_path = "/etc/bloonix/agent"
        self.debug = False
        self.flags = ""
        self.examples = []
        self.option = {}  # options from sys.argv or json-input
        self.option_rules = {} # options added via add_option()
        self.options_order = []
        self.options_no_mixin = []
        self.metrics = {}
        self.strings = {}
        self._has_json_tree = None
        self.item_id = []
        self.exit_status = "OK"
        self.exit_message = None
        self._pass_options_is_now_forbidden = False
        # Default IDs to store statics in json files.
        self.internal_host_id = 0
        self.internal_service_id = 0
        # _get_mtr_by_addr is used to save the hostname
        # to get an mtr before exit() is called
        self._get_mtr_by_addr = False
        self._get_mtr_timeout = None
        # For measuring code executions
        self._runtime = _time.time()

    def no_mixin(self, *opts):
        self.options_no_mixin.append(list(opts))

    def add_example(self, message):
        self.examples.append(message)

    def _set_plugin_cache_dir(self):
        if os.path.isdir(self.option["internal-plugin-cache-dir"]):
            self.plugin_cache_dir = self.option["internal-plugin-cache-dir"]
        else:
            self.plugin_cache_dir = "/tmp"

    def runcmd(self, **kwargs):
        if "timeout" not in kwargs:
            kwargs["timeout"] = 10
        ret = RunCmd(**kwargs)
        if ret.has_failed:
            if ret.has_timed_out:
                self.exit(
                    status="CRITICAL",
                    message=(
                        "command '%s' timed out after %s seconds" %
                            (kwargs["command"][0], kwargs["timeout"])
                    ))
            self.exit(
                status="UNKNOWN",
                message=ret.error)
        return ret

    def libfile(self):
        # this file is used to store statistics for
        # which deltas must be calculated
        return self.basefile() + ".json"

    def cachefile(self):
        # this file is used to store any cache data
        return self.basefile() + ".cache"

    def basefile(self):
        """
        Returns the path + name to store caching data for the plugin.
        As example if the plugin name is check-http then
        /tmp/blx-check-http-0-0 is returned.
        """
        basefile = "-".join([
            "blx",
            self.shortname,
            str(self.option["internal-host-id"]),
            str(self.option["internal-service-id"])
        ])
        basefile = "/".join([
            self.plugin_cache_dir,
            basefile
        ])
        return basefile

    def add_option(self, default=None, **kwargs):
        """
        name
            The command option, but human readable, eg.
           "The hostname"
        option
            The command option, eg. "host" for "--host"
        description
            A description
        optional
            True | False
            Default: True
        multiple
            True | False
            Default: False
        default
            A default value to use if no value is passed.
            Optional: True
        datatype
            This parameter is necessary if the option expects
            a value. Valid settings are:
                bool                - just a flag
                int                 - no floating point number
                float               - floating point number
                str                 - string is displayed as a input
                                      field in the webgui
                lstr                - string (long) is displayed
                                      as a textarea in the webgui
                json-kv             - json-kv is a simple key-value list,
                                      in the webgui displayed as a table
                                      with 2 columns on each row for.
                                      The main element starts a s dict.
                json-raw            - json-raw is displayed as a textarea
                                      in the webgui. The main element 
                                      starts as a dict.
                string_condition    - [key,op,value] condition for strings
                metric_condition    - [key,op,value] condition for numbers
                json_condition      - [str,op,value] condition to parse json values
        hint
            The type or a very, very short description of the value.
            Optional: True
        regex
            A regular expression to validate the value.
            Optional: True
        example
            An example how to use this option. This can be shown as
            placeholder in a input field or textarea of a webform.
            Optional: True
        command_line_only
            0 | 1 | 2
            Set to 1 or 2 if the option should not be exported by
            --plugin-info. Set to 2 if the option shouldn't be
            listed in the help.
            Default: False
        """
        if self._pass_options_is_now_forbidden is True:
            self.exit(
                status="UNKNOWN",
                message=(
                    "after calling parse_options() it's not "
                    "allowed to pass new options (%s)"))

        validate(kwargs, {
            "name": {
                "type": str
            },
            "option": {
                "type": str,
                "regex": r"^[a-z]+(?:-[a-z0-9]+)*\Z"
            },
            "description": {
                "type": str
            },
            "optional": {
                "type": bool,
                "default": True
            },
            "multiple": {
                "type": bool,
                "default": False
            },
            "datatype": {
                "type": str,
                "regex": r"^(?:bool|int|float|str|lstr|json-kv|json-raw|wtrm|(?:string|metric|json)_condition)\Z"
            },
            "hint": {
                "type": str,
                "optional": True
            },
            "command_line_only": {
                "type": int,
                "default": 0
            },
            "regex": {
                "type": str,
                "optional": True
            },
            "minvalue": {
                "type": (int, float),
                "optional": True
            },
            "maxvalue": {
                "type": (int, float),
                "optional": True
            },
            "example": {
                "type": str,
                "optional": True
            },
            "unit": {
                "type": str,
                "regex": r"^(?:bytes|percent|ms|temperature)\Z",
                "default": None
            }
        })
        kwargs["default"] = default
        if ((kwargs["minvalue"] is not None or kwargs["maxvalue"] is not None)
                and kwargs["datatype"] not in ["int", "float"]):
            raise Exception(
                "datatype must be set to int or float "
                "in combination with minvalue or maxvalue")
        if kwargs["datatype"] == "bool" and kwargs["optional"] is False:
            raise Exception(
                "optional=False in combination "
                "with datatype=bool makes no sense")
        if (kwargs["multiple"] is True
                and kwargs["datatype"] not in [
                    "int", "float", "str", "lstr", "json-raw",
                    "json_condition",
                    "string_condition",
                    "metric_condition"]):
            raise Exception(
                "option 'multiple' in only valid in combination "
                "with datatype int, float and string")
        if kwargs["option"] in self.option_rules:
            self.exit(
                status="UNKNOWN",
                message=(
                    "Internal plugin error: option '%s' "
                    "already added" % kwargs["option"]
                ))
        self.option_rules[kwargs["option"]] = kwargs
        self.options_order.append(kwargs["option"])

    def parse_options(self):
        """
        Parse the command line options and the options from STDIN.
        The options are returned as dict.
        """

        self._add_default_options()
        if len(self.metrics):
            self._add_metric_options()
        if len(self.strings):
            self._add_string_matches()
        self._parse_command_line_arguments(self.argv)

        # help, version, plugin_info and stdin are command line options
        # only and will not be processed if they are passed over stdin.
        # So they existance has to be checked before _parse_stdin_arguments
        # and before _validate_options.
        if "help" in self.option:
            self._print_help()
        if "version" in self.option:
            self._print_version()
        if "plugin-info" in self.option:
            self._print_plugin_info()
        if "plugin-dstat" in self.option:
            self._print_plugin_dstat()
        if "stdin" in self.option:
            self._parse_stdin_arguments()
        if "input-file" in self.option:
            self._parse_input_file_arguments()
        self._validate_options()
        self._set_plugin_cache_dir()
        self._pass_options_is_now_forbidden = True
        return self.option

    def _parse_command_line_arguments(self, argv):
        while len(argv):
            arg = argv.pop(0)
            # Normally only options with long names exist,
            # in this case we make an exception
            if arg == "-h":
                arg = "--help"
            if not re.search(r"^--[a-zA-Z0-9][a-zA-Z0-9\-]*\Z", arg):
                self.exit(
                    status="UNKNOWN",
                    message="Invalid option: %s" % arg
                )
            arg = re.sub(r"^--", "", arg)
            value = None
            if arg not in self.option_rules:
                self.exit(
                    status="UNKNOWN",
                    message="Unknown option: %s" % arg)
            if self.option_rules[arg]["datatype"] == "bool":
                value = True
            else:
                if len(argv) == 0:
                    self.exit(
                        status="UNKNOWN",
                        message="missing value for option '%s'" % arg)
                value = argv.pop(0)
            if (self.option_rules[arg]["datatype"] in [
                    "string_condition",
                    "metric_condition",
                    "json_condition",
                    "json-kv"]):
                value = json.loads(value)
            if self.option_rules[arg]["multiple"] is True:
                if arg not in self.option:
                    self.option[arg] = []
                self.option[arg].append(value)
            elif arg not in self.option:
                self.option[arg] = value
            else:
                self.exit(
                    status="UNKNOWN",
                    message="passing option '%s' multiple times" % arg)

    def _parse_stdin_arguments(self):
        data = sys.stdin.read()
        if not data or re.search(r"^[\s\r\n]+\Z", data):
            return
        self._add_data_to_options(data)

    def _parse_input_file_arguments(self):
        f = open(self.option["input-file"], "r")
        data = f.read()
        f.close()
        self._add_data_to_options(data)

    def _add_data_to_options(self, data):
        try:
            data = json.loads(data)
        except Exception as err:
            self.exit(
                status="UNKNOWN",
                message="unable to parse json data: %s" % err)

        if not isinstance(data, dict):
            self.exit(
                status="UNKNOWN",
                message="invalid json input data - expect a dict")
        for arg in data:
            if arg not in self.option_rules:
                self.exit(
                    status="UNKNOWN",
                    message="Unknown option: %s" % arg)
            if self.option_rules[arg]["command_line_only"] == 1:
                self.exit(
                    status="UNKNOWN",
                    message="option %s is a command line argument" % arg)
            self.option[arg] = data[arg]

    def _validate_options(self):
        for opt in self.option:
            if opt not in self.option_rules:
                self.exit(
                    status="UNKNOWN",
                    message="Unknown option: %s" % opt)
        for opt in self.option_rules:
            rules = self.option_rules[opt]
            self._validate_datatype_bool(opt, rules)
            if opt in self.option:
                self._validate_multiple_option(opt, rules)
                self._validate_datatype(opt, rules)
            else:
                self._validate_optional_option(opt, rules)

        if len(self.options_no_mixin):
            for arr in self.options_no_mixin:
                hits = 0
                for key in arr:
                    if key in self.option and self.option[key] is not None:
                        hits += 1
                if hits > 1:
                    self.exit(
                        status="UNKNOWN",
                        message=(
                            "mixin the following options is "
                            "not allowed: %s" % ", ".join(arr))
                        )

    def _validate_datatype_bool(self, opt, rules):
        if rules["datatype"] == "bool":
            if self.option.get(opt, None) in [None, False, 0, "0"]:
                self.option[opt] = False
            else:
                self.option[opt] = True

    def _validate_optional_option(self, opt, rules):
        if rules["default"] is not None:
            self.option[opt] = rules["default"]
        elif rules["optional"] is True:
            self.option[opt] = None
        elif "suggest-options" not in self.option:
            self.exit(
                status="UNKNOWN",
                message="Missing mandatory option '%s'" % opt)

    def _validate_multiple_option(self, opt, rules):
        if rules["multiple"] and not isinstance(self.option[opt], list):
            self.exit(
                status="UNKNOWN",
                message="Option '%s' must be a list" % opt)

    def _validate_datatype(self, opt, rules):
        new_values = []
        values = self._get_values_as_list(opt, rules)

        for value in values:
            err = False

            if rules["datatype"] == "str" or rules["datatype"] == "lstr":
                if not isinstance(value, str):
                    # Convert int/float to string, no need to return an error.
                    if isinstance(value, (int, float)):
                        value = str(value)
                    else:
                        self._log("err datatype str", opt)
                        err = True
            elif rules["datatype"] == "int":
                if not isinstance(value, int):
                    # Fix int. Maybe a string is passed accidentally.
                    if isinstance(value, str) and re.search(r"^-?\d+\Z", value):
                        value = int(value)
                    else:
                        self._log("err datatype int", opt)
                        err = True
            elif rules["datatype"] == "float":
                if not isinstance(value, (float, int)):
                    # Fix float. Maybe a string is passed accidentally.
                    if (isinstance(value, str) 
                            and re.search(r"^-?\d+(\.\d+){0,1}\Z", value)):
                        if re.search(r"^\d+\Z", value):
                            value = int(value)
                        else:
                            value = float(value)
                    else:
                        self._log("err datatype float", opt)
                        err = True
            elif rules["datatype"] == "json-kv" or rules["datatype"] == "json-raw":
                if not isinstance(value, dict):
                    if (isinstance(value, str)
                            and re.search(r"^\s*{.*}\s*$", value, re.S)):
                        try:
                            value = json.loads(value)
                        except Exception as err:
                            self._log("err datatype json-*", opt)
                            err = True
                    else:
                        self._log("err datatype dict", opt)
                        err = True
                if err != True and rules["datatype"] == "json-kv":
                    for k,v in value.items():
                        if (not isinstance(k, str)
                                or not isinstance(v, (str, int, float))):
                            self._log("err datatype json-kv")
                            err = True
            elif rules["datatype"] == "wtrm":
                if not isinstance(value, list):
                    self._log("err datatype list", opt)
                    err = True
            elif rules["datatype"] == "metric_condition":
                if not isinstance(value, list) or len(value) != 3:
                    self._log("err datatype list", opt)
                    err = True
                if (value[0] not in self.metrics
                        or value[1] not in ["lt","le","gt","ge","eq","ne"]
                        or not isinstance(value[2], (str, int, float))):
                    self._log("err invalid metric_condition", opt)
                    err = True
                if err is not True:
                    _val = str(value[2])
                    if re.search(r"^\d+(?:\.\d+){0,1}\s*[KkMmGgTtPpEeZzYyCc]B", _val):
                        value[2] = self.convert_to_bytes(value[2])
                    elif re.search(r"^\d+(?:\.\d+){0,1}\s*[KkMmGgTtPpEeZzYyCc]", _val):
                        value[2] = self.convert_to_bits(_val)
                    elif not re.search(r"^\d+(?:\.\d+){0,1}\Z", _val):
                        self._log("err invalid metric_condition value", opt)
                        err = True
            elif rules["datatype"] == "string_condition":
                if not isinstance(value, list) or len(value) != 3:
                    self._log("err datatype list", opt)
                    err = True
                if (value[0] not in self.strings
                        or value[1] not in ["eq", "ne", "re", "rn"]
                        or not isinstance(value[2], str)):
                    self._log("err invalid string_condition", opt)
                    err = True
            elif rules["datatype"] == "json_condition":
                if not isinstance(value, list) or len(value) != 3:
                    self._log("err datatype list", opt)
                    err = True
                if (not isinstance(value[0], str)
                        or value[1] not in ["lt","le","gt","ge","eq","ne","re","rn"]
                        or not isinstance(value[2], (str, int, float))):
                    self._log("err invalid json_condition", opt)
                    err = True

            if (err == False
                    and (rules["minvalue"] is not None and value < rules["minvalue"])
                    or (rules["maxvalue"] is not None and value > rules["maxvalue"])):
                self._log("min/max exceeded", opt)
                err = True

            if rules["regex"] is not None:
                if isinstance(rules["regex"], list):
                    hit = False
                    for rx in rules["regex"]:
                        if re.search(rx, str(value)):
                            hit = True
                            break
                    if hit is False:
                        self._log("err regex list", opt)
                        err = True
                elif not re.search(rules["regex"], str(value)):
                    self._log("err regex str", opt)
                    err = True

            if err is True:
                self.exit(
                    status="UNKNOWN",
                    message="Invalid value passed to option '%s'" % opt)
            new_values.append(value)
        self._set_new_values(opt, rules, new_values)

    def _get_values_as_list(self, opt, rules):
        if rules["multiple"] is True:
            return self.option[opt]
        return [self.option[opt]]

    def _set_new_values(self, opt, rules, values):
        if rules["multiple"] is True:
            self.option[opt] = values
        else:
            self.option[opt] = values[0]

    def _print_help(self):
        print("Usage:", self.shortname, "[ OPTIONS ]", "\n")
        if self.description:
            if isinstance(self.description, str):
                print(self.description)
            else:
                print("\n".join(self.description))
        print("\nOptions:")
        for key in self.options_order:
            opt = self.option_rules[key]
            if opt["command_line_only"] == 2:
                continue
            print("\n  --%s" % opt["option"], end="")
            if "hint" in opt:
                print(" <" + str(opt["hint"]) + ">", end="")
            self._print_maxlen(re.split(r"\s+", opt["description"]), 4)
            print()
            if "minvalue" in opt:
                print("    Min:", opt["minvalue"])
            if "maxvalue" in opt:
                print("    Max:", opt["maxvalue"])
            if "default" in opt:
                print("    Default:", opt["default"])
        if len(self.examples):
            print("\nExamples:\n")
            for example in self.examples:
                print("    %s" % example)
        if len(self.metrics):
            print(
                "\nMetric options\n\n"
                "  Example: --warning [metric,operator,threshold]\n\n"
                "  The value is a json list.\n\n"
                "  Keys:")
            for key in sorted(self.metrics):
                print("    %s" % key, end="")
                self._print_maxlen(
                    re.split(r" ", self.metrics[key]["description"]),
                    6)
                print("\n      Unit: %s" % self.metrics[key]["unit"])
            print(
                "\n"
                "  Operators:\n"
                "    lt = less than\n"
                "    le = less than or equal\n"
                "    gt = greater than\n"
                "    ge = greater than or equal\n"
                "    eq = equal\n"
                "    ne = not equal\n")
        if len(self.strings):
            print(
                "String matches:\n\n"
                "  Example: --match-warning [key,operator,string]\n\n"
                "  The value is a json list.\n\n"
                "  Keys:")
            for key in sorted(self.strings):
                print("    %s" % key, end="")
                self._print_maxlen(
                    re.split(r" ", self.strings[key]["description"]),
                    6)
                print()
            print(
                "\n"
                "  Operators:\n"
                "    eq = equal\n"
                "    ne = not equal\n"
                "    re = match regex\n"
                "    rn = match regex negated\n")
        if self._has_json_tree:
            print(
                "JSON matches:\n\n"
                "  Example: --match-json-warning [a->b->c,operator,string]\n\n"
                "  The value is a json list.\n")
            print(
                "  Operators:\n"
                "    eq = equal\n"
                "    ne = not equal\n"
                "    re = match regex\n"
                "    rn = match regex negated\n"
                "    lt = less than\n"
                "    le = less than or equal\n"
                "    gt = greater than\n"
                "    ge = greater than or equal\n")
        print(
            "\nPassing parameter over STDIN as JSON string:\n\n"
            "  It's possible to pass all options as JSON string\n"
            "  to STDIN of the check. The options are passed\n"
            "  without the prefix '--' and minus signs between\n"
            "  them are used as underscore. The values of options\n"
            "  which can be specified multiple times are passed as\n"
            "  a list.\n\n"
            "    %s --stdin <<EOT\n"
            '    {"option1":"value1","option2":["value2","value3"]}\n'
            "    EOT\n" % self.shortname)
        sys.exit(0)

    def _print_maxlen(self, parts, ws, maxlen=50):
        cur_len = 0
        prefix = " " * ws
        print("\n%s" % prefix, end="")
        for part in parts:
            if cur_len >= maxlen:
                print("\n%s" % prefix, end="")
                cur_len = 0
            elif cur_len != 0:
                print(" ", end="")
                cur_len += 1
            print(part, end="")
            cur_len += len(part)

    def _print_version(self):
        print(self.progname, "v" + self.version)
        sys.exit(0)

    def _print_plugin_info(self):
        for chart in self.charts:
            for item in chart["series"]:
                item["item_id"] = self.metrics[item["name"]]["item_id"]
        info = {
            "command": self.shortname,
            "version": self.version,
            "service": self.service,
            "category": self.category,
            "abstract": self.abstract,
            "description": self.description,
            "info": self.info,
            "frequency": self.frequency,
            "network_service": self.network_service,
            "default_location": self.default_location,
            "examples": self.examples,
            "options": {},
            "options_order": [],
            "options_no_mixin": self.options_no_mixin,
            "metrics": self.metrics,
            "strings": self.strings,
            "charts": self.charts,
        }
        for opt in self.options_order:
            option = self.option_rules[opt]
            if option["command_line_only"] > 0:
                continue
            info["options_order"].append(opt)
            info["options"][opt] = {}
            for key in option:
                #if key in ["hint", "regex", "option"]:
                #    continue
                info["options"][opt][key] = option[key]
        if "pretty" in self.option and self.option["pretty"] is True:
            print(json.dumps(info, indent=4, sort_keys=True))
        else:
            print(json.dumps(info))
        sys.exit(0)

    def _print_plugin_dstat(self):
        data = {"num_metrics": len(self.metrics)}
        """
        The storage requirements for a metric depend on various factors.
        The stat table has the fields
            time 8: bytes
            service_id: 8 bytes
            satellite_id: 8 bytes
            item_id: 2 bytes
            value: 8 bytes
        In total, that's 34 bytes per metric per row. Then there's the
        size of 2 indexes, which can't be calculated in general terms.
        In our tests, we filled an empty PostgreSQL database with metrics
        for 365 days and an interval of 60 seconds, and came up with a
        size of 41 MB. If you break this down, one metric per interval
        requires around 82 bytes.
            42991616 / 365 / 1440 = rounded up 82 bytes
        """
        data["bytes_per_interval"] = data["num_metrics"] * 82
        data["metrics_item_id"] = {}
        for key, opts in self.metrics.items():
            data["metrics_item_id"][key] = opts["item_id"]
        if "pretty" in self.option and self.option["pretty"] is True:
            print(json.dumps(data, indent=4, sort_keys=True))
        else:
            print(json.dumps(data))
        #fmt = "#  %10s  %4s  %8s  %8s"
        #print(fmt % ("DS SIZE", "DAYS", "SERVICES", "INTERVAL"))
        #for interval in [60, 120, 240, 300]:
        #    metrics_per_day = 86400 / interval
        #    for services in [1, 25, 50, 100, 250, 500, 1000]:
        #        for days in [1, 10, 30, 60, 90, 180, 365]:
        #            size = self.convert_bytes_to_string(
        #                services * days * metrics_per_day * data["bytes_per_interval"],
        #                ws=" ")
        #            print(fmt % (size, days, services, interval))
        sys.exit(0)

    def check_metrics(self, stats, upshot_keys=None, factor=None):
        """
        check_thresholds(
            stats=stats,
            upshot_keys=["foo", "bar", "baz"])
        """
        seen_keys = []
        for opt in ["critical", "alert", "warning", "notice"]:
            if self.option[opt] is None:
                continue
            status = opt.upper()
            for rule in self.option[opt]:
                key, op, threshold = rule
                if key in seen_keys:
                    continue
                if factor is not None and factor > 0:
                    threshold = float(threshold) * factor
                if self.compare(stats[key], op, threshold) is True:
                    message = []
                    message.append(
                        "%s = %s"
                        % (
                            self.metrics[key]["name"],
                            self._gen_hr_value(key, stats[key])
                        ))
                    message.append(self.convert_operator_to_string(op))
                    message.append(self._gen_hr_value(key, threshold))
                    if factor is not None:
                        message.append("(factor=%s)" % factor)
                    message.append("[%s]" % status)
                    seen_keys.append(key)
                    self.add_exit_message(" ".join(message))
                    self.set_exit_status(status)
        if upshot_keys is not None:
            for key in upshot_keys:
                if key not in seen_keys:
                    message = (
                        "%s = %s"
                        % (self.metrics[key]["name"], self._gen_hr_value(key, stats[key])))
                    self.add_exit_message(message)

    def check_matches(self, data, upshot_keys=None):
        seen_keys = []
        for opt in [
                "match-critical", "match-alert",
                "match-warning", "match-notice"]:
            if self.option[opt] is None:
                continue
            status = re.split(r"-", opt)[1].upper()
            for rule in self.option[opt]:
                key, op, match = rule
                if self.compare(data[key], op, match) is True:
                    if op in ["eq", "re"]:
                        op_hr = "found"
                    if op in ["ne", "rn"]:
                        op_hr = "not found"
                    self.set_exit_status(status)
                    self.add_exit_message(
                        "string '%s' %s in %s [%s]"
                        % (match, op_hr, key, status))

        if upshot_keys is not None:
            for key in upshot_keys:
                if key in data:
                    self.add_exit_message("%s=%s" % (key, data[key]))

    def check_json_tree(self, data):
        for status in ["critical", "alert", "warning", "notice"]:
            opt = "match-json-%s" % status
            if (opt not in self.option
                    or self.option[opt] == None
                    or not len(self.option[opt])):
                continue
            # convert once
            if isinstance(data, str):
                try:
                    data = json.loads(data)
                except Exception:
                    self.exit(
                        status="CRITICAL",
                        message="unable to parse json data")
            for item in self.option[opt]:
                parse_date = None
                string, op, value = item
                tree = re.split(r"\s*->\s*", string)
                tree_value = data
                for key in tree:
                    if isinstance(tree_value, dict) and key in tree_value:
                        tree_value = tree_value[key]
                    elif isinstance(tree_value, list) and re.search(r"^\d+\Z", key):
                        tree_value = tree_value[int(key)]
                    else:
                        self.add_exit_message(
                            "unable to parse json data, "
                            "stopped at key %s" % key)
                        self.set_exit_status("CRITICAL")
                if not isinstance(tree_value, (str, int, float)):
                    self.add_exit_message("value in json data is not of type int/float/str")
                    self.set_exit_status("CRITICAL")

                if isinstance(value, str) and isinstance(tree_value, str):
                    m = re.search(r"^DATE\(([\+\-])(\d+)d\)\Z", value)
                    n = re.search(r"(\d\d\d\d-\d\d-\d\d)", tree_value)
                    if m and n:
                        try: tree_value = utils.unixtime(n.group(1))
                        except Exception: pass
                        if isinstance(tree_value, int):
                            value = int(_time.time())
                            if m.group(1) == "+":
                                value += 86400 * int(m.group(2))
                            else:
                                value -= 86400 * int(m.group(2))
                            parse_date = True

                if op in ["ge", "le", "gt", "lt"]:
                    if isinstance(value, str):
                        if re.search(r"^\d+\Z", value):
                            value = int(value)
                        elif re.search(r"^\d+(:?\.\d+)?\Z", value):
                            value = float(value)
                        else:
                            self.add_exit_message("value must be of type int/float")
                            self.set_exit_status("UNKNOWN")
                    if isinstance(tree_value, str):
                        if re.search(r"^\d+(:?\.\d+)?\Z", tree_value):
                            tree_value = float(tree_value)
                        else:
                            self.add_exit_message("value found in json is not of type int/float")
                            self.set_exit_status("CRITICAL")
                else:
                    tree_value = str(tree_value)
                    value = str(value)

                if self.compare(tree_value, op, value) is True:
                    if parse_date:
                        tree_value = utils.timestamp(tree_value, time_format="%Y-%m-%d")
                        value = utils.timestamp(value, time_format="%Y-%m-%d")
                    self.add_exit_message(
                        "json tree: %s=%s %s %s" % (
                            string,
                            tree_value,
                            self.convert_operator_to_string(op),
                            value
                        ))
                    self.set_exit_status(status)
                elif parse_date:
                    tree_value = utils.timestamp(tree_value, time_format="%Y-%m-%d")
                    value = utils.timestamp(value, time_format="%Y-%m-%d")
                    self.add_exit_message(
                        "json tree %s=%s %s %s" % (
                            string,
                            tree_value,
                            self.convert_operator_to_string(op),
                            value
                        ))

    def _gen_hr_value(self, key, value):
        if self.metrics[key]["unit"] == "bytes":
            return self.convert_bytes_to_string(value)
        if self.metrics[key]["unit"] == "percent":
            return str(value) + "%"
        if self.metrics[key]["unit"] == "ms":
            return str(value) + "ms"
        if self.metrics[key]["unit"] == "seconds":
            return str(value) + "s"
        return str(value)

    def get_deltas(self, data=None, ratio_data=None):
        time = int(_time.time())
        init = self.load_json()
        if data is not None:
            deltas = self.calculate_deltas(init, data, time)
        else:
            deltas = {}
        if ratio_data is None:
            self.save_json({"data": data, "time": time})
            return deltas

        if init is None:
            init = {}
        ratio = {}
        for range_time in ratio_data:
            range_time_str = str(range_time)
            ratio[range_time_str] = self._update_data_list(
                data=ratio_data[range_time],
                data_list=init.get("ratio", {}).get(range_time_str, None),
                range_time=range_time,
                time=time)
            deltas.update(
                self.calculate_ratio(
                    ratio[range_time_str][0]["data"],
                    ratio[range_time_str][-1]["data"]))
        self.save_json({"data": data, "time": time, "ratio": ratio})
        return deltas

    def _update_data_list(self, data, data_list, range_time, time):
        expire_time = time - int(range_time)
        if data_list is None:
            data_list = []
        while len(data_list) > 0:
            if data_list[0]["time"] < expire_time:
                data_list.pop(0)
            else:
                break
        if len(data_list) > 0:
            init_data = data_list[0]
        else:
            init_data = None
        data_list.append({"data": data, "time": time})
        return data_list

    def calculate_ratio(self, a, b):
        ret = {}
        diff = {}
        total = 0
        for key in b:
            if key in a:
                diff[key] = b[key] - a[key]
                total += diff[key]
            else:
                diff[key] = 0
        for key in diff:
            if total > 0:
                ret[key] = float("%.2f" % (100 * diff[key] / total))
            else:
                ret[key] = 0.0
        return ret

    def calculate_deltas(self, init, data, time):
        ret = {}
        if (init is None
                or len(init) == 0
                or "data" not in init
                or len(init["data"]) == 0
                or "time" not in init
                or not isinstance(init["time"], int)
                or int(time) - int(init["time"]) == 0):
            for key in data:
                ret[key] = 0
        else:
            delta_time = int(time) - int(init["time"])
            init_data = init["data"]
            for key in data:
                # New key doesn't exist in old cache
                if key in init_data:
                    init_value = float(init_data[key])
                else:
                    init_value = 0
                cur_value = float(data[key])
                # Handle counter? No!
                # 32bit = 2147483647
                # 64bit = 9223372036854775807
                # The problem is that if the host is rebooted... the difference
                # that would be calculated could be too high in this case!
                # For this reason the statistic is set to 0.00.
                # If anybody has another idea... let me know it.
                if init_value >= cur_value:
                    ret[key] = 0
                else:
                    ret[key] = float("%.2f" % ((cur_value - init_value) / delta_time))
        return ret

    def compare(self, value_a, op, value_b):
        if op == "ge":
            return float(value_a) >= float(value_b)
        if op == "le":
            return float(value_a) <= float(value_b)
        if op == "gt":
            return float(value_a) > float(value_b)
        if op == "lt":
            return float(value_a) < float(value_b)
        if op == "eq":
            return str(value_a) == str(value_b)
        if op == "ne":
            return str(value_a) != str(value_b)
        if op == "re":
            if re.search(value_b, value_a, re.S):
                return True
            return False
        if op == "rn":
            if re.search(value_b, value_a, re.S):
                return False
            return True

    def convert_bytes_to_string(self, value, ws=""):
        value = float(value)
        unit = None
        if value >= 1208925819614629174706176:
            value /= 1208925819614629174706176
            unit = "YB"
        elif value >= 1180591620717411303424:
            value /= 1180591620717411303424
            unit = "ZB"
        elif value >= 1152921504606846976:
            value /= 1152921504606846976
            unit = "EB"
        elif value >= 1125899906842624:
            value /= 1125899906842624
            unit = "PB"
        elif value >= 1099511627776:
            value /= 1099511627776
            unit = "TB"
        elif value >= 1073741824:
            value /= 1073741824
            unit = "GB"
        elif value >= 1048576:
            value /= 1048576
            unit = "MB"
        elif value >= 1024:
            value /= 1024
            unit = "KB"
        if unit is None:
            return str(int(value))
        return "%.1f%s%s" % (value, ws, unit)

    def convert_to_bytes(self, value):
        match = re.match(r"^(\d+(?:\.\d+){0,1})\s*([KkMmGgTtPpEeZzYyCc])B\Z", value)
        if match:
            value = float(match.group(1))
            unit = match.group(2).upper()
            if unit == "K":
                return value * 1024
            if unit == "M":
                return value * 1048576
            if unit == "G":
                return value * 1073741824
            if unit == "T":
                return value * 1099511627776
            if unit == "P":
                return value * 1125899906842624
            if unit == "E":
                return value * 1152921504606846976
            if unit == "Z":
                return value * 1180591620717411303424
            if unit == "Y":
                return value * 1208925819614629174706176
        return float(value)

    def convert_to_bits(self, value):
        match = re.match(r"^(\d+(?:\.\d+){0,1})\s*([KkMmGgTtPpEeZzYyCc])b?\Z", value)
        if match:
            value = float(match.group(1))
            unit = match.group(2).upper()
            if unit == "K":
                return value * 1000
            if unit == "M":
                return value * 1000000
            if unit == "G":
                return value * 1000000000
            if unit == "T":
                return value * 1000000000000
            if unit == "P":
                return value * 1000000000000000
            if unit == "E":
                return value * 1000000000000000000
            if unit == "Z":
                return value * 1000000000000000000000
            if unit == "Y":
                return value * 1000000000000000000000000
        return float(value)

    def convert_operator_to_string(self, op):
        if op == "ge":
            return ">="
        if op == "le":
            return "<="
        if op == "gt":
            return ">"
        if op == "lt":
            return "<"
        if op == "eq":
            return "=="
        if op == "ne":
            return "!="
        if op == "re":
            return "=~"
        if op == "rn":
            return "!~"

    def runtime(self):
        """
        Return the runtime between two calls of plugin.runtime().

            import time
            plugin.runtime()
            time.sleep(3)
            runtime1 = plugin.runtime()
            time.sleep(2)
            runtime2 = plugin.runtime()

            print("runtime1:", runtime1)
            print("runtime2:", runtime2)

        The output should be (as example):

            runtime1: 3.000
            runtime2: 2.000

        """
        ret = format(_time.time() - self._runtime, ".3f")
        self._runtime = _time.time()
        return ret

    def get_time_of_day(self):
        """
        Returns the time of day in unix time as floating
        point number with 3 digits after point.
        """
        return format(_time.time(), ".3f")

    def get_secret_file(self):
        """
        Get the content of a secret file:
            (username, password) = plugin.get_secret_file()

        The format of the secret file must be:
            user = username
            password = passord

        Or
            username = username
            password = password

        Whitepaces within the username or password are not allowed.
        """

        if "secret-file" not in self.option:
            self.exit(
                status="UNKNOWN",
                message="no secret file defined")
        if not re.search(r"^[\w\.\-]+\Z", self.option["secret-file"]):
            self.exit(
                status="UNKNOWN",
                message=(
                    "invalid characters found in file name. "
                    "Allowed signs are: a-zA-Z0-9_.-")
            )
        filename = "%s/%s" % (self.config_path, self.option["secret-file"])
        username = None
        password = None

        try:
            with open(filename, encoding="utf-8") as f:
                for line in f:
                    match = re.search(
                        r"^(user|username|password)\s*=\s*([^\s]+)",
                        line
                    )
                    if match:
                        if (match.group(1) == "user"
                                or match.group(1) == "username"):
                            username = match.group(2)
                        elif match.group(1) == "password":
                            password = match.group(2)
        except Exception:
            self.exit(
                status="UNKNOWN",
                message="unable to open secret file '" + filename + "'")

        return username, password

    def add_mtr(self, host, timeout=30):
        self._get_mtr_by_addr = host
        self._get_mtr_timeout = timeout

    def get_mtr(self, ipaddr, timeout=30):
        """
        Execute an MTR and return the result as dict.

            plugin.get_mtr(ipaddr=ipaddr, timeout=timeout)

        The default timeout is 30 seconds.
        """
        mtr_data = []
        result = {
            "result": mtr_data,
            "ipaddr": ipaddr,
            "status": "ok",
        }
        runcmd = RunCmd(
            command=["mtr", "--no-dns", "-trc", "3", ipaddr],
            timeout=timeout)
        if runcmd.has_failed:
            result["status"] = "err"
            if runcmd.has_timed_out:
                result["message"] = "MTR timed out after %s seconds" % timeout
            else:
                result["message"] = (
                    "an unexpected error occurs, "
                    "please contact the administrator")
            return result

        lines = runcmd.stdout()
        header = lines.pop(0)

        for line in lines:
            data = line.split()
            data[0] = re.sub(r"\|.+", "", data[0])
            data[2] = re.sub(r"%", "", data[2])
            mtr_data.append({
                "step": data[0],
                "ipaddr": data[1],
                "loss": data[2],
                "snt": data[3],
                "last": data[4],
                "avg": data[5],
                "best": data[6],
                "wrst": data[7],
                "stdev": data[8]
            })

        return result

    def load_json(self, filename=None):
        """
        Load json data from a temporary file.
        """
        if filename is None:
            filename = self.libfile()
        try:
            f = open(filename, "r")
            content = f.read()
            f.close()
            return json.loads(content)
        except Exception:
            return None

    def save_json(self, data, filename=None):
        """
        Save json data to a temporary file.
        """
        if filename is None:
            filename = self.libfile()

        try:
            data = json.dumps(data)
        except Exception:
            self.exit(
                status="UNKNOWN",
                message=(
                    "Internal plugin error: unable to "
                    "convert data into json string"
                ))
        try:
            f = open(filename, "w")
            f.truncate()
            f.write(data)
            f.close()
        except Exception as err:
            self.exit(
                status="UNKNOWN",
                message=(
                    "Internal plugin error: unable to write "
                    "json string to lib file %s: %s" % (filename, str(err))
                ))

    def load_cache(self):
        filename = self.cachefile()
        try:
            f = open(filename, "r")
            if platform.system() != "Windows":
                fcntl.flock(f, fcntl.LOCK_EX)
            content = f.read()
            if platform.system() != "Windows":
                fcntl.flock(f, fcntl.LOCK_UN)
            f.close()
            return json.loads(content)
        except Exception:
            return None

    def save_cache(self, data):
        filename = self.cachefile()
        try:
            data = json.dumps(data)
        except Exception:
            self.exit(
                status="UNKNOWN",
                message=(
                    "Internal plugin error: unable to "
                    "convert data into json string"
                ))
        try:
            f = open(filename, "w")
            if platform.system() != "Windows":
                fcntl.flock(f, fcntl.LOCK_EX)
            f.truncate()
            f.write(data)
            if platform.system() != "Windows":
                fcntl.flock(f, fcntl.LOCK_UN)
            f.close()
        except Exception:
            self.exit(
                status="UNKNOWN",
                message=(
                    "Internal plugin error: unable to "
                    "write json string to cache file"
                ))

    def set_exit_status(self, status):
        if status not in mstatus.CODE:
            self.exit(
                status="UNKNOWN",
                message="plugin error: invalid status '%s'" % status)
        if mstatus.CODE[status] > mstatus.CODE[self.exit_status]:
            self.exit_status = status
            return True
        return False

    def get_exit_status(self):
        return mstatus.CODE[self.exit_status]

    def add_exit_message(self, message):
        if self.exit_message is None:
            self.exit_message = []
        if isinstance(message, list):
            self.exit_message += message
        else:
            self.exit_message.append(message)

    def exit(
            self, status=None, message=None,
            stats=None, debug=None, result=None,
            return_status=None):
        """
        exit(
            status  = OK | 0 | WARNING | 1 | CRITICAL | 2 | UNKNOWN | 3
            message = a short message as string or list, if it's a list
                      then the message is joined via ", "
            stats   = metrics as dict
            debug   = debug information as dict
            result  = result data, as example from web transactions
            return  = return instead print and exit, for debugging
        )
        """
        if ("validate-metrics" in self.option
                and self.option["validate-metrics"]):
            self._validate_metrics(stats)
        if status is None:
            status = self.exit_status
        if status not in mstatus.CODE:
            self.exit(
                status="UNKNOWN",
                message="plugin error: invalid status '%s'" % status)
        if message is None:
            message = []
        elif not isinstance(message, list):
            message = [str(message)]
        if self.exit_message is not None:
            message += self.exit_message
        if len(message) == 0:
            message.append(status)
        output = {
            "status": mstatus.STRING[status],
            "message": ", ".join(message)}
        if len(output["message"]) > 500:
            output["message"] = "%s..." % output["message"][0:497]
        if stats is not None:
            output["stats"] = stats
        if isinstance(debug, dict):
            output["debug"] = debug
        if result is not None and isinstance(result, list):
            output["result"] = result
        if self._get_mtr_by_addr is not False:
            if "debug" not in output:
                output["debug"] = {}
            output["debug"]["mtr"] = self.get_mtr(
                self._get_mtr_by_addr,
                self._get_mtr_timeout)
        if return_status == True:
            output["exitcode"] = mstatus.CODE[status]
            return output
        if "pretty" in self.option and self.option["pretty"] is True:
            print(json.dumps(output, indent=4, sort_keys=True))
        else:
            print(json.dumps(output))
        sys.exit(mstatus.CODE[status])

    def convert_to_number(self, num):
        if re.search(r"^\d+\.\d+", num):
            return float(num)
        return int(num)

    def _log(self, *args):
        if self.debug is not False:
            msg = []
            for arg in args:
                msg.append(str(arg))
            if isinstance(self.debug, list):
                self.debug.append(" ".join(msg))
            else:
                print(msg)

    def _add_default_options(self):
        self._add_option_version()
        self._add_option_help()
        self._add_option_plugin_info()
        self._add_option_plugin_dstat()
        self._add_option_pretty()
        self._add_option_suggest_options()
        self._add_option_stdin()
        self._add_option_input_file()
        self._add_option_host_id()
        self._add_option_service_id()
        self._add_option_validate_metrics()
        self._add_option_plugin_cache_dir()

    def _add_metric_options(self):
        for status in ["Notice", "Warning", "Alert", "Critical"]:
            self.add_option(
                name="%s threshold" % status,
                option=status.lower(),
                description=(
                    "Trigger a event with status '%s' if the threshold "
                    "is exceeded. Large values can be specified in a "
                    "human-readable format with a unit, such as 10KB "
                    "for 10 kilobytes or 10K for 10 kilos."
                    % status.upper()),
                datatype="metric_condition",
                hint='["key", "op", "value"]',
                multiple=True)

    def _add_string_matches(self):
        for status in ["Notice", "Warning", "Alert", "Critical"]:
            self.add_option(
                name="%s match" % status,
                option="match-%s" % status.lower(),
                description=(
                    "With this option it's possible "
                    "to check content for matches."),
                datatype="string_condition",
                hint='["key", "op", "value"]',
                multiple=True)   

    def _add_option_version(self):
        self.add_option(
            name="Version",
            option="version",
            description="Print the version.",
            datatype="bool",
            command_line_only=1)

    def _add_option_help(self):
        self.add_option(
            name="Help",
            option="help",
            description="Print the help.",
            datatype="bool",
            command_line_only=1)

    def _add_option_plugin_info(self):
        self.add_option(
            name="Plugin information",
            option="plugin-info",
            description="Print plugin information as JSON string.",
            datatype="bool",
            command_line_only=1)

    def _add_option_plugin_dstat(self):
        self.add_option(
            name="Plugin information",
            option="plugin-dstat",
            description="Print plugin information about the database table size.",
            datatype="bool",
            command_line_only=1)

    def _add_option_pretty(self):
        self.add_option(
            name="Pretty output",
            option="pretty",
            description="Print the plugin information in pretty format",
            datatype="bool",
            command_line_only=1)

    def _add_option_suggest_options(self):
        self.add_option(
            name="Suggest options",
            option="suggest-options",
            description="Suggest options for auto discovery.",
            datatype="bool",
            command_line_only=1)

    def _add_option_stdin(self):
        self.add_option(
            name="STDIN",
            option="stdin",
            description="Read options as json string from stdin.",
            datatype="bool",
            command_line_only=1)

    def _add_option_input_file(self):
        self.add_option(
            name="Input File",
            option="input-file",
            description="Read options as json string from a file.",
            datatype="str",
            command_line_only=1)

    def _add_option_host_id(self):
        self.add_option(
            name="Host ID",
            option="internal-host-id",
            hint="id",
            datatype="str",
            regex=r"^[a-z0-9\-\.]+\Z",
            description="The host ID of the check for internal usage.",
            command_line_only=1,
            default=self.internal_host_id)

    def _add_option_service_id(self):
        self.add_option(
            name="Service ID",
            option="internal-service-id",
            hint="id",
            datatype="int",
            description="The service ID of the check for internal usage.",
            command_line_only=1,
            default=self.internal_service_id)

    def _add_option_validate_metrics(self):
        self.add_option(
            name="Validate metrics",
            option="internal-validate-metrics",
            description=(
                "Validate if all defined metrics are found in "
                "statistic output. Useful for testing purposes."),
            datatype="bool",
            command_line_only=1)

    def _add_option_plugin_cache_dir(self):
        self.add_option(
            name="Cache directory",
            option="internal-plugin-cache-dir",
            hint="dir",
            datatype="str",
            description="The path for cache files.",
            command_line_only=1,
            default="/tmp")

    def _add_option(self, args, **kwargs):
        kwargs.update(args)
        self.add_option(**kwargs)

    def _validate_metrics(self, stats):
        if stats is None:
            print("no metrics found")
            sys.exit(1)
        if not isinstance(stats, dict):
            print("wrong stats format")
            sys.exit(1)
        exitcode = 0
        for key in stats:
            if key not in self.metrics:
                print("metric %s not defined" % key)
                exitcode = 1
        for key in self.metrics:
            if key not in stats:
                print("metric %s not in stats" % key)
                exitcode = 1
        if exitcode == 0:
            print("all metrics found")
        sys.exit(exitcode)

    def has_metric(
            self, key, item_id, name, description,
            datatype="float", unit="none",
            metric_type="gauge"):
        if not re.search(r"^[a-zA-Z_0-9\.\-]+\Z", key):
            raise Exception("invalid key format '%s'" % key)
        if item_id is None:
            raise Exception("metric %s has no item_id" % key)
        if not isinstance(item_id, int):
            raise Exception("invalid item_id %s" % item_id)
        if item_id in self.item_id:
            raise Exception("item_id %s already exist" % item_id)
        if datatype not in ["int", "float"]:
            self.exit(
                status="UNKNOWN",
                message=(
                    "Internal plugin error: "
                    "invalid datatype for metric %s"
                    % key))
        if unit not in ["none", "bytes", "percent", "ms", "seconds", "temperature"]:
            self.exit(
                status="UNKNOWN",
                message=(
                    "Internal plugin error: "
                    "invalid unit for metric %s" % key))
        self.metrics[key] = {
            "name": name,
            "item_id": item_id,
            "datatype": datatype,
            "description": description,
            "unit": unit,
            "metric_type": metric_type,
        }

    def has_string(self, key, name, description):
        if not re.search(r"^[a-zA-Z_0-9\.\-]+\Z", key):
            raise Exception("invalid key format '%s'" % key)
        self.strings[key] = {
            "name": name,
            "description": description
        }

    def has_json_tree(self, description=None):
        self._has_json_tree = True
        if description == None:
            description = (
                "Check a JSON tree for specific values. "
                "To access a JSON tree down to the value, "
                "an arrow notation (dash, greater than sign) "
                "must be used. For example, to parse the "
                'JSON strucrute {"a":{"b":[{"c":10}]}}, the '
                "following string is required: a->b->[0]->c"
            )
        for status in ["Critical", "Alert", "Warning", "Notice"]:
            self.add_option(
                name="Match JSON %s" % status,
                option="match-json-%s" % status.lower(),
                datatype="json_condition",
                hint='[json->tree, op, value]',
                multiple=True,
                description=description)

    def has_notice(self, **kwargs):
        self._add_option(
            kwargs,
            name="Notice threshold",
            option="notice",
            hint="seconds",
            datatype="float",
            description=(
                "A value in seconds. When the check takes longer "
                "than this time then a warning status is triggered."
            ))

    def has_warning(self, **kwargs):
        self._add_option(
            kwargs,
            name="Warning threshold",
            option="warning",
            hint="seconds",
            datatype="float",
            description=(
                "A value in seconds. When the check takes longer "
                "than this time then a warning status is triggered."
            ))

    def has_alert(self, **kwargs):
        self._add_option(
            kwargs,
            name="Alert threshold",
            option="alert",
            hint="seconds",
            datatype="float",
            description=(
                "A value in seconds. When the check takes longer "
                "than this time then a warning status is triggered."
            ))

    def has_critical(self, **kwargs):
        self._add_option(
            kwargs,
            name="Critical threshold",
            option="critical",
            hint="seconds",
            datatype="float",
            description=(
                "A value in seconds. When the check takes "
                "longer than this time then a critical status is triggered."
            ))

    def has_timeout(self, **kwargs):
        self._add_option(
            kwargs,
            name="Timeout",
            option="timeout",
            hint="seconds",
            datatype="int",
            minvalue=1,
            maxvalue=30,
            description=(
                "A timeout in seconds after its expiration the check is "
                "aborted and a critical status is triggered. NOTE: the "
                "timeout may not be higher than 30 seconds!"
            ))

    def has_host(self, **kwargs):
        self._add_option(
            kwargs,
            name="Hostname or IP address",
            option="host",
            hint="hostname or ip address",
            datatype="str",
            regex=r"^[a-zA-Z0-9:.\-]+\Z",
            description="A hostname or IP address to connect to.")

    def has_ipaddr(self, **kwargs):
        self._add_option(
            kwargs,
            name="IP address",
            option="ipaddr",
            hint="address",
            datatype="str",
            regex=bloonix.rx.IPADDR,
            description="The IP address to connect to.")

    def has_port(self, **kwargs):
        # Port 0 does not exist, but maybe 0 can be
        # used as autodiscovery property.
        allow_zero = kwargs.get("allow_zero", False)

        if allow_zero is True:
            zero = "0|"
        else:
            zero = ""

        self._add_option(
            kwargs,
            name="Port number",
            option="port",
            hint="port",
            datatype="int",
            regex=(
                r"^(?:%s"
                    r"6553[0-5]|"        # 65530 - 65535
                    r"655[0-2][0-9]|"    # 65500 - 65529
                    r"65[0-4][0-9]{2}|"  # 65000 - 65499
                    r"6[0-4][0-9]{3}|"   # 60000 - 64999
                    r"[1-5][0-9]{4}|"    # 10000 - 59999
                    r"[1-9][0-9]{0,3}"   # 1 - 9999
                r")\Z" % zero
            ),
            description="A port number to connect to.")

    def has_bind_to_ipaddr(self, **kwargs):
        self._add_option(
            kwargs,
            name="Bind to IP address",
            option="bind",
            hint="ipaddr",
            datatype="str",
            regex=r"^[a-fA-F0-9\.:]+\Z",
            description="A local IP address to bind to.")

    def has_url(self, **kwargs):
        self._add_option(
            kwargs,
            name="URL",
            option="url",
            hint="url",
            datatype="str",
            regex=r"^https{0,1}://[^']+(/[^\s']*){0,1}",
            description=(
                "This is the HTTP or HTTPS request you want to check. "
                "Please enter the full URL with the query string."),
            example="http://test.example/")

    def has_use_ipv4(self, **kwargs):
        self._add_option(
            kwargs,
            name="Use IPv4",
            option="use-ipv4",
            description="Use IPv4 to connect to the host.",
            datatype="bool")

    def has_use_ipv6(self, **kwargs):
        self._add_option(
            kwargs,
            name="Use IPv6",
            option="use-ipv6",
            description="Force IPv6 to connect to the host.",
            datatype="bool")

    def has_use_ssl(self, **kwargs):
        self._add_option(
            kwargs,
            name = "Use secure connection via SSL",
            option = "use-ssl",
            description = "Use secure connection via SSL.",
            datatype="bool")

    def has_auth_basic(self, **kwargs):
        self._add_option(
            kwargs,
            name="Auth basic",
            option="auth-basic",
            hint="username:password",
            datatype="str",
            regex=r"^[^\s:]+:[^\s:]+\Z",
            description=(
                "A HTTP Auth-Basic authentication "
                "string in format username:password."
            ))

    def has_login_username(self, **kwargs):
        self._add_option(
            kwargs,
            name="Username",
            option="username",
            hint="username",
            datatype="str",
            regex=r"^[^\s]+\Z",
            description="The username to use for the login.")

    def has_login_password(self, **kwargs):
        self._add_option(
            kwargs,
            name="Password",
            option="password",
            hint="password",
            datatype="str",
            regex=r"^[^\s]+\Z",
            description="The password for the user to login.")

    def has_database_name(self, **kwargs):
        self._add_option(
            kwargs,
            name="Database",
            option="database",
            hint="database",
            datatype="str",
            regex=r"^[^\s]+\Z",
            description="Set the database to connect to.")

    def has_login_secret_file(self, **kwargs):
        self._add_option(
            kwargs,
            name="Secret file",
            option="secret-file",
            hint="filename",
            datatype="str",
            default="%s.conf" % self.shortname,
            description=(
                "This is a file to set the username and "
                "password to connect to the service. The file "
                "is placed in /etc/bloonix/agent. Example content: "
                "username=root\npassword=secret"
            ))

    def has_mailbox(self, **kwargs):
        self._add_option(
            kwargs,
            name="Mailbox",
            option="mailbox",
            hint="mailbox",
            datatype="str",
            regex=r"^[a-zA-Z]+\Z",
            description="The mailbox to query.")

    def has_debug(self, **kwargs):
        self._add_option(
            kwargs,
            name="Debug",
            option="debug",
            description=(
                "Turn on debugging. Just useful if you want to "
                "test something on the command line."),
            command_line_only=1)

    def has_test(self, **kwargs):
        self._add_option(
            kwargs,
            name="Test data",
            option="test",
            datatype="str",
            description="Pass a filename to test the plugin.",
            command_line_only=2)

    def has_test_init(self, **kwargs):
        self._add_option(
            kwargs,
            name="Initial test data",
            option="test-init",
            datatype="str",
            description="Pass a filename to test the plugin.",
            command_line_only=2)

    def get_file_content(self, filename, split=True):
        try:
            f = open(filename, "r")
            if split is True:
                data = f.readlines()
            else:
                data = f.read()
            f.close()
            return data
        except Exception:
            self.exit(
                status="UNKNOWN",
                message="unable to open file '%s' for reading" % filename)

    def has_chart(self, **chart):
        chart = validate(chart, {
            "chart_id": {
                "type": str,
            },
            "title": {
                "type": str,
            },
            "ylabel": {
                "type": str,
            },
            "chart_type": {
                "type": str,
                "options": ["line", "area"],
                "default": "line",
            },
            "units": {
                "type": str,
                "default": None,
            },
            "stack": {
                "type": bool,
                "default": None,
            },
            "series": {
                "type": dict,
                "multiple": True,
            }
        })
        if chart["stack"] is True:
            chart["stack"] = 1
        else:
            chart["stack"] = 0
        for i, series in enumerate(chart["series"]):
            chart["series"][i] = validate(series, {
                "name": {
                    "type": str,
                },
                "color": {
                    "type": str,
                    "regex": r"^#[a-fA-F0-9]{6}\Z",
                    "optional": True
                },
                "opposite": {
                    "type": bool,
                    "default": False,
                }
            })
        self.charts.append(chart)
