"""
Only pysnmp 7.0.3 compatible! Please install pysnmp:
  pip install pyasn1==0.4.8
  pip install pysnmp==7.0.3
This module should only be used on windows systems.
"""
import asyncio
import bloonix.plugin.base
import bloonix.plugin.has_snmp
import pysnmp.hlapi.v3arch.asyncio as snmp
from pysnmp.proto.rfc1905 import NoSuchObject

class Check(bloonix.plugin.base.Check, bloonix.plugin.has_snmp.HasSNMP):
    def snmp_walk(self, oid, empty_ok=False):
        return asyncio.run(self.__snmp_request(oid, snmp.walkCmd, empty_ok))

    async def __snmp_request(self, oid, snmp_object, empty_ok=False):
        data = {}

        if self.option["snmp-version"] == "1":
            version = 0
        elif self.option["snmp-version"] == "2c":
            version = 1

        g = snmp_object(
            snmp.SnmpEngine(maxMessageSize=65535),
            snmp.CommunityData(self.option["community"], mpModel=version),
            snmp.UdpTransportTarget((self.option["host"], self.option["port"])),
            snmp.ContextData(),
            snmp.ObjectType(snmp.ObjectIdentity(oid)),
            lookupMib=False,
            lexicographicMode=False)

        async for error_indication, error_status, error_index, var_binds in g:
            if error_indication:
                self.exit(
                    status="CRITICAL",
                    message="SNMP error: %s" % error_indication)
            if error_status:
                self.exit(
                    status="CRITICAL",
                    message="SNMP status: %s" % error_status)

            for oid, value in var_binds:
                oid = str(oid)

                if isinstance(value, NoSuchObject):
                    if empty_ok is True:
                        return None
                    self.exit(
                        status="CRITICAL",
                        message="no such object")

                if (isinstance(value, snmp.Integer32) or
                        isinstance(value, snmp.Integer) or
                        isinstance(value, snmp.Counter32) or
                        isinstance(value, snmp.Counter64) or
                        isinstance(value, snmp.Gauge32) or
                        isinstance(value, snmp.Unsigned32) or
                        isinstance(value, snmp.TimeTicks)):
                    data[oid] = int(value)
                else:
                    data[oid] = value.prettyPrint()

        if len(data) == 0:
            if empty_ok is True:
                return None
            self.exit(
                status="CRITICAL",
                message="oid %s not found" % oid)

        return data

    def snmp_get(self, oid, empty_ok=False):
        res = asyncio.run(self.__snmp_get(oid))
        if empty_ok is True and res is None:
            return None
        if oid in res:
            return res[oid]
        return None

    async def __snmp_get(self, oid):
        if self.option["snmp-version"] == "1":
            version = 0
        elif self.option["snmp-version"] == "2c":
            version = 1

        error_indication, error_status, error_index, var_binds = await snmp.getCmd(
            snmp.SnmpEngine(maxMessageSize=65535),
            snmp.CommunityData(self.option["community"], mpModel=version),
            snmp.UdpTransportTarget((self.option["host"], self.option["port"])),
            snmp.ContextData(),
            snmp.ObjectType(snmp.ObjectIdentity(oid)),
            lexicographicMode=False)

        if error_indication:
            self.exit(
                status="CRITICAL",
                message="SNMP error: %s" % error_indication)

        if error_status:
            self.exit(
                status="CRITICAL",
                message="SNMP status: %s" % error_status)

        if not len(var_binds):
            return

        data = {}
        oid, value = var_binds[0]
        oid = str(oid)
        if (isinstance(value, snmp.Integer32) or
                isinstance(value, snmp.Integer) or
                isinstance(value, snmp.Counter32) or
                isinstance(value, snmp.Counter64) or
                isinstance(value, snmp.Gauge32) or
                isinstance(value, snmp.Unsigned32) or
                isinstance(value, snmp.TimeTicks)):
            data[oid] = int(value)
        else:
            data[oid] = value.prettyPrint()
        return data
