from multiprocessing import Process, Value
from ctypes import c_longlong, c_byte
import os
import signal
import sys
import time

from bloonix.logger import logger

class ProcManager:
    def __init__(self,
            exit_on_sig_term=True,
            signal_hup_callback=None,
            send_hup_to_pools=None,
            restart_pools_on_hup=None):
        self.pools = []
        self.done = False
        self.log = logger("bloonix")
        self.callback = None
        self.exit_on_sig_term = exit_on_sig_term
        self.signal_hup_callback = signal_hup_callback
        self.send_hup_to_pools = send_hup_to_pools
        self.restart_pools_on_hup = restart_pools_on_hup
        self.restart_pools = None

    def add_pool(
            self, name, worker, callback,
            signal_hup_child_callback=None,
            args=[], kwargs={}):
        self.log.notice("register pool %s with %s worker" % (name, worker))
        if worker > 0:
            self.pools.append(
                Pool(name, worker, callback, signal_hup_child_callback, args, kwargs))

    def add_callback(self, callback):
        """
        This callback is called on each loop within daemonize()
        and can be used to debug things, like
            print( p.process_status() )
        """
        self.callback = callback

    def daemonize(self):
        self.log.info("start main proc manager")
        signal.signal(signal.SIGHUP,  self._sig_hup_handler)
        signal.signal(signal.SIGTERM, self._sig_term_handler)
        signal.signal(signal.SIGINT,  self._sig_int_handler)

        while self.done is False:
            self._check_pools()
            #try:
            #    self._check_pools()
            #except Exception:
            #    # Don't wanna see any error messages on SIG{INT|TERM|KILL}.
            #    # So check at first if done is True.
            #    if self.done is False:
            #        self.log.exception()
            #        time.sleep(0.2)

        self._kill_pools()
        self.log.info("stop proc manager")

    def _check_pools(self):
        callback = self.callback
        while self.done is False:
            if self.restart_pools == True:
                self.restart_pools = None
                for pool in self.pools:
                    for pid in list(pool.children):
                        child = pool.children[pid]
                        child.quit = True

            for pool in self.pools:
                self._check_pool(pool)

            if callback: callback()
            # sleep only if done is False, so dieing as fast
            # as possilbe on sig{int|term}
            if self.done is False:
                time.sleep(1)

    def _check_pool(self, pool):
        self._spawn_worker(pool)
        self._quit_worker(pool)
        self._reap_worker(pool)

    def _spawn_worker(self, pool):
        num = 0
        for pid in pool.children:
            child = pool.children[pid]
            if child.quit == None:
                num += 1

        for i in range(pool.worker - num):
            if not len(pool.status_objects) or not len(pool.request_objects):
                self.log.warning("not enough value objects in pool")
                return
            self.log.notice("start worker [pool=%s]" % pool.name)
            # Locking is not necessary because the forked worker
            # is the only one who ever writes to the shared memory.
            status = pool.status_objects.pop()
            request = pool.request_objects.pop()
            status.value = 1
            p = Process(
                target=self._run_worker,
                args=(
                    pool.callback,
                    pool.signal_hup_child_callback,
                    pool.name,
                    status,
                    request,
                    pool.args,
                    pool.kwargs,
                ))
            p.start()
            pool.children[p.pid] = Child(p, status, request)

    def _quit_worker(self, pool):
        for pid in list(pool.children):
            child = pool.children[pid]
            if child.quit == True:
                os.kill(pid, 1)

    def _reap_worker(self, pool):
        for pid in list(pool.children):
            child = pool.children[pid]
            if not child.process.is_alive():
                child.status.value = 0
                pool.status_objects.append(child.status)
                pool.request_objects.append(child.request)
                child.process.join()
                del pool.children[pid]
                if child.process.exitcode > 0:
                    self.log.error(
                        "child pid %s died with exitcode %s [pool=%s]"
                        % (pid, child.process.exitcode, pool.name))

    def _kill_pools(self):
        children = {}

        for pool in self.pools:
            for pid in list(pool.children):
                child = pool.children[pid]
                if child.process.is_alive():
                    self.log.info("send sigterm to pid %s [pool=%s]" % (
                        pid, pool.name))
                    try:
                        os.kill(pid, 15)
                    except Exception:
                        pass
                    children[pid] = child
                    del pool.children[pid]

        timeout = time.time() + 15

        while len(children) > 0 and timeout > time.time():
            self.log.info("children still alive:", list(children))
            for pid in list(children):
                child = children[pid]
                if not child.process.is_alive():
                    child.process.join()
                    del children[pid]

            time.sleep(0.1)

        for pid in list(children):
            self.log.info("send sigkill to pid %s [pool=%s]" % (pid, pool.name))
            try:
                os.kill(pid, 9)
            except Exception:
                pass

    def _run_worker(
            self, callback, signal_hup_child_callback,
            pool_name, status, requests, args, kwargs):
        self.signal_hup_child_callback = signal_hup_child_callback
        self.pool_name = pool_name
        signal.signal(signal.SIGHUP, self._sig_hup_handler_child)
        if self.exit_on_sig_term:
            signal.signal(signal.SIGTERM, self._sig_term_handler_child)
            signal.signal(signal.SIGINT, self._sig_int_handler_child)
        self.status = status
        self.requests = requests
        try:
            callback(*args, **kwargs)
        except Exception:
            self.log.exception()
        sys.exit(0)

    def is_waiting(self):
        self.status.value = 2

    def is_reading(self):
        self.status.value = 3

    def is_processing(self):
        self.status.value = 4
        self.requests.value += 1

    def is_sending(self):
        self.status.value = 5

    def process_status(self):
        data = {}
        for pool in self.pools:
            data[pool.name] = {
                "requests": 0,
                "initiazing": 0,
                "waiting": 0,
                "reading": 0,
                "processing": 0,
                "sending": 0,
                "total": 0,
            }
            for o in pool.all_status_objects:
                if   o.value == 1: data[pool.name]["initiazing"] += 1
                elif o.value == 2: data[pool.name]["waiting"] += 1
                elif o.value == 3: data[pool.name]["reading"] += 1
                elif o.value == 4: data[pool.name]["processing"] += 1
                elif o.value == 5: data[pool.name]["sending"] += 1
                data[pool.name]["total"] += 1
            for o in pool.all_request_objects:
                data[pool.name]["requests"] += o.value
        return data

    def send_signal_hup_to_children(self):
       for pool in self.pools:
            for pid in list(pool.children):
                child = pool.children[pid]
                if child.process.is_alive():
                    try:
                        os.kill(pid, 1)
                    except Exception:
                        pass

    def _sig_term_handler(self, signum, frame):
        self.log.info("signal TERM received - quitting")
        self.done = True

    def _sig_int_handler(self, signum, frame):
        self.log.info("signal INT received - quitting")
        self.done = True

    def _sig_hup_handler(self, signum, frame):
        if self.signal_hup_callback:
            if self.signal_hup_callback() == True:
                if self.restart_pools_on_hup:
                    self.restart_pools = True
                elif self.send_hup_to_pools:
                    self.send_signal_hup_to_children()
        else:
            self.log.info("signal HUP received - ignoring")

    # Subprocesses immediate quit on TERM or INT
    def _sig_hup_handler_child(self, signum, frame):
        if self.restart_pools_on_hup:
            self.done = True
        elif self.signal_hup_child_callback:
            self.signal_hup_child_callback()
        else:
            self.log.info("signal HUP received - ignoring [pool=%s]" % self.pool_name)
    def _sig_int_handler_child(self, signum, frame):
        self.log.info("signal INT received - quitting [pool=%s]" % self.pool_name)
        sys.exit(0)
    def _sig_term_handler_child(self, signum, frame):
        self.log.info("signal TERM received - quitting [pool=%s]" % self.pool_name)
        sys.exit(0)

    def process_size(self):
        return Size()


class Pool:
    def __init__(
            self, name, worker, callback,
            signal_hup_child_callback, args, kwargs):
        self.name = name
        self.worker = worker
        self.callback = callback
        self.signal_hup_child_callback = signal_hup_child_callback
        self.args = args
        self.kwargs = kwargs
        self.children = {}
        self.status_objects = []
        self.request_objects = []
        for i in range(worker):
            self.status_objects.append(
                Value(c_byte, 0, lock=False))
            self.request_objects.append(
                Value(c_longlong, 0, lock=False))
        self.all_status_objects = list(self.status_objects)
        self.all_request_objects = list(self.request_objects)


class Child:
    def __init__(self, process, status, request):
        self.process = process
        self.status = status
        self.request = request
        self.quit = None


class Size:
    def __init__(self):
        self.size = 0
        self.resident = 0
        self.share = 0
        self.trs = 0
        self.lrs = 0
        self.drs = 0
        self.dtp = 0
        filename = "/proc/%s/statm" % os.getpid()
        try:
            with open(filename, encoding="utf-8") as f:
                line = f.readline()
                f.close()
                parts = list(map((lambda x: int(x) * 4096), line.split()))
                self.size = parts[0]
                self.resident = parts[1]
                self.shared = parts[2]
                self.text = parts[3]
                self.lib = parts[4]
                self.data = parts[5]
                self.dt = parts[6]
        except Exception:
            pass
