import os
import re
import platform
import select
import signal
import sys
import time
import traceback
import subprocess
from subprocess import Popen, PIPE
from bloonix.validator import validate
from bloonix.genattr import AddAttr
#from bloonix.logger import logger

class RunCmd:
    def __init__(self, **kwargs):
        AddAttr(self, self.validate(kwargs))
        self.has_max_bytes_exceeded = False
        self.has_timed_out = False
        self.has_os_error = False
        self.has_unknown_error = False
        self.has_exit_failure_code = False
        self.has_failed = False
        self.error = []
        self.exitcode = -1
        self.pid = None
        self.stdout = ""
        self.stderr = ""
        self.start_time = time.time()
        #self.log = logger("bloonix")

        try:
            self.run()
        except OSError as err:
            self.error.append("OS error: {0}".format(err))
            self.has_os_error = True
        except Exception as msg:
            self.error.append(
                "An unexpected error occurs: %s" % traceback.format_exc())
            self.has_unknown_error = True

        self.runtime = float(format(time.time() - self.start_time, ".3f"))

        if len(self.error) > 0:
            self.error = "; ".join(self.error)
            self.has_failed = True
        else:
            self.error = None

    def run(self):
        if platform.system() == "Windows":
            self.run_win()
        else:
            self.run_ix()

        if self.exitcode is None:
            self.error.append("command %s is still running" % self.command)
            self.has_unknown_error = True
        elif (self.ignore_exit_code == False
                and self.exitcode not in self.exit_success_codes):
            self.error.append(
                "command %s exited with %s" % (self.command, self.exitcode))

        if self.split_lines is True:
            self.stdout = re.split(r"\n", self.stdout)
            if len(self.stderr):
                self.stderr = re.split(r"\n", self.stderr)
            else:
                self.stderr = []

    def run_win(self):
        self.command = [str(e) for e in self.command]
        stdin = None
        if self.to_stdin is not None:
            stdin = self.to_stdin.encode('utf-8')
        try:
            p = subprocess.run(
                self.command,
                input=stdin,
                capture_output=True,
                timeout=self.timeout)
            self.stdout = p.stdout.decode("utf-8", "ignore")
            self.stderr = p.stderr.decode("utf-8", "ignore")
            self.exitcode = p.returncode
        except subprocess.TimeoutExpired:
            self.has_timed_out = True
            self.error.append("command timed out after %s seconds" % self.timeout)
        except Exception as err:
            self.error.append("an unexpected error occurs: %s" % err)

    def run_ix(self):
        self.command = [str(e) for e in self.command]
        p = Popen(
            self.command, bufsize=self.bufsize,
            stdin=PIPE, stdout=PIPE, stderr=PIPE,
            close_fds=True, shell=False)
        os.set_blocking(p.stdout.fileno(), os.O_NONBLOCK)
        os.set_blocking(p.stderr.fileno(), os.O_NONBLOCK)
        if self.to_stdin is not None:
            p.stdin.write(self.to_stdin.encode('utf-8'))
        p.stdin.close()
        self.read(p)
        self.pid = p.pid
        self.exitcode = self.poll(p)

    def read(self, p):
        """
        This method reads rotatory the output of stdout and stderr.
        Because of non-blocking IO a sleep of 0.05s is implemented
        for a CPU friendly polling. If the execution of the process
        runs on a timeout or if more than max_bytes are read from both
        stdout and stderr, then the process will be killed with the
        signal that is set to kill_signal.
        """

        # If both are True then both are finished and the while loop ends.
        stdout_fileno = p.stdout.fileno()
        stderr_fileno = p.stderr.fileno()
        stdout_done = None
        stderr_done = None
        bytes_read = 0

        # First if both stdout and stderr are finished the while loop ends.
        while not stdout_done or not stderr_done:
            # Max runtime of the command is set with self.timeout.
            # If the process runs on a timeout, the process is killed.
            if self.start_time + self.timeout < time.time():
                self.has_timed_out = True
                self.error.append(
                    "command timed out after %s seconds" % self.timeout)
                self.error.append("send kill %s" % self.kill_signal)
                self.kill(p.pid)
                break

            #self.log.debug("select", stdout_done, stderr_done)
            ready_fileno = select.select(
                [stdout_fileno, stderr_fileno], [], [], 1
            )[0]

            # Yay, we have data :-)
            #self.log.debug("ready fileno", ready_fileno)
            for fileno in ready_fileno:
                if fileno == stdout_fileno:
                    handle = p.stdout
                else:
                    handle = p.stderr

                # Read them, but only self.bufsize bytes.
                #self.log.debug("read size", self.bufsize)
                buf = handle.read(self.bufsize)
                #self.log.debug("read", buf)

                if len(buf):
                    #self.log.debug("read", len(buf), "bytes from", handle)
                    bytes_read += len(buf)

                    # Check if max_bytes is reached and kill the process
                    # because the process maybe is still running and is
                    # producing more and more output.
                    if self.max_bytes > 0 and bytes_read > self.max_bytes:
                        self.has_max_bytes_exceeded = True
                        self.error.append(
                            "command output exceeds max bytes %s"
                            % self.max_bytes)
                        self.error.append("send kill %s" % self.kill_signal)
                        self.kill(p.pid)
                        break

                    if fileno == stdout_fileno:
                        self.stdout += buf.decode("utf-8", "ignore")
                    else:
                        self.stderr += buf.decode("utf-8", "ignore")
                elif fileno == stdout_fileno:
                    #self.log.debug("stdout done")
                    stdout_done = True
                else:
                    #self.log.debug("stderr done")
                    stderr_done = True

        p.stdout.close()
        p.stderr.close()

    def kill(self, pid):
        """
        Try to kill the process if a pid is available.
        """
        if isinstance(pid, int):
            kill_signal = self.kill_signal

            if kill_signal == 9:
                kill_signal = signal.SIGKILL
            elif kill_signal == 15:
                kill_signal = signal.SIGTERM

            try:
                os.kill(pid, kill_signal)
            except Exception:
                pass

    def poll(self, p, wait=10, poll=0.5):
        while wait > 0:
            ret = p.poll()
            if ret is not None:
                return ret
            time.sleep(poll)
            wait -= poll

        return None

    def validate(self, config):
        config = validate(config, {
            "command": {
                "type": list,
                "allow_empty": False,
            },
            #"arguments": {
            #    "type": list,
            #    "allow_empty": True,
            #    "default": None
            #},
            "to_stdin": {
                "type": str,
                "default": None
            },
            "timeout": {
                "type": int,
                "default": 15
            },
            "kill_signal": {
                "type": int,
                "minvalue": 1,
                "default": 9
            },
            "max_bytes": {
                "type": int,
                "default": 10485760
            },
            "bufsize": {
                "type": int,
                "minvalue": 1,
                "default": 4096
            },
            "debug": {
                "type": bool,
                "default": False
            },
            "shell": {
                "type": bool,
                "default": False
            },
            "split_lines": {
                "type": bool,
                "default": False
            },
            "exit_success_codes": {
                "type": list,
                "default": [0]
            },
            "ignore_exit_code": {
                "type": bool,
                "default": False
            }
        })

        if config["max_bytes"] > 0 and config["max_bytes"] < config["bufsize"]:
            config["bufsize"] = config["max_bytes"]

        return config
