"""
Fork the process and run a function. The forked process will
be killed with signal 9 if the timeout is exceeded.

Call:
    ret = RunJob(timeout, callback, args, kwargs)

Arguments:
    timeout         - in secconds
    callback        - function or method
    args            - passed to the callback
    kwargs          - passed to the callback

Return value:
    ret.has_failed  - True or None. True if something wents wrong.
                      If the timeout is exceeded then ret.timed_out
                      is set to true. If the callback throws an
                      Exception, then the error message is set to
                      ret.error.
    ret.timed_out   - True or None
    ret.error       - The error message if the callback throws
    ret.exitcode    - The exit code of the Process
                      an exception.
    ret.data        - Return value of callback.

"""

from multiprocessing import Process, Queue
from queue import Empty
import os
import time

class RunJob:
    def __init__(self, timeout, callback, *args, **kwargs):
        self._timeout = timeout
        self._callback = callback
        self._args = args
        self._kwargs = kwargs
        self.timed_out = None
        self.data = None
        self.exitcode = None
        self.has_failed = None
        self.error = None
        self._run_parent()

    def _run_parent(self):
        self._queue = Queue(0)
        self._process = Process(target=self._run_child)
        self._process.start()
        self._watch_process()

    def _run_child(self):
        try:
            ret = self._callback(*self._args, **self._kwargs)
            self._queue.put((None, ret))
        except Exception as err:
            self._queue.put((err, None))

    def _watch_process(self):
        expire_time = time.time() + self._timeout
        
        while True:
            if not self._process.is_alive():
                self._read_data()
                self._reap_child()
                break

            if time.time() > expire_time:
                self.has_failed = True
                self.timed_out = True
                self._kill_process()
                break

            time.sleep(0.1)

    def _kill_process(self):
        try:
            os.kill(self._process.pid, 9)
        except Exception:
            pass
        while True:
            if not self._process.is_alive():
                self._reap_child()
                break
            time.sleep(0.1)

    def _read_data(self):
        try:
            self.error, self.data = self._queue.get(block=False)
            if self.error:
                self.has_failed = True
        except Empty:
            pass

    def _reap_child(self):
        self._process.join()
        self.exitcode = self._process.exitcode
