"""
The most important part of serialized.py is dealing with timeouts.
Whether it's establishing a connection, the SSL handshake, sending
or receiving data. A slow connection can cause the client or server
to hang in an unwanted way. While you can limit the runtime of
functions using signals on Unix systems, it's much more difficult on
Windows. For this reason, serialized.py splits the function calls
for Windows (_win) and Unix-like (_ix) systems.

Since there are no signals on Windows that can be used to control the
runtime of functions, socket.settimeout is used for Windows systems.
socket.settimeout is great because it can be used across platforms.
The big disadvantage, however, is that socket.settimeout can only be
used to a limited extent for SSL handshakes. You can set socket.settimeout
before a handshake, but the timeout is renewed for each traffic, which
can mean that extremely slow connections cannot be limited in time.
Therefore, signals via RunCode are used for Unix-like systems.

This module should therefore be used with caution on Windows systems
and only used where long timeouts are not a problem. This applies,
for example, to the Bloonix agent, where it is not a problem if the
connection to the server takes a little longer. The Bloonix server
runs on a Unix system and would abort the connection attempt if it
takes too long.
"""

import os
import platform
import re
import socket
import ssl
import struct
import sys
import time
import zlib
from importlib import import_module

from bloonix.exception import BloonixError, BloonixTimeout
from bloonix.logger import logger
from bloonix.validator import validate
from bloonix.json import json
from bloonix import utils

# 1. module name to import
# 2. compress function of module
# 3. decompress function of module
# 4. 1-Byte unsigned int, 0 means no compression
# zlib is default and always available
COMPRESSION = {
    "zlib": ["zlib", zlib.compress, zlib.decompress, 1],
    "gzip": ["gzip", "compress", "decompress", 2],
    "brotli": ["brotli", "compress", "decompress", 3],
    "lz4": ["lz4.frame", "compress", "decompress", 4],
}
COMPRESSION_MARKERS = {}

for method in list(COMPRESSION):
    module, compress, decompress, marker = COMPRESSION[method]
    if not isinstance(compress, str):
        COMPRESSION_MARKERS[marker] = method
        continue
    try:
        module = import_module(module)
        COMPRESSION[method][1] = getattr(module, compress)
        COMPRESSION[method][2] = getattr(module, decompress)
        COMPRESSION_MARKERS[marker] = method
    except Exception as err:
        del COMPRESSION[method]

if platform.system() != "Windows":
    from bloonix.runcode import RunCode

class __Base:
    def __init__(self):
        self.log = logger("bloonix")
        self.listen_sock = None
        self.sock = None
        self.addr = None
        self.port = None

    # Without self
    def get_available_compression_methods():
        return list(COMPRESSION)

    def send(self, data, timeout=None, max_bytes=None):
        if timeout is None:
            timeout = self.config["send_timeout"]

        if max_bytes is None:
            max_bytes = self.config["send_max_bytes"]

        self.log.info(
            "send data timeout %s max_bytes %s"
            % (timeout, max_bytes))

        if isinstance(data, dict):
            data = json.dumps(data)
        elif not isinstance(data, str):
            raise BloonixError("send: data must be of type dict/str")
        data = bytes(data, "utf-8")

        if max_bytes > 0 and len(data) > max_bytes:
            raise BloonixError(
                "send: the data length (%s bytes) exceeds max_bytes"
                % len(data))

        if self.compression is None:
            # Old style, plain JSON String,
            # newline masks the end of the data stream.
            self.log.debug("send: using old style plain json data exchange")
            data = data + b"\n"
        else:
            # New style with header data.
            data = self._check_send_compression(data)

        if platform.system() == "Windows":
            self._send_win(data, timeout)
        else:
            self._send_ix(data, timeout)

    def _check_send_compression(self, data):
        """
        Server > Client:
            Header
                4-Byte unsigned int, Length of the data
                1-Byte unsigned int, Compression marker
            Data

        Client > Server:
            Header
                4-Byte unsigned int, Length of the data
                1-Byte unsigned int, Compression marker
                1-Byte unsigned int, Available compression marker
            Data

        The first 4 bytes is used to pack(">I") the data length.
        The compression marker is used to set if the data are
        compressed or not. The available compression marker
        is only used by the client to tell the server which
        compression method to use in the response. The data is
        only compressed if the marker in not 0 and if the data
        has a length of at least 1000 Bytes. The listen_sock
        attribute checks whether it is a client or server connection.
        """
        length = len(data)
        marker = None
        available_marker = None

        if self.compression == "disabled":
            marker = 0
            available_marker = 0
            self.log.debug("send: compression disabled")
        elif length < 1000:
            marker = 0
            available_marker = COMPRESSION[self.compression][3]
            self.log.debug("send: compression disabled, data lower than 1000 bytes")
        else:
            marker = COMPRESSION[self.compression][3]
            available_marker = COMPRESSION[self.compression][3]
            self.log.debug("send: compress data with '%s'" % self.compression)
            data = COMPRESSION[self.compression][1](data)
            self.log.info(
                "send data compressed(%s) decompressed(%s) bytes"
                % (len(data), length))
            length = len(data)

        self.log.debug(
            "send: length = '%s', marker = '%s', available_marker = '%s'"
            % (length, marker, available_marker))

        if self.listen_sock is None:
            header = struct.pack(">IBB", length, marker, available_marker)
            #self.log.debug("send header", header)
            #self.log.debug("send data", data)
            return header + data

        header = struct.pack(">IB", length, marker)
        #self.log.debug("send header", header)
        #self.log.debug("send data", data)
        return header + data

    def _send_ix(self, data, timeout):
        if timeout:
            ret = RunCode(
                timeout,
                self._send_ix_sub,
                args=(data,))
            if ret.timed_out is True:
                raise BloonixTimeout(
                    "send runs on a timeout after %ss"
                    % timeout)
        else:
            self.sock.sendall(data)

    def _send_ix_sub(self, data):
        if self.config["test_type"] == "send":
            time.sleep(self.config["test_time"])
        self.sock.sendall(data)

    def _send_win(self, data, timeout):
        try:
            if self.config["test_type"] == "send":
                time.sleep(self.config["test_time"])
            self.sock.settimeout(timeout)
            self.sock.sendall(data)
        except TimeoutError as err:
            self.log.warning(
                "send runs on a timeout after %ss"
                % timeout)
            raise err
        finally:
            self.sock.settimeout(None)

    def recv(self, timeout=None, max_bytes=None):
        if timeout is None:
            timeout = self.config["recv_timeout"]

        if max_bytes is None:
            max_bytes = self.config["recv_max_bytes"]

        self.log.info(
            "recv data, settings: timeout %s, max_bytes %s"
            % (timeout, max_bytes))

        if platform.system() == "Windows":
            data = self._recv_win(timeout, max_bytes)
        else:
            data = self._recv_ix(timeout, max_bytes)

        try:
            self.log.notice("received", len(data), "bytes")
            data = data.decode("utf-8")
            data = json.loads(data)
        except Exception as err:
            self.log.warning(
                "unable to de-serialize data: %s, addr %s"
                % (err, self.addr))

        if not isinstance(data, dict):
            raise BloonixError("recv: data must be of type dict")

        return data

    def _recv_win(self, timeout, max_bytes):
        start_time = time.time()
        header = self._recv_win_bytes(5, timeout)
        length, marker = struct.unpack(">IB", header)

        if max_bytes > 0 and length > max_bytes:
            raise BloonixError(
                "recv: the data length (%s bytes) exceeds max_bytes"
                % length)

        if marker != 0 and marker not in COMPRESSION_MARKERS:
            raise BloonixError("recv: compression marker %s not supported" % marker)

        self.log.debug(
            "recv: length = '%s', marker = '%s'"
            % (length, marker))

        if length <= 0:
            raise BloonixError("recv: no data received")

        remaining_time = max(timeout - (time.time() - start_time), 0)
        if remaining_time == 0:
            raise BloonixTimeout("recv runs on a timeout after %ss" % timeout)

        data = self._recv_win_bytes(length, remaining_time)
        if not data:
            raise BloonixError("recv: no data received")

        if marker == 0:
            return data

        method = COMPRESSION_MARKERS[marker]
        self.log.debug("recv: decompress data with '%s'" % method)
        return COMPRESSION[method][2](data)

    def _recv_win_bytes(self, length, timeout):
        try:
            return self._recv_win_sub_bytes(length, timeout)
        finally:
            self.sock.settimeout(None)

    def _recv_win_sub_bytes(self, length, timeout):
        data = b""
        max_time = time.time() + timeout
        last_timeout = None

        while len(data) < length:
            if self.config["test_type"] == "recv":
                time.sleep(self.config["test_time"])

            t = max_time - time.time()
            if t <= 0:
                raise BloonixTimeout("recv runs on a timeout")

            if last_timeout is None or abs(t - last_timeout) > 0.1:
                self.sock.settimeout(t)
                last_timeout = t

            try:
                buf = self.sock.recv(length - len(data))
            except TimeoutError:
                raise BloonixTimeout(
                    "recv runs on a timeout after %ss"
                    % timeout)
            except Exception as err:
                raise BloonixError(
                    "recv failed: %s, addr %s"
                    % (err, self.addr))

            if len(buf) == 0:
                if len(data) == 0:
                    raise BloonixError(
                        "connection closed without receiving any data")
                break

            data += buf

        return data

    def _recv_ix(self, timeout, max_bytes):
        data = None
        if timeout:
            ret = RunCode(
                timeout,
                self._recv_ix_sub,
                args=(max_bytes,))
            if ret.timed_out is True:
                raise BloonixTimeout(
                    "recv runs on a timeout after %ss"
                    % timeout)
            return ret.data
        return self._recv_ix_sub(max_bytes)

    def _recv_ix_sub(self, max_bytes):
        first_byte = self.sock.recv(1)

        if not first_byte:
            raise BloonixError("recv: connection closed without receiving any data")

        # Old style, plain JSON.
        if first_byte == b"{":
            self.log.debug("recv: using old style plain json data exchange")
            data = first_byte
            while True:
                if self.config["test_type"] == "recv":
                    time.sleep(self.config["test_time"])
                buf = self.sock.recv(4096)
                if len(buf) == 0:
                    break
                data += buf
                if max_bytes > 0 and len(data) > max_bytes:
                    raise BloonixError(
                        "recv: the data length (%s bytes) exceeds max_bytes"
                        % len(data))
                if data[-1:] == b"\n":
                    break
            return data

        if self.listen_sock is None:
            header = first_byte + self._recv_ix_sub_bytes(4)
            length, marker = struct.unpack(">IB", header)
            available_marker = None
        else:
            header = first_byte + self._recv_ix_sub_bytes(5)
            length, marker, available_marker = struct.unpack(">IBB", header)

            # self.compression is set for send(), so the server knows
            # which compression method to use to send data to the client.
            if available_marker == 0:
                self.compression = "disabled"
            elif available_marker in self.allowed_compression_markers:
                self.compression = COMPRESSION_MARKERS[available_marker]
            else:
                self.log.notice(
                    "recv: compression available_marker", available_marker,
                    "not supported, fallback to zlib")
                self.compression = "zlib"

        if max_bytes > 0 and length > max_bytes:
            raise BloonixError(
                "recv: the data length (%s bytes) exceeds max_bytes"
                % len(length))

        if length <= 0:
            raise BloonixError("recv: no data received")

        if marker != 0:
            if self.listen_sock is None:
                allowed_markers = COMPRESSION_MARKERS
            else:
                allowed_markers = self.allowed_compression_markers
            if marker not in allowed_markers:
                raise BloonixError("compression marker %s not supported" % marker)

        self.log.debug(
            "recv: length = '%s', marker = '%s', available_marker = '%s'"
            % (length, marker, available_marker))

        data = self._recv_ix_sub_bytes(length)
        if marker == 0:
            self.log.debug("recv: data not compressed")
            return data

        method = COMPRESSION_MARKERS[marker]
        self.log.debug("recv: decompress data with '%s'" % method)
        data = COMPRESSION[method][2](data)
        self.log.info(
            "recv data compressed(%s) decompressed(%s) bytes"
            % (length, len(data)))
        return data

    def _recv_ix_sub_bytes(self, length):
        data = b""
        while len(data) < length:
            if self.config["test_type"] == "recv":
                time.sleep(self.config["test_time"])
            #self.log.debug("recv(%s)" % (length - len(data)))
            buf = self.sock.recv(length - len(data))
            if len(buf) == 0:
                raise BloonixError(
                    "recv: connection closed before receiving %s bytes"
                    % length)
            data += buf
        return data

    def disconnect(self):
        self.sock.close()
        self.sock = None


class Server(__Base):
    def __init__(self, **kwargs):
        super().__init__()
        self.config = validate_server(kwargs)
        self.allowed_compression_markers = []
        self.allowed_compression = []

        for method in self.config["compression"]:
            if method in COMPRESSION:
                self.log.notice("enable compression method '%s'" % method)
                marker = COMPRESSION[method][3]
                self.allowed_compression_markers.append(marker)
                self.allowed_compression.append(method)

        if self.config["use_ssl"] is True:
            self.log.debug("create server ssl context")
            self.log.debug("ssl_key_file=%s" % self.config["ssl_key_file"])
            self.log.debug("ssl_cert_file=%s" % self.config["ssl_cert_file"])

            if self.config["ssl_protocol"] is not None:
                self.context = ssl.SSLContext(
                    getattr(ssl, self.config["ssl_protocol"]))
            else:
                self.context = ssl.create_default_context(
                    ssl.Purpose.CLIENT_AUTH)

            self.context.load_cert_chain(
                keyfile=self.config["ssl_key_file"],
                certfile=self.config["ssl_cert_file"])

            if self.config["ssl_cipher_suite"]:
                self.context.set_ciphers(self.config["ssl_cipher_suite"])

            if self.config["ssl_dh_params_file"]:
                self.context.load_dh_params(self.config["ssl_dh_params_file"])

    def get_allowed_compression_methods(self):
        return list(self.allowed_compression)

    def listen(self):
        if re.search(":", self.config["bind"]):
            sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        else:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self.log.debug(
            "bind listen socket to",
            self.config["bind"],
            "port", self.config["port"])
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((self.config["bind"], self.config["port"]))
        if socket.SOMAXCONN < 4096:
            self.log.warning(
                "socket.SOMAXCONN=%s is maybe too low"
                % socket.SOMAXCONN)
        sock.listen(socket.SOMAXCONN)
        self.listen_sock = sock

    def accept(self, timeout=None):
        self.compression = None
        self.sock = None
        self.addr = None
        self.port = None

        try:
            self.listen_sock.settimeout(timeout)
            sock, addr = self.listen_sock.accept()
            self.listen_sock.settimeout(None)
        except TimeoutError:
            return
        except InterruptedError:
            return
        except Exception as err:
            self.log.warning("unable to accept connection: %s" % err)
            return

        self.addr, self.port = addr[0:2]
        if self.addr.startswith("::ffff:"):
            self.addr = self.addr.replace("::ffff:", "")
        self.log.notice("connect from", self.addr, self.port)

        if self.config["use_ssl"] is True:
            if platform.system() == "Windows":
                sock = self._wrap_context_win(sock)
            else:
                sock = self._wrap_context_ix(sock)
            if not sock:
                return
            cipher, version, bits = sock.cipher()
            self.log.info(
                "addr %s cipher %s ssl version %s bits %s"
                % (addr, cipher, version, bits))

        self.sock = sock
        return self

    def _wrap_context_win(self, sock):
        try:
            if self.config["test_type"] == "wrap_socket":
                time.sleep(self.config["test_time"])
            sock.settimeout(10)
            sock = self.context.wrap_socket(sock, server_side=True)
        except Exception as err:
            self.log.warning(
                "upgrading socket failed: %s, addr %s"
                % (err, addr))
            return None
        finally:
            sock.settimeout(None)
        return sock

    def _wrap_context_ix(self, sock):
        if self.config["test_type"] == "wrap_socket":
            time.sleep(self.config["test_time"])
        ret = RunCode(
            timeout=10,
            callback=self.context.wrap_socket,
            args=(sock,),
            kwargs={"server_side": True})
        if ret.timed_out is True:
            self.log.warning("ssl handshake timed out")
            return None
        return ret.data

    def stop_listen(self):
        self.listen_sock.close()
        self.listen_sock = None


class Client(__Base):
    def __init__(self, **kwargs):
        super().__init__()
        self.config = validate_client(kwargs)

        if self.config["compression"] is None or self.config["compression"] == "none":
            self.compression = None
            self.log.notice("please enable compression in configuration")
        elif self.config["compression"] in COMPRESSION:
            self.compression = self.config["compression"]
            self.log.notice("compression method %s enabled" % self.compression)
        else:
            self.compression = "zlib"
            self.log.warning("falling back to compression method zlib")

    def connect(self, timeout=None, retry=None):
        self.addr = None
        self.port = None

        if timeout == None:
            timeout = self.config["connect_timeout"]

        if retry == None:
            self._connect_timeout(timeout=timeout)
            return

        tries = -1
        failed = True
        while tries < retry:
            try:
                self._connect_timeout(timeout=timeout)
                failed = False
                break
            except Exception as err:
                tries = tries + 1
                time.sleep(1)

        if failed == True:
            raise Exception(
                "unable to connect to any host after %s tries"
                % tries)

    def _connect_timeout(self, timeout=None):
        if platform.system() == "Windows" or not timeout:
            self._connect(timeout=timeout)
        else:
            ret = RunCode(
                timeout=timeout,
                callback=self._connect,
                kwargs={"timeout": None})
            if ret.timed_out is True:
                raise BloonixTimeout(
                    "connect runs on a timeout after %ss"
                    % timeout)

    def _connect(self, timeout=None):
        if self.config["test_type"] == "connect":
            time.sleep(self.config["test_time"])

        if len(self.config["host"]) == 1:
            self._connect_to_host_port(
                self.config["host"][0],
                self.config["port"],
                timeout)
            return

        if self.config["mode"] == "balanced":
            self.config["host"].append( self.config["host"].pop(0) )

        arrlen = len(self.config["host"])
        for _ in range(arrlen):
            try:
                self._connect_to_host_port(
                    self.config["host"][0],
                    self.config["port"],
                    timeout)
                return
            except Exception as err:
                self.log.error(err)
                self.config["host"].append(self.config["host"].pop(0))

        raise Exception("unable to connect to any host")

    def _connect_to_host_port(self, host, port, timeout):
        self.addr = host
        self.port = port

        self.log.notice("connect to", host, "port", port)
        # The timeout is multiplied by the number of ip addresses
        # which are resolved by a host.
        sock = socket.create_connection(
            address=(host, port),
            timeout=timeout, # always None on unix
            source_address=self.config["bind"])
        self.log.info("connection established to", host, "port", port)

        if self.config["use_ssl"] is True:
            self.log.debug("create client ssl context")

            if self.config["ssl_protocol"] is not None:
                self.context = ssl.SSLContext(
                    getattr(ssl, self.config["ssl_protocol"]))
            else:
                self.context = ssl.create_default_context(
                    ssl.Purpose.SERVER_AUTH)

            if self.config["ssl_verify_mode"] == "peer":
                self.context.check_hostname = True
                self.context.verify_mode = ssl.CERT_REQUIRED
            else:
                self.context.check_hostname = False
                self.context.verify_mode = ssl.CERT_NONE

            if self.config["ssl_ca_bundle"] is not None:
                for path in self.config["ssl_ca_bundle"]:
                    if os.path.isfile(path):
                        self.context.load_verify_locations(cafile=path)
                    elif os.path.isdir(path):
                        self.context.load_verify_locations(capath=path)

            try:
                if platform.system() == "Windows":
                    sock.settimeout(10)
                if self.config["test_type"] == "wrap_socket":
                    time.sleep(self.config["test_time"])
                if self.config["ssl_server_hostname"] is not None:
                    sock = self.context.wrap_socket(
                        sock, server_hostname=self.config["ssl_server_hostname"])
                else:
                    sock = self.context.wrap_socket(sock, server_hostname=host)
            except Exception as err:
                self.log.warning(err)
                raise err
            finally:
                if platform.system() == "Windows":
                    sock.settimeout(None)

        self.sock = sock


def _base_parameter():
    return {
        "bind": {
            "type": str,
            "optional": True
        },
        "port": {
            "type": int
        },
        "use_ssl": {
            "type": bool,
            "default": False
        },
        "ssl_cipher_suite": {
            "type": str,
            "optional": True
        },
        "recv_max_bytes": {
            "type": str,
            "regex": r"^(?:unlimited|\d+(?:b{0,1}|[kmgt]b{0,1}))\Z",
            "default": "512k"
        },
        "send_max_bytes": {
            "type": str,
            "regex": r"^(?:unlimited|\d+(?:b{0,1}|[kmgt]b{0,1}))\Z",
            "default": "0"
        },
        "recv_timeout": {
            "type": int,
            "default": 15
        },
        "send_timeout": {
            "type": int,
            "default": 15
        },
        "timeout": {
            "type": int,
            "optional": True
        },
        "test_time": {
            "type": int,
            "default": 10,
        },
        "test_type": {
            "type": str,
            "default": "none"
        },
    }


def _client_parameter():
    return {
        "host": {
            "type": str
        },
        "ssl_ca_bundle": {
            "type": str,
            "default": "/etc/ssl/certs, /etc/pki/tls/certs/ca-bundle.crt"
        },
        "ssl_verify_mode": {
            "type": str,
            "regex": r"^(none|peer)\Z",
            "default": "peer"
        },
        "ssl_server_hostname": {
            "type": str,
            "optional": True
        },
        "ssl_protocol": {
            "type": str,
            "optional": True,
            "regex": r"^(?:PROTOCOL_\w+|TLSv\d(?:_|\.)\d)\Z",
            "default": "PROTOCOL_TLS_CLIENT"
        },
        "mode": {
            "type": str,
            "regex": r"^(failover|balanced)\Z",
            "default": "balanced"
        },
        "connect_timeout": {
            "type": int,
            "default": 10
        },
        "compression": {
            "type": str,
            "default": "zlib"
        },
    }

def _server_parameter():
    return {
        "ssl_cert_file": {
            "type": str,
            "optional": True
        },
        "ssl_key_file": {
            "type": str,
            "optional": True
        },
        "ssl_dh_params_file": {
            "type": str,
            "optional": True
        },
        "ssl_protocol": {
            "type": str,
            "regex": r"^(?:PROTOCOL_\w+|TLSv\d(?:_|\.)\d)\Z",
            "default": "PROTOCOL_TLS_SERVER",
        },
        "compression": {
            "type": str,
            "regex_cl": r"^(zlib|gzip|brotli|lz4)\Z",
            "default": "zlib,gzip",
        },
    }

def _validate(config, params):
    config = validate(config, params)

    if config["timeout"] is not None:
        config["recv_timeout"] = config["timeout"]
        config["send_timeout"] = config["timeout"]

    for key in ["recv_max_bytes", "send_max_bytes"]:
        value = str(config[key])

        if value == "unlimited":
            config[key] = 0
        else:
            config[key] = utils.string_to_bytes(value)

    return config

def validate_client(config):
    params = _base_parameter()
    params.update(_client_parameter())
    config = _validate(config, params)

    hosts = re.sub(r"\s", "", config["host"])
    config["host"] = re.split(",", hosts)

    if config["bind"] is not None:
        bind = re.sub(r"\s", "", config["bind"])
        config["bind"] = re.split(",", bind)

    config["ssl_ca_bundle"] = re.sub(r"\s", "", config["ssl_ca_bundle"])
    config["ssl_ca_bundle"] = re.split(",", config["ssl_ca_bundle"])
    config["ssl_protocol"] = validate_ssl_protocol(config["ssl_protocol"])

    return config

def validate_server(config):
    params = _base_parameter()
    params.update(_server_parameter())
    config = _validate(config, params)

    if config["bind"] is None or config["bind"] == "*":
        config["bind"] = "::"

    config["ssl_protocol"] = validate_ssl_protocol(config["ssl_protocol"])
    config["compression"] = re.sub(r"\s", "", config["compression"])
    config["compression"] = re.split(",", config["compression"])

    return config

def validate_ssl_protocol(proto):
    if re.search(r"^PROTOCOL_\w+\Z", proto):
        return proto
    if re.search(r"^TLSv\d\.\d\Z", proto):
        return "PROTOCOL_%s" % re.sub(r"\.", "_", proto)
    raise Exception("invalid protocol")
