import re
from bloonix.json import json

class SQL:
    _quotes = {
        "pg": '"',
        "mysql": '`',
        "mariadb": '`',
        "oracle": "'",
        "none": ""
    }

    def __init__(self, driver="none"):
        self._driver = driver
        self._qc = self._quotes[driver]

    def select(self, **kwargs):
        stmt = ["select"]
        bind = []
        close_bracket = None
        from_table = None

        if isinstance(kwargs["table"], list):
            from_table = kwargs["table"][0]
        elif isinstance(kwargs["table"], str):
            from_table = kwargs["table"]
            if "columns" in kwargs:
                kwargs["table"] = [kwargs["table"], kwargs["columns"]]
        else:
            raise Exception("no table set")

        if "count" in kwargs:
            if isinstance(kwargs["count"], str):
                if re.search(r"\.", kwargs["count"]):
                    add_columns = self.quote(kwargs["count"])
                else:
                    add_columns = self.quote(
                        "%s.%s" % (from_table, kwargs["count"]))
            else:
                cts = []
                for ct in kwargs["count"]:
                    if re.search(r"\.", ct):
                        cts.append(self.quote(ct))
                    else:
                        cts.append(self.quote("%s.%s" % (from_table, ct)))
                add_columns = ", ".join(cts)
            if "distinct" in kwargs and kwargs["distinct"] is True:
                stmt.append("count(distinct(%s)) AS count" % add_columns)
            elif "count" in kwargs:
                stmt.append("count(%s) AS count" % add_columns)
        elif "distinct" in kwargs and kwargs["distinct"] is True:
            stmt.append("distinct")

        if isinstance(kwargs["table"], list):
            if "count" in kwargs:
                stmt[-1] += ","
            add_columns = self._table(kwargs["table"], bind)
            stmt.append(add_columns)

        stmt.append("from %s" % self.quote(from_table))

        if "join" in kwargs:
            stmt.append(self._join(kwargs["join"]))
        if "condition" in kwargs:
            conds, binds = self._condition(kwargs["condition"])
            stmt.append(conds)
            bind += binds
        if "group" in kwargs:
            stmt.append("group by")
            stmt.append(self._group_columns(kwargs["group"]))
        if "order" in kwargs:
            stmt.append(self._order(kwargs["order"]))
        if "limit" in kwargs:
            stmt.append("limit %s")
            bind.append(kwargs["limit"])
        if "offset" in kwargs:
            stmt.append("offset %s")
            bind.append(kwargs["offset"])
        if "comment" in kwargs and isinstance(kwargs["comment"], str):
            stmt.append("--")
            stmt.append(kwargs["comment"])

        return " ".join(stmt), bind

    def _table(self, tables, bind):
        ret = []
        i = 0
        while i < len(tables):
            table = tables[i]
            i += 1
            columns = tables[i]
            i += 1

            if columns is None:
                continue

            if not isinstance(columns, list):
                columns = [columns]

            for column in columns:
                if column == 1 or column == "1":
                    ret.append("1")
                elif column == "*":
                    ret.append("%s.%s" % (self.quote(table), column))
                else:
                    ret.append(self._sel_column(table, column, bind))

        return ", ".join(ret)

    def _sel_column(self, table, column, bind):
        qc = self._qc

        # column = {
        #    "func": "coalesce",
        #    "value": {
        #        "func": "max",
        #        "column" "column",
        #    },
        #    "as": "name"
        # }
        if isinstance(column, dict):
            # coalesce( max(col), 1 ) as "status"
            bind.append(column["default"])
            if re.search(r"\.", column["value"]["column"]):
                col = self.quote(column["value"]["column"])
            else:
                col = self.quote("%s.%s" % (table, column["value"]["column"]))
            if "as" in column:
                return "%s(%s(%s),%%s) as %s" %(
                    column["func"],
                    column["value"]["func"],
                    col,
                    self.quote(column["as"]))
            else:
                return "%s(%s(%s),%%s)" %(
                    column["func"],
                    column["value"]["func"],
                    col)

        m = re.search(r"^(min|max|sum|avg|count|distinct)\(([a-zA-Z_0-9]+)\)(.*)\Z", column)
        if m:
            func = m.group(1)
            col = self.quote("%s.%s" % (table, m.group(2)))
            rest = m.group(3)
            m = re.search(r"\s+(?:AS|as)\s+([a-zA-Z_0-9]+)", rest)
            if m:
                return "%s(%s) as %s" % (func, col, self.quote(m.group(1)))
            return "%s(%s)" % (func, col)

        if re.search(" (?:AS|as) ", column):
            a, b = re.split(" (?:AS|as) ", column)
            if re.search(r"\.", column):
                a = self.quote(a)
            else:
                a = self.quote(table +"."+ a)
            b = self.quote(b)
            return "%s as %s" % (a, b)

        return self.quote(table +"."+ column)

    def _group_columns(self, columns):
        """
        sql.select(columns=["a", "b", "c"])
        sql.select(columns="a")
        """
        if isinstance(columns, str):
            return self.quote(columns)

        cols = []
        for col in columns:
            if col == "*":
                cols.append(col)
            else:
                cols.append(self.quote(col))
        return ", ".join(cols)

    def _join(self, joins):
        ret = []

        i = 0
        while i < len(joins):
            join_type = joins[i]
            i += 1
            join_def = joins[i]
            i += 1

            if len(join_def) == 4:
                table, left, op, right = join_def
            elif len(join_def) == 3:
                table, left, right = join_def
                op = "="

            ret.append(
                "%s join %s on %s %s %s"
                % (
                    join_type,
                    self.quote(table),
                    self.quote(left),
                    op,
                    self.quote(right)
                ))

        return " ".join(ret)

    def _condition(self, conditions):
        conds = []
        binds = []

        if (len(conditions) > 0
                and (conditions[0] == "and" or conditions[0] == "or")):
            conditions[0] = "where"

        next_sub_select = False

        for cond in conditions:
            if next_sub_select:
                next_sub_select = False
                s, b = self.select(**cond)
                value = "(%s)" % s
                binds = binds + b
                conds.append(value)
                continue

            if isinstance(cond, str):
                if re.search(r"[()]", cond):
                    conds.append(cond)
                elif re.search(r"(where|and|or)\s+(not\s+){0,1}exists", cond):
                    conds.append(cond)
                    next_sub_select = True
                else:
                    # where, and, or
                    conds.append(cond)
                continue

            # cond tuple: (col, op, col)
            if isinstance(cond, tuple):
                if len(cond) == 3:
                    col1, op, col2 = cond
                elif len(cond) == 2:
                    col1, col2 = cond
                    op = "="
                conds.append(
                    "%s %s %s"
                    % (self.quote(col1), op, self.quote(col2)))
                continue

            # SPECIAL
            # where, {
            #   "math": ["column1", "*", 1000, "-", 10, ">"]
            # }
            if isinstance(cond, dict):
                if "math" not in cond:
                    raise Exception(cond)
                for item in cond["math"]:
                    if isinstance(item, str) and item in ["+", "-", "*", "/", ">", ">=", "<", "<=", "="]:
                        conds.append(item)
                    elif isinstance(item, (int, float)):
                        conds.append("%s")
                        binds.append(item)
                    elif isinstance(item, str):
                        conds.append(self.quote(item))
                continue

            if not isinstance(cond, list) or len(cond) > 3 or len(cond) < 2:
                raise Exception(cond)

            # cond list: [col, op, value]
            if len(cond) == 3:
                column, op, value = cond
            elif len(cond) == 2:
                column, value = cond
                if isinstance(value, list) or isinstance(value, dict):
                    op = "in"
                elif value is None:
                    op = "is"
                    value = None
                else:
                    op = "="

            if isinstance(column, dict):
                column = self._cond_column(column)
            else:
                column = self.quote(column)

            if isinstance(value, list):
                binds += value
                value = "(%s)" % ", ".join(["%s"] * len(value))
                conds.append("%s %s %s" % (column, op, value))
            elif isinstance(value, dict):
                s, b = self.select(**value)
                value = "(%s)" % s
                binds = binds + b
                conds.append("%s %s %s" % (column, op, value))
            elif (op == "is" or op == "is not"):
                if value is None or re.search(r"^null\Z", value, re.IGNORECASE):
                    conds.append("%s %s null" % (column, op))
                else:
                    binds.append(value)
                    conds.append("%s %s %s" % (column, op, "%s"))
            elif (op == "ilike" and self._driver in ["mysql", "mariadb"]):
                binds.append(value)
                conds.append("lower(%s) like lower(%s)" % (column, "%s"))
            else:
                binds.append(value)
                conds.append("%s %s %s" % (column, op, "%s"))

        return " ".join(conds), binds

        if isinstance(columns, list):
            for column in columns:
                self.order(order_type, column)
        else:
            self.order_by.append("%s %s" % (self.quote(columns), order_type))

    def _cond_column(self, column):
        if "concat" in column:
            delimiter = column.get("delimiter", " ")
            lower = column.get("lower", True)
            if lower:
                string = "lower(concat_ws('%s', %s))"
            else:
                string = "concat_ws(%s, %s)"
            return string % (delimiter, ",".join(column["concat"]))
        else:
            raise Exception("invalid column %s" % column)

    def _order(self, order):
        order_by = []

        i = 0
        while i < len(order):
            order_type = order[i]
            i += 1
            columns = order[i]
            i += 1

            if order_type != "asc" and order_type != "desc":
                raise Exception("wrong order type %s" % order_type)

            if isinstance(columns, list):
                for column in columns:
                    order_by.append("%s %s" % (self.quote(column), order_type))
            else:
                order_by.append("%s %s" % (self.quote(columns), order_type))

        return "order by %s" % ", ".join(order_by)

    def quote(self, string):
        qc = self._qc

        if string == "*":
            return string

        if not re.search(r"^[a-zA-Z_0-9]+(?:\.[a-zA-Z_0-9]+)?\Z", string):
            raise Exception("invalid characters found in string:", string)

        # table.column
        if re.search(r"\.", string):
            a, b = re.split(r"\.", string)
            return "%s%s%s.%s%s%s" % (qc, a, qc, qc, b, qc)

        # column
        return "%s%s%s" % (qc, string, qc)

    def insert(self, table, data):
        cols = []
        placeholders = []
        stmt = None
        bind = []

        if isinstance(data, list):
            bulk_placeholders = []
            for column in data[0]:
                cols.append(column)
                placeholders.append("%s")
            for _data in data:
                bulk_placeholders.append("(%s)" % ",".join(placeholders))
                for column in cols:
                    if isinstance(_data[column], (dict, list)):
                        bind.append(json.dumps(_data[column]))
                    else:
                        bind.append(_data[column])

            stmt = "insert into %s (%s) values %s" % (
                self.quote(table),
                ", ".join(cols),
                ", ".join(bulk_placeholders)
            )
        else:
            for column in data:
                cols.append(self.quote(column))
                placeholders.append("%s")
                if isinstance(data[column], (dict, list)):
                    bind.append(json.dumps(data[column]))
                else:
                    bind.append(data[column])

            stmt = "insert into %s (%s) values (%s)" % (
                self.quote(table),
                ", ".join(cols),
                ", ".join(placeholders)
            )

        return stmt, bind

    def update(self, table, data=None, condition=None, sdata=None):
        if data is None and sdata is None:
            raise Exception("missing data for update")

        stmt, bind = self._update(table, data, sdata)

        if condition:
            cond, cond_bind = self._condition(condition)
            stmt += " " + cond
            bind += cond_bind

        return stmt, bind

    def update_by_id(self, table, value, data):
        return self.update_by_col(table, "id", value, data)

    def update_by_col(self, table, column, value, data):
        stmt, bind = self._update(table, data, None)
        cond, cond_bind = self._condition(["where", [column, "=", value]])
        stmt += " " + cond
        bind += cond_bind
        return stmt, bind

    def _update(self, table, data, sdata):
        cols = []
        bind = []

        if data:
            for column in data:
                cols.append(self.quote(column) + " = %s")
                if isinstance(data[column], (dict, list)):
                    bind.append(json.dumps(data[column]))
                else:
                    bind.append(data[column])

        if sdata:
            for col in sdata:
                sub_stmt, sub_binds = self.select(**sdata[col])
                cols.append("%s = (%s)" % (self.quote(col), sub_stmt))
                bind += sub_binds

        stmt = "update %s set %s" % (
            self.quote(table),
            ", ".join(cols)
        )

        return stmt, bind

    def delete_by_id(self, table, value):
        return self.delete_by_col(table, "id", value)

    def delete_by_col(self, table, column, value, op="="):
        return self.delete(table, ["where", [column, op, value]])

    def delete(self, table, condition=None):
        stmt = "delete from %s" % self.quote(table)
        bind = []

        if condition:
            cond, bind = self._condition(condition)
            stmt += " " + cond

        return stmt, bind

    def get(self, table, value, column="id", columns="*"):
        """
        *get* can be used to select a rows by id.

        Example:
            get("table", "value1")

        SQL output:
            select * from table where id = 'value1'
        """
        return self.select(
            table=table,
            columns=columns,
            condition=["where", [column, value]]
        )

    def find(self, table, *args, columns="*", condition=None, combine="and"):
        """
        *find* can be used to select rows by multiple where clauses

        Example:
            find(
                "table",
                "column1", "value1",
                "column2", "value2",
            )

        SQL output:
            select * from table where column1 = 'value1' and column2 = 'value2'

        For complex conditions:
            find(
                "table",
                condition=[
                    "where", ["column1", "value1"],
                    "and", ["column2", "!=", "value2"]
                ]
            )

        SQL output:
            select * from table where column1 = 'value1' and column2 != 'value2'
        """
        if combine != "and" and combine != "or":
            raise Exception("invalid combine for where condition")

        if condition is None:
            cond = []
            args = list(args)
            while len(args):
                column = args.pop(0)
                value = args.pop(0)
                if len(cond):
                    cond.append(combine)
                else:
                    cond.append("where")
                cond.append([column, value])
            return self.select(
                table=table,
                columns=columns,
                condition=cond
            )

        return self.select(
            table=table,
            columns=columns,
            condition=condition
        )
