from bloonix import rx
from bloonix import utils
import re
import time

rx_time = r"%s:%s" % (rx.HOUR_BASE, rx.SECOND_BASE)
rx_month_with_day = r"%s\s+%s" % (rx.MONTH_NAME_BASE, rx.DAY_BASE)
rx_date = r"%s-%s-%s" % (rx.YEAR_BASE, rx.MONTH_BASE, rx.DAY_BASE)
rx_time_to_time = r"%s\s*-\s*%s" % (rx_time, rx_time)
rx_time_list = r"%s(?:\s*,\s*%s)*" % (rx_time_to_time, rx_time_to_time)

name2num = {
    "jan": 1,  "feb": 2,  "mar": 3,
    "apr": 4,  "may": 5,  "jun": 6,
    "jul": 7,  "aug": 8,  "sep": 9,
    "oct": 10, "nov": 11, "dec": 12,
    "mon": 0,  "tue": 1,  "wed": 2,
    "thu": 3,  "fri": 4,  "sat": 5,
    "sun": 6
}

# Monday            09:00 - 23:00
# Mon               09:00 - 23:00
# Monday - Friday   09:00 - 23:00
# Mon - Fri         09:00 - 23:00
f_weekday = r"^(-\s*)?(%s(?:\s*-\s*%s){0,1})\s+(%s)\Z" % (
    rx.WEEKDAY_BASE, rx.WEEKDAY_BASE, rx_time_list)

# January           09:00 - 23:00
# January - July    09:00 - 23:00
# Jan - Jul         09:00 - 23:00
# Jan               09:00 - 23:00
f_month = r"^(-\s*)?(%s(?:\s*-\s*%s){0,1})\s+(%s)\Z" % (
    rx.MONTH_NAME_BASE, rx.MONTH_NAME_BASE, rx_time_list)

# January 01            09:00 - 23:00
# Jan                   09:00 - 23:00
# January 01 - July 01  09:00 - 23:00
# Jan 01 - Jul 01       09:00 - 23:00
f_month_with_day = r"^(-\s*)?(%s(?:\s*-\s*%s){0,1})\s+(%s)\Z" % (
    rx_month_with_day, rx_month_with_day, rx_time_list)

# 2016-01               09:00 - 23:00
# 2016-01               09:00 - 23:00
# 2016-01 - 2016-01     09:00 - 23:00
# 2016-01 - 2016-01     09:00 - 23:00
f_year_month = r"^(-\s*)?(%s-%s(?:\s*-\s*%s-%s){0,1})\s+(%s)\Z" % (
    rx.YEAR_BASE, rx.MONTH_BASE, rx.YEAR_BASE, rx.MONTH_BASE, rx_time_list)

# 2016-04-05                09:00 - 23:00
# 2016-04-05 - 2015-01-02   09:00 - 23:00
f_date = r"^(-\s*)?(%s(?:\s*-\s*%s){0,1})\s+(%s)\Z" % (
    rx_date, rx_date, rx_time_list)

formats = {
    "weekday": f_weekday,
    "month": f_month,
    "month_with_day": f_month_with_day, 
    "year_month": f_year_month,
    "date": f_date
}

# validate a timeslice or list of timeslices
def validate(period, raise_error=None):
    if not isinstance(period, list):
        period = [period]

    for p in period:
        match = _parse(p)

        if not match:
            if raise_error:
                raise BloonixError("invalid timeslice '%s'" % p)
            return False

    return True

def check(period, unixtime=None, tz=None):
    if unixtime is None:
        unixtime = time.time()

    dt = utils.datetime(unixtime, tz)

    if not isinstance(period, list):
        period = [period]

    hit = None

    for p in period:
        match = _parse(p)

        if not _check_time(dt, match["time"]):
            continue
        if not match:
            raise Exception("invalid timeperiod format '%s'" % p)

        if match["alias"] == "month":
            if _check_month(dt, match["date"]):
                hit = True
        elif match["alias"] == "month_with_day":
            if _check_month_with_day(dt, match["date"]):
                hit = True
        elif match["alias"] == "weekday":
            if _check_weekday(dt, match["date"]):
                hit = True
        elif match["alias"] == "year_month":
            if _check_year_month(dt, match["date"]):
                hit = True
        elif match["alias"] == "date":
            if _check_date(dt, match["date"]):
                hit = True

        if match["exclude"] and hit:
            return False

    return hit

def _parse(string):
    date = None
    time = None
    alias = None

    for _alias in formats:
        if not re.search(r"%s\Z" % rx_time_to_time, string):
            string += " 00:00 - 23:59"

        m = re.search(formats[_alias], string)
        if m:
            exclude = m.group(1)
            date = m.group(2)
            time = m.group(3)
            alias = _alias
            break

    if alias is None:
        return None

    return {"alias": alias, "date": date, "time": time, "exclude": exclude}

def _check_time(dt, time):
    parts = re.split(",", time)
    curtime = dt.tm_hour * 60 + dt.tm_min

    for p in parts:
        m = re.search(r"^\s*(\d{2}):(\d{2})\s*-\s*(\d{2}):(\d{2})\s*\Z", p)
        if m:
            fhour = int(m.group(1))
            fmin = int(m.group(2))
            thour = int(m.group(3))
            tmin = int(m.group(4))
            ftime = fhour * 60 + fmin
            ttime = thour * 60 + tmin

            if curtime >= ftime and curtime <= ttime:
                return True

    return False

def _check_month(dt, date):
    m = re.search(r"^\s*([^\s]+)\s*-\s*([^\s]+)\s*\Z", date)
    if m:
        fmon = m.group(1)[0:3].lower()
        tmon = m.group(2)[0:3].lower()
    else:
        m = re.search(r"^\s*([^\s]+)\s*\Z", date)
        if m:
            fmon = tmon = m.group(1)[0:3].lower()

    fmon = name2num[fmon]
    tmon = name2num[tmon]

    if dt.tm_mon >= fmon and dt.tm_mon <= tmon:
        return True

    return False

def _check_month_with_day(dt, date):
    m = re.search(r"^\s*([^\s]+)\s+(\d+)\s*-\s*([^\s]+)\s+(\d+)\s*\Z", date)
    if m:
        fmon = m.group(1)[0:3].lower()
        fday = int(m.group(2))
        tmon = m.group(3)[0:3].lower()
        tday = int(m.group(4))
    else:
        m = re.search(r"^\s*([^\s]+)\s+(\d+)\s*\Z", date)
        if m:
            fmon = tmon = m.group(1)[0:3].lower()
            fday = tday = int(m.group(2))

    fmon = int(name2num[fmon])
    tmon = int(name2num[tmon])

    ftime = int("1%02d%02d" % (fmon, fday))
    ttime = int("1%02d%02d" % (tmon, tday))
    curtime = int("1%02d%02d" % (dt.tm_mon, dt.tm_mday))

    if curtime >= ftime and curtime <= ttime:
        return True

    return False

def _check_weekday(dt, date):
    m = re.search(r"^\s*([^\s]+)\s*-\s*([^\s]+)\s*\Z", date)
    if m:
        fwday = m.group(1)[0:3].lower()
        twday = m.group(2)[0:3].lower()
    else:
        m = re.search(r"^\s*([^\s]+)\s*\Z", date)
        if m:
            fwday = twday = m.group(1)[0:3].lower()

    fwday = name2num[fwday]
    twday = name2num[twday]

    #print("if", dt.tm_wday, ">=", fwday, "and", dt.tm_wday, "<=", twday)
    if dt.tm_wday >= fwday and dt.tm_wday <= twday:
        return True

    return False

def _check_year_month(dt, date):
    m = re.search(r"^\s*(-{0,1}\d+)-(\d{2})\s*-\s*(-{0,1}\d+)-(\d{2})\s*\Z", date)
    if m:
        fyear = m.group(1)
        fmon = m.group(2)
        tyear = m.group(3)
        tmon = m.group(4)
    else:
        m = re.search(r"^\s*(-{0,1}\d+)-(\d{2})\s*\Z", date)
        if m:
            fyear = tyear = m.group(1)
            fmon = tmon = m.group(2)

    ftime = int(fyear + fmon)
    ttime = int(tyear + tmon)
    curtime = int(time.strftime("%Y%m", dt))

    #print("if", curtime, ">=", ftime, "and", curtime, "<=", ttime)
    if curtime >= ftime and curtime <= ttime:
        return True

    return False

def _check_date(dt, date):
    drx = r"(-{0,1}\d+)-(\d{2})-(\d{2})"
    from_to_regex = r"^\s*%s\s*-\s*%s\s*\Z" % (drx, drx)
    date_regex = r"^\s*%s\s*\Z" % drx

    m = re.search(from_to_regex, date)
    if m:
        fyear = m.group(1)
        fmon = m.group(2)
        fday = m.group(3)
        tyear = m.group(4)
        tmon = m.group(5)
        tday = m.group(6)
    else:
        m = re.search(date_regex, date)
        if m:
            fyear = tyear = m.group(1)
            fmon = tmon = m.group(2)
            fday = tday = m.group(3)
        else:
            raise Exception("invalid timeperiod date stamp '%s'" % date)

    ftime = int(fyear + fmon + fday)
    ttime = int(tyear + tmon + tday)
    curtime = int(time.strftime("%Y%m%d", dt))

    if curtime >= ftime and curtime <= ttime:
        return True

    return False
