"""
Utils module with simple functions which can be re-used everywhere.
"""
import base64
import hashlib
import os
import random
import re
import signal
import string
import time
import smtplib
import email

from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

gsm_7bit_chars = (
    "@£$¥èéùìòÇ\nØø\rÅåΔ_ΦΓΛΩΠΨΣΘΞ\x1BÆæßÉ !\"#¤%&'()*+,-./0123456789:;<=>?"
    "¡ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÑÜ§¿abcdefghijklmnopqrstuvwxyzäöñüà"
)

def string_to_bytes(s):
    """ Convert a string like 1k to 1024. """
    _bytes = {
        "b": 1,
        "k": 1024,
        "m": 1048576,
        "g": 1073741824,
        "t": 1099511627776
    }
    m = re.search(r"^(\d+)([bkmgt]{0,1})\Z", s)
    if m:
        num = m.group(1)
        byt = m.group(2)
        if len(byt) == 0:
            byt = "b"
        return float(num) * _bytes[byt]
    return 0

def quickmail(sender, to, subject, message, cc=None, bcc=None):
    """
    Simple mail without attachments.
    """
    if isinstance(to, list):
        to = ", ".join(to)

    if isinstance(message, list):
        message = "\r\n".join(message)

    msg = MIMEText(message)
    msg["Subject"] = subject
    msg["From"] = sender
    msg["To"] = to

    if cc is not None:
        msg["Cc"] = cc
    if bcc is not None:
        msg["Bcc"] = bcc

    s = smtplib.SMTP("localhost")
    s.send_message(msg)
    s.quit()

def sendmail(
        sender, to, subject, message=None,
        cc=None, bcc=None, attachments=None):
    """
    E-Mail with attachments.
    attachments = [{
        "filetype": filetype,
        "filename": filename,
        "data": data
    }]
    """
    if isinstance(to, list):
        to = ", ".join(to)

    mail = MIMEMultipart()
    mail["Subject"] = subject
    mail["From"] = sender
    mail["To"] = to

    if cc is not None:
        mail["Cc"] = cc
    if bcc is not None:
        mail["Bcc"] = bcc

    if isinstance(message, list):
        message = "\r\n".join(message)

    if message is not None:
        msg = MIMEText(message)
        mail.attach(msg)

    if isinstance(attachments, list):
        for att in attachments:
            base_filename = re.sub(r".*/", "", att["filename"])
            maintype, subtype = att["filetype"].split("/")
            msg = MIMEBase(maintype, subtype)

            if "data" in att and att["data"] is not None:
                # remove the data url prefix if exist
                data = re.sub(r"^data:.+?/.+?;base\d*,", "", att["data"])
                # add newlines each 76 characters if not exist
                data = base64.b64decode(data)
                msg.set_payload(data)
            else:
                f = open(att["filename"], "rb")
                msg.set_payload(f.read())
                f.close()

            email.encoders.encode_base64(msg)
            msg.add_header(
                "Content-Disposition",
                "attachment",
                filename=base_filename)
            mail.attach(msg)

    s = smtplib.SMTP("localhost")
    s.send_message(mail)
    s.quit()

def timestamp(
        unixtime=None, timezone=None, dateonly=None, timeonly=None,
        date_as_number=None, time_format="%Y-%m-%d %H:%M:%S"):

    """ Generate a timestamp in format 2000-01-01 00:00:00 """
    if timezone is not None:
        _tz = os.environ.get("TZ", None)
        os.environ["TZ"] = timezone

    if date_as_number is True:
        fmt = "%Y%m%d"
    elif dateonly is True:
        fmt = "%Y-%m-%d"
    elif timeonly is True:
        fmt = "%H:%M:%S"
    else:
        fmt = time_format

    if unixtime is not None:
        timestamp = time.strftime(fmt, time.localtime(unixtime))
    else:
        timestamp = time.strftime(fmt)

    if timezone is not None:
        if _tz is None:
            del os.environ["TZ"]
        else:
            os.environ["TZ"] = _tz

    return timestamp

def ymd_time(unixtime=None):
    """ Return year, month, day as tuple """
    t = None

    if unixtime is not None:
        t = time.localtime(unixtime)
    else:
        t = time.localtime()

    return t.tm_year, t.tm_mon, t.tm_mday

try:
    # Python >= 3.6
    import secrets
    def pwgen(length=30, special_signs=False):
        """ Create a random string """
        char = string.ascii_letters + string.digits
        if special_signs:
            char += "!\"#$%&()*+,-./:;<=>?@[]^_`{|}~'"
        return "".join(secrets.choice(char) for _ in range(length))
except Exception:
    # Python < 3.6
    def pwgen(length=30, special_signs=False):
        """ Create a random string """
        char = list(string.ascii_letters)
        char += list(string.digits)
        if special_signs == True:
            char += list('!"#$%&()*+,-./:;<=>?@[]^_`{|}~')
            char.append("'")
        charlen = len(char)
        pwd = ""
        for _ in range(length):
            pwd += char[random.randrange(charlen)]
        return pwd

def sha256(s):
    """ Returns hashlib.sha256().hexdigest() """
    return hashlib.sha256(s.encode("utf-8")).hexdigest()

def sha512(s):
    """ Returns hashlib.sha512().hexdigest() """
    return hashlib.sha512(s.encode("utf-8")).hexdigest()

def pbkdf2_sha512_base64(password, salt=None, rounds=None):
    """
    Creates a PBKDF2 hash and returns
        hashed password, salt, rounds
    """
    if salt is None:
        salt = pwgen(64)
    if rounds is None:
        rounds = 100000
    dk = hashlib.pbkdf2_hmac(
        "sha512",
        bytes(password, "utf-8"),
        bytes(salt, "utf-8"),
        rounds)
    pwd = base64.standard_b64encode(dk).decode("utf-8")
    return pwd, salt, rounds

def validate_pv_dict(data):
    """
    Validation of simple dicts parsed from an INI file, for example.
    Keys may only consist of 'a-zA-Z_0-9.' and values may only have
    strings, integer or float set.
    """
    if not isinstance(data, dict):
        return False
    for key in list(data):
        if (not isinstance(key, str)
                or not isinstance(data[key], (str, int, float))):
            return False
        if re.search(r"^\s*\Z", key) and re.search(r"^\s*\Z", data[key]):
            del data[key]
            continue
        if (not re.search(r"^[a-zA-Z_0-9\.]+\Z", key) or
                not re.search(r"[^\s]", str(data[key]))):
            return False
    return True

def unixtime(timestring, tz=None):
    """
    Convert a timestring in format '2000-01-01 00:00:00' to unixtime.
    """
    tz = set_tz(tz)
    try:
        if re.search(r"^\d\d\d\d-\d\d-\d\d\Z", timestring):
            timestring += " 00:00:00"
        t = time.strptime(timestring, "%Y-%m-%d %H:%M:%S")
        unixtime = time.mktime(t)
    finally:
        set_tz(tz)
    return int(unixtime)

def datetime(unixtime=None, tz=None):
    """ Returns struct from time.localtime() """
    if unixtime is None:
        unixtime = time.time()
    tz = set_tz(tz)
    try:
        dt = time.localtime(unixtime)
    finally:
        set_tz(tz)
    return dt

def set_tz(tz=None):
    _tz = os.environ.get("TZ", None)

    if tz is not None:
        os.environ["TZ"] = tz
    elif tz is None and "TZ" in os.environ:
        del os.environ["TZ"]

    time.tzset()
    return _tz

def mhd_to_seconds(s):
    """
    Convert a string to seconds and returns
    the value as int(). Examples:
        1d = 86400
        1h = 3600
        1m = 60
        1 = 1
    """
    m = re.search(r"^(\d+)([mhd])\Z", s)
    if m:
        t = float(m.group(1))
        u = m.group(2)
        if u == "m":
            return int(t * 60)
        if u == "h":
            return int(t * 3600)
        if u == "d":
            return int(t * 86400)
    return int(s)

def get_last_day_of_month(year, month):
    if not isinstance(year, int) or not isinstance(month, int):
        raise Exception("get_last_day_of_month() expecting integer values")
    if month in [4,6,9,11]:
        return 30
    if month == 2:
        if (year % 400 == 0
                or (year % 100 > 0 and year % 4 == 0)):
            return 29
        return 28
    return 31

def euro(s, lang="de"):
    """
    Replaces dot and returns currency in euro. Examples:
        2 = 2,00
        2.0 = 2,00
        2.255 = 2,26
    """
    if lang == "de":
        string = '{0},{1}'
    else:
        string = '{0}.{1}'
    return string.format(*('{:.2f}'.format(s).split(".")))

def version(a):
    """
    Convert a version number to a number which can be compared
    to other version numbers.
    """
    m = re.search(r"(.+)-(.+)", a)
    if m:
        vl, b = m.group(1), m.group(2)
        vl = re.split(r"\.", vl)
        b = b.zfill(8)
    else:
        vl = re.split(r"\.", a)
        b = "00000000"
    for i,_ in enumerate(vl):
        vl[i] = vl[i].zfill(8)
    while len(vl) < 4:
        vl.append("00000000")
    vl.append(b)
    return int("".join(vl))

def stdin():
    lines = ""
    while True:
        try:
            line = input()
        except EOFError:
            break
        lines += line
    return lines
