import copy
import re

from bloonix.validator import static_vars as static
import bloonix.utils as utils
from bloonix.json import json
from bloonix.genattr import GenAttr

def validate(data, rules, get_resultset=None, ignore_missing=None):
    v = Validator(get_resultset=get_resultset)
    v.set(rules)
    return v.validate(data, ignore_missing=ignore_missing)

class Validator:
    def __init__(
            self, rules=None, once=None, get_resultset=None,
            log=None, no_mixin=None, post_validation=None):
        self._once = once
        self._get_resultset = get_resultset
        self._options = {}
        self._defaults = {}
        self._rules = {}
        self._depends_on_keys = []
        self._ignore_missing = None
        self._no_mixin = no_mixin
        self._post_validation = post_validation

        # Example usage: self.log("message", level="DEBUG")
        self._log = log

        # __name__ is not the same in all python versions
        if rules is not None:
            self.set(rules)

    def set(self, rules=None, **kwargs):
        if rules is not None:
            self._check_dict("set", rules)
            rules.update(kwargs)
        else:
            rules = kwargs

        # do not change original passed data
        #rules = copy.deepcopy(rules)

        for key, rule in rules.items():
            if "function" in rule:
                __function = rule["function"]
                del rule["function"]
                rules[key] = copy.deepcopy(rule)
                rule["function"] = __function
            else:
                rules[key] = copy.deepcopy(rule)

            if "validator" in rule:
                if "type" in rule and not isinstance(rule["type"], dict):
                    raise Exception("invalid type for option 'validator'")
                else:
                    rule["type"] = dict
                for opt in rule:
                    if opt not in ["type", "validator", "optional", "default"]:
                        raise Exception(
                            "mixin option '%s' with option "
                            "'validator' is invalid"
                            % opt)
                self._options[key] = rule["validator"].get_options()
                self._defaults[key] = rule["validator"].get_defaults()
                self._rules[key] = rule
                continue    
            self._check_type(key, rule)
            self._check_options(key, rule)
            self._rules[key] = rule
            if "optional" not in rule:
                rule["optional"] = False
            if rule.get("maxvalue", 0) == "bigint":
                rule["type"] = int
                rule["maxvalue"] = 9223372036854775807
            if rule.get("minvalue", 0) == "bigint":
                rule["type"] = int
                rule["minvalue"] = -9223372036854775807
            for opt in static.LIST_OPTIONS:
                if opt in rule:
                    if key not in self._options:
                        self._options[key] = {}
                        self._options[key]["options"] = rule[opt]
            for opt in ["minlength", "maxlength", "minvalue", "maxvalue"]:
                if opt in rule:
                    if key not in self._options:
                        self._options[key] = {}
                    self._options[key][opt] = rule[opt]
            if "default" in rule:
                self._defaults[key] = rule["default"]
            for opt in ["regex", "regex_cl"]:
                if opt in rule:
                    rule[opt] = self._compile_re(opt, rule[opt])
            if "depends_on" in rule:
                if not isinstance(rule["depends_on"], dict):
                    raise Exception("invalid value type for rule 'depends_on'")
                for skey, srule in rule["depends_on"].items():
                    for skey in srule:
                        if skey not in static.ALLOWED_DEPENDS_ON_OPTIONS:
                            raise Exception("option '%s' does not exists" % skey)
                if "max_diff" in srule and "lower_or_greater" not in srule:
                    raise Exception(
                        "missing mandatory option 'lower_or_greater' "
                        "in combination with 'max_diff'")
                if ("lower_or_greater" in srule
                        and srule["lower_or_greater"] != "greater"
                        and srule["lower_or_greater"] != "lower"):
                    raise Exception("invalid value for rule lower_or_greater")
                self._depends_on_keys.append(key)
        return True

    def validate(self, data, ignore_missing=None):
        if self._rules == None or len(self._rules) == 0:
            raise Exception("no rules defined")
        self._ignore_missing = ignore_missing
        self._check_dict("validate", data)
        self._result = result = Result(data)
        self._data = data
        try:
            self._validate()
        except Exception:
            raise
        finally:
            self._data = None
            self._result = None
            if self._once is True:
                self._rules = None
                self._options = None
                self._defaults = None
                self._depends_on_keys = None
        if self._get_resultset is True:
            return result
        if result.has_failed is True:
            raise Exception("validation failed\nERR -> %s\n" % json.dumps(result.failed, indent=2))
        return result.data

    def get_options(self):
        return copy.deepcopy(self._options)

    def get_defaults(self):
        return copy.deepcopy(self._defaults)

    def get_params(self):
        return list(self._rules)

    def _compile_re(self, key, value):
        if type(value).__name__ == static.RE_TYPE_NAME:
            return value

        try:
            return re.compile(value)
        except Exception:
            pass

        raise Exception("unable to compile regex (key=%s)" % key, value)

    def _check_type(self, key, rule):
        if "type" not in rule:
            for opt in static.LIST_OPTIONS:
                if opt in rule:
                    rule["type"] = (list,)
                    return

        if "type" not in rule:
            raise Exception("missing type (key=%s)" % key)

        if isinstance(rule["type"], list):
            rule["type"] = tuple(rule["type"])
        elif not isinstance(rule["type"], tuple):
            rule["type"] = (rule["type"],)

        for t in rule["type"]:
            if t not in [str, int, float, bool, list, dict]:
                raise Exception("invalid type (key=%s)" % key)

        if len(rule["type"]) > 2:
            raise Exception("invalid number of types for (key=%s)" % key)
        if (len(rule["type"]) > 1 and (
                str in rule["type"]
                or bool in rule["type"]
                or (int in rule["type"] and not float in rule["type"])
                or (float in rule["type"] and not int in rule["type"])
                or (list in rule["type"] and not dict in rule["type"])
                or (dict in rule["type"] and not list in rule["type"]))):
            raise Exception("invalid mixin of types (key=%s)" % key)

    def _check_options(self, key, rule):
        for opt in list(rule):
            if opt not in static.ALLOWED_OPTIONS:
                raise Exception("option '%s' does not exists" % opt)

        if (str in rule["type"] and (
                "minvalue" in rule
                or "maxvalue" in rule
                or "minsize" in rule
                or "maxsize" in rule)):
            raise Exception(
                "mixin min-/maxvalue or min-/maxsize "
                "with type str is invalid")

        if ((int in rule["type"] or float in rule["type"]) and (
                "minlength" in rule
                or "maxlength" in rule
                or "minsize" in rule
                or "maxsize" in rule)):
            raise Exception(
                "mixin min-/maxlength or min-/maxsize "
                "with type int or float is invalid")

        if ((dict in rule["type"] or list in rule["type"]) and (
                "minvalue" in rule or "maxvalue" in rule)):
            raise Exception(
                "mixin min-/maxvalue "
                "with type dict or list is invalid")
        if ((dict in rule["type"] or bool in rule["type"])
                and "regex" in rule):
            raise Exception(
                "mixin type dict and bool with regex is invalid")

        if "regex_cl" in rule and str not in rule["type"]:
            raise Exception("regex_cl is only supported with type str")

        if (rule.get("force_list", None) is True
                and list not in rule["type"]):
            raise Exception("force_list is only supported with type list")

        hits = 0
        for opt in static.LIST_OPTIONS:
            if opt in rule:
                if hits == 1:
                    raise Exception(
                        "mixin options, multioptions, list_of_dicts_options "
                        "and list_of_dicts_multioptions is invalid")
                hits += 1

        for rule_type in ["list_of_dicts_options", "list_of_dicts_multioptions"]:
            if rule_type in rule:
                rule["__options"] = self._check_list_of_dicts_options(key, rule, rule_type)

        if ("isa" in rule
                and rule["isa"] != "bytes"
                and rule["isa"] != "seconds"
                and rule["isa"] != "unixtime"
                and rule["isa"] != "json_dump"
                and rule["isa"] != "date"):
            raise Exception("invalid 'isa' value")

    def _check_list_of_dicts_options(self, key, rule, rule_type):
        if "list_of_dicts_key" in rule:
            mandatory_key = rule["list_of_dicts_key"]
        else:
            mandatory_key = key
        arr = []
        for opt in rule[rule_type]:
            if not isinstance(opt, dict) or not mandatory_key in opt:
                raise Exception("invalid data structure (key=%s)" % key)
            arr.append(opt[mandatory_key])
        return arr

    def _check_dict(self, name, obj):
        if not isinstance(obj, dict):
            raise Exception(
                "the object passed to %s() is not a dict" % name)

    def _validate(self):
        for key, value in self._data.items():
            if key not in self._rules:
                raise Exception("unknown key '%s'" % key)
        for key, rule in self._rules.items():
            if not self._validate_exists(key, rule):
                continue
            if "validator" in rule:
                res = rule["validator"].validate(self._data[key])
                if res.has_failed is True:
                    self._result.has_failed = True
                    self._result.failed[key] = {"failed": res.failed}
                    self._result.data = res.data
                continue
            if rule.get("multiple", None) == True:
                data = self._data[key]
                if not isinstance(data, list):
                    data = [data]
                for i, item in enumerate(data):
                    # overwrite _data temporary
                    self._data[key] = item
                    self._validate_key_rule(key, rule)
                    data[i] = self._data[key]
                self._data[key] = data
            else:
                self._validate_key_rule(key, rule)

        if len(self._depends_on_keys):
            for key in self._depends_on_keys:
                self._validate_depends_on(key)

        if self._no_mixin != None:
            for arr in self._no_mixin:
                if not isinstance(arr, list):
                    raise Exception("invalid no_mixin format")
                hits = []
                for key in arr:
                    if key in self._data and self._data[key] != None:
                        hits.append(key)
                if len(hits) > 1:
                    for key in hits:
                        self._set_failed(key, "err-719")

        if self._post_validation:
            ret = self._post_validation(self._data)
            if isinstance(ret, dict):
                for key, status in ret.items():
                    if key not in self._result.failed:
                        self._set_failed(key, status)

    def _validate_key_rule(self, key, rule):
        if self._validate_none(key, rule):
            return
        if not self._validate_type(key, rule) or not self._validate_empty(key, rule):
            return
        if "minvalue" in rule and not self._validate_minvalue(key, rule):
            return
        if "maxvalue" in rule and not self._validate_maxvalue(key, rule):
            return
        if "minlength" in rule and not self._validate_minlength(key, rule):
            return
        if "maxlength" in rule and not self._validate_maxlength(key, rule):
            return
        if "minsize" in rule and not self._validate_minsize(key, rule):
            return
        if "maxsize" in rule and not self._validate_maxsize(key, rule):
            return
        if "regex" in rule and not self._validate_regex(key, rule):
            return
        if "regex_cl" in rule and not self._validate_regex_cl(key, rule):
            return
        if "function" in rule and not self._validate_function(key, rule):
            return
        if "options" in rule and not self._validate_options(key, rule):
            return
        if "multioptions" in rule and not self._validate_multioptions(key, rule):
            return
        if ("list_of_dicts_options" in rule
                and not self._validate_list_of_dicts_options(key, rule)):
            return
        if ("list_of_dicts_multioptions" in rule
                and not self._validate_list_of_dicts_multioptions(key, rule)):
            return
        if "isa" in rule and not self._convert_isa(key, rule):
            return
        if "isa_unixtime" in rule and not self._convert_to_unixtime(key, rule["isa_unixtime"]):
            return

    def _validate_exists(self, key, rule):
        if key in self._data:
            return True
        if self._ignore_missing == True:
            return False
        # default values will not be validated !!!
        if "default" in rule:
            self._data[key] = rule["default"]
            if (self._data[key] is None
                    or (isinstance(self._data[key], (dict,list))
                        and len(self._data[key]) == 0)):
                return False
            return True
        if rule.get("optional", None) is True:
            self._data[key] = None
            return False
        self._set_failed(key, "err-700")

    def _validate_type(self, key, rule):
        if (rule.get("force_list", None) is True
                and not isinstance(self._data[key], list)):
            self._data[key] = [self._data[key]]
        # If the type is a string, then the value will be converted to
        # 0 or None in _validate_empty.
        if (isinstance(self._data[key], str)
                and self._data[key] == ""
                and ("empty_to_none" in rule or "empty_to_null" in rule)):
            return True
        if not isinstance(self._data[key], rule["type"]):
            if not self._typecast(key, rule):
                return self._set_failed(key, "err-712")
        return True

    def _typecast(self, key, rule):
        if bool in rule["type"]:
            if self._data[key] is not True and self._data[key] is not False:
                if re.search(r"^(?:1|[Yy]es|[Tt]rue)\Z", str(self._data[key])):
                    self._data[key] = True
                elif re.search(r"^(?:0|[Nn]o|[Ff]alse)\Z", str(self._data[key])):
                    self._data[key] = False
                else:
                    return False
            return True

        if (isinstance(self._data[key], (int, float))
                and str in rule["type"]):
            self._data[key] = str(self._data[key])
            return True

        if isinstance(self._data[key], str):
            if (int in rule["type"]
                    and re.search(r"^\d+\Z", self._data[key])):
                self._data[key] = int(self._data[key])
                return True
            if (float in rule["type"]
                    and re.search(r"^\d+([,.]\d+)?\Z", self._data[key])):
                v = re.sub(",", ".", self._data[key])
                self._data[key] = float(v)
                return True

        return False

    def _validate_none(self, key, rule):
        allow_none = rule.get("allow_none", None)
        if (rule.get("allow_none", None) is True
                and self._data[key] is None):
            return True

    def _validate_empty(self, key, rule):
        allow_empty = rule.get("allow_empty", None)
        minlength = rule.get("minlength", None)
        minsize = rule.get("minsize", None)
        empty_to_null = rule.get("empty_to_null", None)
        empty_to_none = rule.get("empty_to_none", None)

        if (isinstance(self._data[key], str)
                and (str in rule["type"]
                    or int in rule["type"]
                    or float in rule["type"])):
            if re.search(r"^\s*\Z", self._data[key]):
                # If allow_empty is set or minlength is 0,
                # then just return False without an error
                # to stop further validation.
                if empty_to_null is True:
                    self._data[key] = 0
                    return False
                elif empty_to_none is True:
                    self._data[key] = None
                    return False
                else:
                    self._data[key] = "" # remove whitespaces
                if allow_empty is True: # minlength or maxlength is ignored
                    return False

        if ((list in rule["type"] or dict in rule["type"])
                and len(self._data[key]) == 0
                and (minsize is None or minsize > 0)
                and allow_empty is not True):
            self._set_failed(key, "err-716")
            return False

        return True

    def _validate_minvalue(self, key, rule):
        if self._data[key] < rule["minvalue"]:
            self._set_failed(key, "err-702")
            return False
        return True

    def _validate_maxvalue(self, key, rule):
        if self._data[key] > rule["maxvalue"]:
            self._set_failed(key, "err-703")
            return False
        return True

    def _get_string_length(self, key, rule):
        if str in rule["type"]:
            return len(self._data[key])
        if dict in rule["type"] or list in rule["type"]:
            return len(json.dumps(self._data[key]))
        raise Exception(
            "unexpected type '%s' (key=%s)"
            % (rule["type"], key))

    def _validate_minlength(self, key, rule):
        if self._get_string_length(key, rule) < rule["minlength"]:
            if str in rule["type"]:
                self._set_failed(key, "err-704")
            else:
                self._set_failed(key, "err-713")
            return False
        return True

    def _validate_maxlength(self, key, rule):
        if self._get_string_length(key, rule) > rule["maxlength"]:
            if str in rule["type"]:
                self._set_failed(key, "err-705")
            else:
                self._set_failed(key, "err-714")
            return False
        return True

    def _validate_minsize(self, key, rule):
        if len(self._data[key]) < rule["minsize"]:
            self._set_failed(key, "err-706")
            return False
        return True

    def _validate_maxsize(self, key, rule):
        if len(self._data[key]) > rule["maxsize"]:
            self._set_failed(key, "err-707")
            return False
        return True

    def _validate_regex(self, key, rule):
        if str in rule["type"] or int in rule["type"] or float in rule["type"]:
            if not rule["regex"].search(str(self._data[key])):
                self._set_failed(key, "err-710")
                return False
        elif list in rule["type"]:
            for value in self._data[key]:
                if (not isinstance(value, str)
                        or not rule["regex"].search(value)):
                    self._set_failed(key, "err-711")
                    return False
        return True

    def _validate_regex_cl(self, key, rule):
        if "regex_cl" not in rule:
            return True

        values = re.sub(r"\s", "", self._data[key])
        for value in re.split(",", values):
            if not rule["regex_cl"].search(value):
                self._set_failed(key, "err-711")
                return False

        return True

    def _validate_function(self, key, rule):
        args = rule.get("function_args", ())
        kwargs = rule.get("function_kwargs", {})
        ret = rule["function"](self._data[key], *args, **kwargs)
        if ret is True:
            return True
        if isinstance(ret, str):
            self._set_failed(key, ret)
        elif ret is False or ret is None:
            self._set_failed(key, "err-710")
        return False

    def _validate_options(self, key, rule):
        if self._data[key] not in rule["options"]:
            self._set_failed(key, "err-710")
            return False
        return True

    def _validate_multioptions(self, key, rule):
        for value in self._data[key]:
            if value not in rule["multioptions"]:
                self._set_failed(key, "err-711")
                return False
        return True

    def _validate_list_of_dicts_options(self, key, rule):
        if self._data[key] not in rule["__options"]:
            self._set_failed(key, "err-710")
            return True
        return False

    def _validate_list_of_dicts_multioptions(self, key, rule):
        if not isinstance(self._data[key], list):
            self._set_failed(key, "err-718")
            return False
        for value in self._data[key]:
            if value not in rule["__options"]:
                self._set_failed(key, "err-711")
                return False
        return True

    def _validate_depends_on(self, key):
        rule = self._rules[key]

        if (key not in self._data
                or self._data[key] is None
                or key in self._result.failed):
            return False

        for pkey, srule in rule["depends_on"].items():
            if pkey in self._result.failed:
                continue

            if (pkey not in self._data or self._data[pkey] is None):
                self._set_failed(pkey, "err-717")
                continue

            if "lower_or_greater" in srule:
                if (srule["lower_or_greater"] == "greater"
                        and self._data[key] < self._data[pkey]):
                    self._set_failed(key, "err-702")
                    self._set_failed(pkey, "err-703")
                    continue
                elif (srule["lower_or_greater"] == "lower"
                        and self._data[key] > self._data[pkey]):
                    self._set_failed(key, "err-702")
                    self._set_failed(pkey, "err-703")
                    continue

            if "max_diff" in srule:
                if (srule["lower_or_greater"] == "greater"
                        and self._data[key] - self._data[pkey] > srule["max_diff"]):
                    self._set_failed(key, "err-703")
                elif (srule["lower_or_greater"] == "lower"
                        and self._data[pkey] - self._data[key] > srule["max_diff"]):
                    self._set_failed(pkey, "err-703")

    def _convert_isa(self, key, rule):
        if rule["isa"] == "bytes":
            return self._convert_bytes(key)
        if rule["isa"] == "seconds":
            return self._convert_seconds(key)
        if rule["isa"] == "unixtime":
            return self._convert_to_unixtime(key)
        if rule["isa"] == "json_dump":
            self._data[key] = json.dumps(self._data[key])
            return True
        if ruls["isa"] == "date":
            return self._parse_date(key)

    def _convert_bytes(self, key):
        m = re.search(
            r"^(\d+)([kmgtpe]{0,1})b{0,1}\Z", self._data[key], re.IGNORECASE)
        if not m:
            self._set_failed(key, "err-710")
            return False
        num = int(m.group(1))
        byt = m.group(2).lower()
        if len(byt) == 0:
            byt = "b"
        self._data[key] = num * static.BYTES[byt]
        return True

    def _convert_seconds(self, key):
        m = re.search(r"(\d+)([smhd]{0,1})\Z", self._data[key], re.IGNORECASE)
        if not m:
            self._set_failed(key, "err-710")
            return False
        num = int(m.group(1))
        sec = m.group(2).lower()
        if sec in static.TIME:
            num *= static.TIME[sec]
        self._data[key] = num
        return True

    def _convert_to_unixtime(self, key, tz=None):
        if isinstance(self._data[key], int):
            # seems to be already in unixtime format
            return True
        if re.search(r"^\d+\Z", self._data[key]):
            # already unixtime, but a string
            self._data[key] = int(self._data[key])
            return True
        if re.search(r"^\d\d\d\d-\d\d-\d\d\Z", self._data[key]):
            return self._convert_string_to_unixtime(
                key, "%s %s" % (self._data[key], "00:00:00"), tz)
        if re.search(r"^\d\d\d\d-\d\d-\d\d\s+\d\d:\d\d\Z", self._data[key]):
            return self._convert_string_to_unixtime(key, self._data[key] +":00", tz)
        if re.search(r"^\d\d\d\d-\d\d-\d\d\s+\d\d:\d\d:\d\d\Z", self._data[key]):
            return self._convert_string_to_unixtime(key, self._data[key], tz)
        self._set_failed(key, "err-710")
        return False

    def _convert_string_to_unixtime(self, key, timestring, tz=None):
        try:
            self._data[key] = utils.unixtime(timestring, tz=tz)
            return True
        except Exception:
            self._set_failed(key, "err-710")
            return False

    def _parse_date(self, key):
        m = re.search(r"^(\d\d\d\d)-(\d{1,2})-(\d{1,2})\Z", self._data[key])
        if not m:
            return False

        year = int(m.group(1))
        month = int(m.group(1))
        day = int(m.group(1))

        if month < 1 or month > 12 or day < 1:
            return False

        if month == 2:
            if (year % 400 == 0 or (year % 4 == 0 and year % 100 > 0)):
                if day > 29:
                    return False
            elif day > 28:
                return False
        elif month == 4 or month == 6 or month == 9 or month == 11:
            if day > 30:
                return False
        elif day > 31:
            return False

        self._data[key] = "%04d-%02d-%02d" % (year, month, day)
        return True

    def _set_failed(self, key, status, value=None):
        if key not in self._rules:
            raise Exception(
                "internal validation error: no rule exists for key '%s'"
                % key)

        if status not in static.ERROR_MESSAGE:
            status = "err-701"

        self._result.has_failed = True

        if self._rules[key].get("multiple", None) is True:
            if key not in self._result.failed:
                self._result.failed[key] = {
                    "status": "err-711",
                    "message": static.ERROR_MESSAGE["err-711"],
                    "values": []}
            if key in self._data:
                self._result.failed[key]["values"].append(self._data[key])
            return

        self._result.failed[key] = {
            "status": status,
            "message": static.ERROR_MESSAGE[status]}

    def log(self, *args, **kwargs):
        # Example usage: self.log("message", level="DEBUG")
        if self._log:
            self._log.log(*args, **kwargs)


class Result:
    def __init__(self, data):
        self.failed = {}
        self.has_failed = None
        self.data = data
