#!/usr/bin/python3

from bloonix.dbi import DBI
import re
import unittest
import warnings

def init_db(driver, port):
    dbi = DBI.setup(
        driver=driver,
        port=port,
        host="127.0.0.1",
        user="bloonix",
        password="bloonix",
        database="bloonix"
    )

    try: dbi.do("drop table t1")
    except Exception: pass
    try: dbi.do("drop table t2")
    except Exception: pass
    try: dbi.do("drop table t3")
    except Exception: pass
    try: dbi.do("drop table t4")
    except Exception: pass
    try: dbi.do("drop table t5")
    except Exception: pass

    dbi.do("create table t1 (a varchar(10))")
    dbi.do("create table t2 (a varchar(10))")
    dbi.do("create table t3 (a varchar(10))")
    dbi.do("create table t4 (a varchar(10), unique(a))")
    dbi.do("create table t5 (a json default '{}')")

    return dbi

def init_pg():
    return init_db("pg", 5432)

def init_my():
    return init_db("mysql", 3306)

def init_ma():
    return init_db("mariadb", 3306)


class TestDBI(unittest.TestCase):
    def _transaction(self, dbi):
        dbi.fetchall("select * from t1")
        dbi.fetchall("select * from t2")
        dbi.do("insert into t3 (a) values ('b')")

    def _transaction_rollback(self, dbi):
        dbi.fetchall("select * from t1")
        dbi.fetchall("select * from t2")
        dbi.do("insert into t3 (a) values ('b')")
        raise Exception("force rollback")

    def _run_transaction(self, dbi):
        dbi.run_transaction(
            {"t1": "write", "t2": "write", "t3": "write"},
            self._transaction,
            dbi
        )
        row = dbi.fetchone("select * from t3")
        self.assertTrue(row["a"] == "b")

    def _run_transaction_rollback(self, dbi):
        try:
            dbi.run_transaction(
                {"t1": "write", "t2": "write", "t3": "write"},
                self._transaction_rollback,
                dbi
            )
        except Exception as e:
            if not re.search("force rollback", str(e)):
                raise

        row = dbi.fetchone("select * from t3")
        self.assertTrue(row is None)

    def test_pg_tranaction(self):
        dbi = init_pg()
        self._run_transaction(dbi)

    def test_pg_tranaction_rollback(self):
        dbi = init_pg()
        self._run_transaction_rollback(dbi)

    def test_pg_json(self):
        self.__test_json(init_pg())

    def test_mariadb_tranaction(self):
        dbi = init_ma()
        self._run_transaction(dbi)

    def test_mariadb_tranaction_rollback(self):
        dbi = init_ma()
        self._run_transaction_rollback(dbi)

    def test_mariadb_json(self):
        self.__test_json(init_ma())

    #def test_mysql_tranaction(self):
    #    dbi = init_my()
    #    self._run_transaction(dbi)

    #def test_mysql_tranaction_rollback(self):
    #    dbi = init_my()
    #    self._run_transaction_rollback(dbi)

    #def test_mysql_json(self):
    #    self.__test_json(init_my())

    def __test_json(self, dbi):
        dbi.do("insert into t5 (a) values ('{}')")
        row = dbi.fetchone("select * from t5")
        self.assertTrue(isinstance(row["a"], dict))


unittest.main(verbosity=2)
