#!/usr/bin/python3

import os
import sys
import time
import unittest
from bloonix.exception import BloonixError, BloonixTimeout
import bloonix.socket.serialized as SIPC

class TestRX(unittest.TestCase):
    def setUp(self):
        if os.path.isfile("tests/scripts/client.py"):
            self.clientcmd = "tests/scripts/client.py"
            self.sslkey = "tests/ssl/server.key"
            self.sslcert = "tests/ssl/server.cert"
        else:
            self.clientcmd = "scripts/client.py"
            self.sslkey = "ssl/server.key"
            self.sslcert = "ssl/server.cert"

    def listen(self, **kwargs):
        kwargs["bind"] = kwargs.get("host", "127.0.0.1")
        kwargs["port"] = kwargs.get("port", 4711)
        kwargs["use_ssl"] = kwargs.get("use_ssl", True)
        kwargs["ssl_key_file"] = kwargs.get("ssl_key_file", self.sslkey)
        kwargs["ssl_cert_file"] = kwargs.get("ssl_cert_file", self.sslcert)
        kwargs["recv_max_bytes"] = kwargs.get("recv_max_bytes", 0)
        sipc = SIPC.Server(**kwargs)
        sipc.listen()
        return sipc

    def test_basic(self):
        sipc = self.listen()
        os.system(self.clientcmd + " &")
        client = sipc.accept()
        request = client.recv()
        client.send({"action":"pong"})
        client.disconnect()
        sipc.stop_listen()
        self.assertTrue(type(request) is dict)
        self.assertTrue("action" in request)
        self.assertTrue(request["action"] == "ping")

    def test_recv_max_bytes(self):
        sipc = self.listen(recv_max_bytes=10)
        os.system(self.clientcmd + " &")
        client = sipc.accept()

        self.assertRaisesRegex(
            BloonixError,
            "recv: the data length \(17 bytes\) exceeds max_bytes",
            client.recv)

        client.disconnect()
        sipc.stop_listen()

    def test_send_max_bytes(self):
        sipc = self.listen(send_max_bytes=10)
        os.system(self.clientcmd + " &")
        client = sipc.accept()
        request = client.recv()

        self.assertRaisesRegex(
            BloonixError,
            "send: the data length \(17 bytes\) exceeds max_bytes",
            client.send, {"action":"pong"})

        client.disconnect()
        sipc.stop_listen()

    def test_timeout(self):
        sipc = self.listen(timeout=2)
        os.system(self.clientcmd + " wait &")
        client = sipc.accept()

        try:
            request = client.recv()
            client.send({"action":"pong"})
        except BloonixTimeout as msg:
            if str(msg) != "recv runs on a timeout after 2s":
                raise

        client.disconnect()
        sipc.stop_listen()

    def test_recv_timeout(self):
        sipc = self.listen(recv_timeout=2)
        os.system(self.clientcmd + " wait &")
        client = sipc.accept()

        try:
            request = client.recv()
            client.send({"action":"pong"})
        except BloonixTimeout as msg:
            if str(msg) != "recv runs on a timeout after 2s":
                raise Exception(msg)

        client.disconnect()
        sipc.stop_listen()

    def test_send_timeout(self):
        sipc = self.listen(send_timeout=2)
        sipc.test_timeout = 5
        os.system(self.clientcmd + " wait &")
        client = sipc.accept()

        try:
            request = client.recv()
            client.send({"action":"pong"})
        except BloonixTimeout as msg:
            if str(msg) != "send runs on a timeout after 2s":
                raise

        client.disconnect()
        sipc.stop_listen()

    def test_connect_timeout(self):
        sipc = SIPC.Client(
            host="127.0.0.1",
            port=4711,
            use_ssl=True,
            ssl_verify_mode="none",
            connect_timeout=2
        )

        try:
            sipc.connect(test_connect_timeout=5, retry=False)
        except BloonixTimeout as msg:
            if str(msg) != "connect runs on a timeout after 2s":
                raise


unittest.main(verbosity=2)
