#!/usr/bin/python3

import unittest
from bloonix.exception import BloonixError
from bloonix.sql import SQL

sql = SQL(driver="pg")

class TestTemplate(unittest.TestCase):
    global sql

    def test_select_01(self):
        stmt, bind = sql.select(
            table="test",
            columns=["a", "b", "c"]
        )
        self.assertTrue(
            stmt == 'select "test"."a", "test"."b", "test"."c" from "test"')
        self.assertTrue(len(bind) == 0)

    def test_select_02(self):
        stmt, bind = sql.select(
            table=[
                "test1", ["a", "b"],
                "test2", ["c", "d"]
            ]
        )
        self.assertTrue(stmt == " ".join([
            'select',
                '"test1"."a", "test1"."b",',
                '"test2"."c", "test2"."d"',
            'from "test1"'
        ]))
        self.assertTrue(len(bind) == 0)

    def test_select_03(self):
        stmt, bind = sql.select(
            distinct=True,
            table=[
                "service", "*",
                "host", ["hostname", "interval", "ipaddr"],
                "company", ["id AS company_id", "company"],
                "status_priority", "priority",
                "plugin", "plugin"
            ],
            join=[
                "inner", ["host", "service.host_id", "host.id"],
                "inner", ["company", "host.company_id", "company.id"],
                "inner", [
                    "status_priority", "service.status",
                    "status_priority.status"],
                "inner", ["host_group", "host.id", "host_group.host_id"],
                "inner", [
                    "user_group", "host_group.group_id",
                    "user_group.group_id"],
                "inner", ["plugin", "service_parameter.plugin_id", "plugin.id"]
            ],
            condition=[
                "where", "(",
                    ["user_group.user_id", "x1"],
                    "and", ["host.id", ["x2","x3", "x4"]],
                ")"
            ],
            order=[
                "desc", [
                    "status_priority.priority",
                    "service.status_nok_since"],
                "asc", ["host.hostname"]
            ],
            limit="x5",
            offset="x6"
        )

        self.assertTrue(self.__test_select_stmt_03(stmt))
        self.assertTrue(self.__test_select_bind_03(bind))

    def __test_select_stmt_03(self, stmt):
        return stmt == (
            'select distinct '
                '"service".*, '
                '"host"."hostname", "host"."interval", "host"."ipaddr", '
                '"company"."id" as "company_id", "company"."company", '
                '"status_priority"."priority", '
                '"plugin"."plugin" '
            'from "service" '
            'inner join "host" on "service"."host_id" = "host"."id" '
            'inner join "company" on "host"."company_id" = "company"."id" '
            'inner join "status_priority" on '
                '"service"."status" = "status_priority"."status" '
            'inner join "host_group" on "host"."id" = "host_group"."host_id" '
            'inner join "user_group" on '
                '"host_group"."group_id" = "user_group"."group_id" '
            'inner join "plugin" on '
                '"service_parameter"."plugin_id" = "plugin"."id" '
            'where ( '
                '"user_group"."user_id" = %s and "host"."id" in (%s, %s, %s) '
            ') '
            'order by '
                '"status_priority"."priority" desc, '
                '"service"."status_nok_since" desc, '
                '"host"."hostname" asc '
            'limit %s '
            'offset %s'
        )

    def __test_select_bind_03(self, bind):
        if (len(bind) == 6 and
                bind[0] == "x1" and bind[1] == "x2" and bind[2] == "x3" and
                bind[3] == "x4" and bind[4] == "x5" and bind[5] == "x6"):
            return True
        return False

    def test_select_04(self):
        stmt, bind = sql.find("table", "a", "b", "c", "d")
        self.assertTrue(
            stmt == (
                'select "table".* from "table" where "a" = %s and "c" = %s'
            ))
        self.assertTrue(len(bind) == 2 and bind[0] == "b" and bind[1] == "d")

    def test_select_05(self):
        stmt, bind = sql.select(
            table=["ticket", "*"],
            condition=[
                "where", ["ticket.status", "x1"],
                "and", ["ticket.ticket_queue_id", "in", {
                    "table": "admin_ticket_queue",
                    "columns": "ticket_queue_id",
                    "condition": ["where", ["admin_id", "x2"]]
                }]
            ],
            order=["desc", "ticket.id"]
        )
        self.assertTrue(stmt == (
            'select "ticket".* from "ticket" '
            'where "ticket"."status" = %s '
            'and "ticket"."ticket_queue_id" in '
            '(select "admin_ticket_queue"."ticket_queue_id" '
                'from "admin_ticket_queue" where "admin_id" = %s) '
            'order by "ticket"."id" desc'))
        self.assertTrue(bind[0] == "x1" and bind[1] == "x2")

    def test_select_05_2(self):
        stmt, bind = sql.select(
            table=["ticket", "*"],
            condition=[
                "where", ["ticket.status", "x1"],
                "and", ["ticket.ticket_queue_id", "not in", {
                    "table": "admin_ticket_queue",
                    "columns": "ticket_queue_id",
                    "condition": ["where", ["admin_id", "x2"]]
                }]
            ],
            order=["desc", "ticket.id"]
        )
        self.assertTrue(stmt == (
            'select "ticket".* from "ticket" '
            'where "ticket"."status" = %s '
            'and "ticket"."ticket_queue_id" not in '
            '(select "admin_ticket_queue"."ticket_queue_id" '
                'from "admin_ticket_queue" where "admin_id" = %s) '
            'order by "ticket"."id" desc'))
        self.assertTrue(bind[0] == "x1" and bind[1] == "x2")

    def test_select_06(self):
        stmt, bind = sql.select(
            table=["ticket", "*"],
            condition=[
                "where not exists", {
                    "table": "admin_ticket_queue",
                    "columns": "ticket_queue_id",
                    "condition": ["where", ["admin_id", "x1"]]
                }
            ],
            order=["desc", "ticket.id"]
        )
        self.assertTrue(stmt == (
            'select "ticket".* from "ticket" '
            'where not exists '
            '(select "admin_ticket_queue"."ticket_queue_id" '
                'from "admin_ticket_queue" where "admin_id" = %s) '
            'order by "ticket"."id" desc'))
        self.assertTrue(bind[0] == "x1")

    def test_select_07(self):
        stmt, bind = sql.select(
            table="table",
            count="id"
        )
        self.assertTrue(
            stmt == 'select count("table"."id") AS count from "table"')

    def test_select_08(self):
        stmt, bind = sql.select(
            table="table",
            count="table2.id",
        )
        self.assertTrue(stmt == (
            'select count("table2"."id") AS count from "table"'
        ))

    def test_select_09(self):
        stmt, bind = sql.select(
            distinct=True,
            table="table",
            count=["id", "name"]
        )
        self.assertTrue(stmt == (
            'select count(distinct("table"."id", "table"."name")) AS count '
            'from "table"'
        ))

    def test_select_10(self):
        stmt, bind = sql.select(
            distinct=True,
            table="table",
            count=["id", "name"],
            columns=["status"],
            group=["status"]
        )
        self.assertTrue(stmt == (
            'select count(distinct("table"."id", "table"."name")) AS count, '
            '"table"."status" from "table" group by "status"'
        ))

    def test_select_11(self):
        stmt, bind = sql.select(
            distinct=True,
            table="table",
            count=["id", "table2.name"],
            columns=["status"],
            group=["status"]
        )
        self.assertTrue(stmt == (
            'select count(distinct("table"."id", "table2"."name")) AS count, '
            '"table"."status" from "table" group by "status"'
        ))

    def test_select_12(self):
        stmt, bind = sql.select(
            distinct=True,
            table="table",
            columns=["id"]
        )
        self.assertTrue(stmt == 'select distinct "table"."id" from "table"')

    def test_select_13(self):
        stmt, bind = sql.select(
            table=["ticket", "*"],
            condition=["where", ("col1", "=", "col2")]
        )
        self.assertTrue(stmt == (
            'select "ticket".* from "ticket" '
            'where "col1" = "col2"'
        ))

    def test_select_14(self):
        stmt, bind = sql.select(
            table="ticket",
            columns=1
        )
        self.assertTrue(stmt == 'select 1 from "ticket"')

    def test_select_15(self):
        stmt, bind = sql.select(
            table=[
                "a", "min(b)",
                "c", "max(d)",
                "e", "sum(f)",
                "g", "avg(h)",
                "a", "min(b) AS ab",
                "c", "max(d) AS cd",
                "e", "sum(f) AS ef",
                "g", "avg(h) AS gh"
            ]
        )
        self.assertTrue(
            stmt == (
                'select min("a"."b"), max("c"."d"), '
                    'sum("e"."f"), avg("g"."h"), '
                    'min("a"."b") as "ab", max("c"."d") as "cd", '
                    'sum("e"."f") as "ef", avg("g"."h") as "gh" '
                'from "a"'
            ))

    def test_select_16(self):
        stmt, bind = sql.select(
            table="test",
            columns="*",
            condition=[
                "where", [
                    "a", 10
                ],
                "and", {
                    "math": ["b", "+", "d", "<", 100]
                },
                "and", [
                    "c", 30
                ]
            ]
        )
        self.assertTrue(stmt == 'select "test".* from "test" where "a" = %s and "b" + "d" < %s and "c" = %s')
        self.assertTrue(bind == [10, 100, 30])

    def test_select_17(self):
        stmt, bind = sql.select(
            table="test",
            columns={
                "func": "coalesce",
                "value": {
                    "func": "max",
                    "column": "a"
                },
                "default": 1,
                "as": "b"
            })
        print(stmt, bind)

    def test_insert(self):
        stmt, bind = sql.insert("test", {"col1": "val1", "col2": "val2"})
        self.assertTrue(
            (stmt == 'insert into "test" ("col1", "col2") values (%s, %s)'
                and bind[0] == "val1" and bind[1] == "val2")
            or (stmt == 'insert into "test" ("col2", "col1") values (%s, %s)'
                and bind[0] == "val2" and bind[1] == "val1")
        )
        self.assertTrue(len(bind) == 2)

    def test_bulk_insert(self):
        stmt, bind = sql.insert(
            table="test",
            data=[
                {"c1": 1, "c2": 2, "c3": 3},
                {"c1": 4, "c2": 5, "c3": 6},
                {"c1": 7, "c2": 8, "c3": 9},
            ])
        self.assertTrue(stmt == (
            'insert into "test" (c1, c2, c3) '
            'values (%s,%s,%s), (%s,%s,%s), (%s,%s,%s)'))
        self.assertTrue(len(bind) == 9)
        for i in range(9):
            self.assertTrue(bind[i] == i+1)

    def test_update_01(self):
        stmt, bind = sql.update(
            "test", # table
            {"col1": "val1", "col2": "val2"},
            ["where", ["col3", "val3"], "and", ["col4", "!=", "val4"]]
        )
        self.assertTrue(
            (stmt == 'update "test" set "col1" = %s, "col2" = %s'
                    ' where "col3" = %s and "col4" != %s' and
                bind[0] == "val1" and bind[1] == "val2" and
                bind[2] == "val3" and bind[3] == "val4")
            or (stmt == 'update "test" set "col2" = %s, "col1" = %s'
                    ' where "col3" = %s and "col4" != %s' and
                bind[0] == "val2" and bind[1] == "val1" and
                bind[2] == "val3" and bind[3] == "val4")
        )
        self.assertTrue(len(bind) == 4)

    def test_update_02(self):
        stmt, bind = sql.update_by_id(
            "test", # table
            12345, # id
            {"col1": "val1", "col2": "val2"} # data
        )
        self.assertTrue(
            (stmt == (
                    'update "test" set "col1" = %s, '
                    '"col2" = %s where "id" = %s')
                and bind[0] == "val1"
                and bind[1] == "val2"
                and bind[2] == 12345)
            or (stmt == (
                    'update "test" set "col2" = %s, '
                    '"col1" = %s where "id" = %s')
                and bind[0] == "val2"
                and bind[1] == "val1"
                and bind[2] == 12345)
        )
        self.assertTrue(len(bind) == 3)

    def test_update_03(self):
        stmt, bind = sql.update_by_col(
            "test", # table
            "col", # column
            "val", # value
            {"col1": "val1", "col2": "val2"} # data
        )
        self.assertTrue(
            (stmt == (
                    'update "test" set "col1" = %s, '
                    '"col2" = %s where "col" = %s')
                and bind[0] == "val1"
                and bind[1] == "val2"
                and bind[2] == "val")
            or (stmt == (
                    'update "test" set "col2" = %s, '
                    '"col1" = %s where "col" = %s')
                and bind[0] == "val2"
                and bind[1] == "val1"
                and bind[2] == "val")
        )
        self.assertTrue(len(bind) == 3)

    def test_update_04(self):
        stmt, bind = sql.update(
            table="t1",
            data={},
            sdata={
                "status": {
                    "table": "t2",
                    "columns": "max(c2)",
                    "condition": ["where", ["c3", 1]]
                }
            },
            condition=["where", ["t3", 2]]
        )
        self.assertTrue(
            stmt == (
                'update "t1" set "status" = '
                '(select max("t2"."c2") from "t2" where "c3" = %s) '
                'where "t3" = %s'
            ))
        self.assertTrue(len(bind) == 2 and bind[0] == 1 and bind[1] == 2)


unittest.main(verbosity=2)
