#!/usr/bin/python3

import unittest
from bloonix.validator import Validator
import bloonix.rx as rx
import re

class TestBloonixValidator(unittest.TestCase):
    def test_init(self):
        v = Validator()
        self.assertTrue(v._get_resultset is None)
        self.assertTrue(v._once is None)
        self.assertTrue(isinstance(v._rules, dict))
        self.assertTrue(isinstance(v._options, dict))
        self.assertTrue(isinstance(v._defaults, dict))

    def test_check_arguments(self):
        v = Validator()
        self.assertRaisesRegex(
            Exception,
            "the object passed to set\(\) is not a dict",
            v.set, "invalid")
        self.assertRaisesRegex(
            Exception,
            "the object passed to set\(\) is not a dict",
            v.set, {"foo": "invalid"})
        self.assertRaisesRegex(
            Exception,
            "the object passed to set\(\)",
            v.set, foo="invalid")
        self.assertRaisesRegex(
            Exception,
            "no rules defined",
            v.validate, {})
        self.assertTrue(v.set(test={"type": int}) is True)
        self.assertRaisesRegex(
            Exception,
            "the object passed to validate\(\) is not a dict",
            v.validate, "foo")

    def test_type(self):
        v = Validator()

        # missing type
        self.assertRaisesRegex(
            Exception,
            "missing type \(key=string\)",
            v.set, string={"default": "10"})

        # set type to list for options
        for opt in ["options", "multioptions"]:
            v.set({opt: {opt: [1,2,3]}})
            self.assertTrue(Exception, v._rules[opt]["type"] is list)

        # set type to list for form options
        for opt in ["list_of_dicts_options", "list_of_dicts_multioptions"]:
            v.set({
                opt: {
                    opt: [{"label": "label", "value": 1}],
                    "list_of_dicts_key": "value",
                }
            })
            self.assertTrue(Exception, v._rules[opt]["type"] is list)

    def test_options(self):
        v = Validator()
        for o in [
                [list, "minvalue"],
                [list, "maxvalue"],
                [dict, "minvalue"],
                [dict, "maxvalue"]]:
            self.assertRaisesRegex(
                Exception,
                "mixin min-/maxvalue with type dict or list is invalid",
                v.set, {"value": {"type": o[0], o[1]: 0}})
        for o in ["minvalue", "maxvalue", "minsize", "maxsize"]:
            self.assertRaisesRegex(
                Exception,
                "mixin min-/maxvalue or min-/maxsize with type str is invalid",
                v.set, {"value": {"type": str, o: 0}})
        for o in [
                [int, "minlength"],
                [int, "maxlength"],
                [int, "minsize"],
                [int, "maxsize"],
                [float, "minlength"],
                [float, "maxlength"],
                [float, "minsize"],
                [float, "maxsize"]]:
            self.assertRaisesRegex(
                Exception,
                ("mixin min-/maxlength or min-/maxsize with "
                "type int or float is invalid"),
                v.set, {"value": {"type": o[0], o[1]: 0}})
        for o in [
                [dict, "regex"],
                [bool, "regex"]]:
            self.assertRaisesRegex(
                Exception,
                "mixin type dict and bool with regex is invalid",
                v.set, {"value": {"type": o[0], o[1]: "rx"}})
        for o in [int, float, bool, list, dict]:
            self.assertRaisesRegex(
                Exception,
                "regex_cl is only supported with type str",
                v.set, {"value": {"type": o, "regex_cl": "rx"}})
        self.assertRaisesRegex(
            Exception,
                ("mixin options, multioptions, list_of_dicts_options "
                 "and list_of_dicts_multioptions is invalid"),
                v.set, {"value": {"options": [], "multioptions": []}})
        self.assertRaisesRegex(
            Exception,
                "invalid 'isa' value",
                v.set, {"value": {"type": str, "isa": "not exists"}})

    def test_typecast_int(self):
        v = Validator()
        v.set(value={"type": int})
        data = v.validate({"value": "10"})
        self.assertTrue(data["value"] == 10)
        self.assertRaisesRegex(
            Exception,
            "err-712",
            v.validate, {"value": "string"})

    def test_typecast_float(self):
        v = Validator()
        v.set(value={"type": float})
        data = v.validate({"value": "1.0"})
        self.assertTrue(data["value"] == 1.0)
        data = v.validate({"value": "1,0"})
        self.assertTrue(data["value"] == 1.0)
        self.assertRaisesRegex(
            Exception,
            "err-712",
            v.validate, {"value": "string"})

    def test_typecast_int_float_int(self):
        v = Validator()
        v.set(value={"type": (int,float)})
        data = v.validate({"value": "10"})
        self.assertTrue(data["value"] == 10)
        self.assertRaisesRegex(
            Exception,
            "err-712",
            v.validate, {"value": "string"})

    def test_typecast_int_float_float(self):
        v = Validator()
        v.set(value={"type": (int,float)})
        data = v.validate({"value": "1.0"})
        self.assertTrue(data["value"] == 1.0)
        data = v.validate({"value": "1,0"})
        self.assertTrue(data["value"] == 1.0)
        self.assertRaisesRegex(
            Exception,
            "err-712",
            v.validate, {"value": "string"})

    def test_force_list_default(self):
        v = Validator()
        v.set(value={
            "type": list,
            "force_list": True,
            "default": "string"
        })
        data = v.validate({})
        self.assertTrue(isinstance(data["value"], list))
        self.assertTrue(len(data["value"]) == 1)
        self.assertTrue(data["value"][0] == "string")

    def test_force_list_default_none(self):
        v = Validator()
        v.set(value={
            "type": list,
            "force_list": True,
            "default": None
        })
        data = v.validate({})
        self.assertTrue(data["value"] is None)

    def test_force_list_optional(self):
        v = Validator()
        v.set(value={
            "type": list,
            "force_list": True,
            "optional": True
        })
        data = v.validate({})
        self.assertTrue(data["value"] is None)

    def test_force_list(self):
        v = Validator()
        v.set(value={"type": list, "force_list": True})
        data = v.validate({"value": "string"})
        self.assertTrue(isinstance(data["value"], list))
        self.assertTrue(len(data["value"]) == 1)
        self.assertTrue(data["value"][0] == "string")

    def test_min_value(self):
        v = Validator()
        v.set(value={"type": int, "minvalue": 10})
        self.assertRaisesRegex(
            Exception,
            "err-702",
            v.validate, {"value": 5})
        data = v.validate({"value": 20})
        self.assertTrue(data["value"] == 20)

    def test_max_value(self):
        v = Validator()
        v.set(value={"type": int, "maxvalue": 10})
        self.assertRaisesRegex(
            Exception,
            "err-703",
            v.validate, {"value": 15})
        data = v.validate({"value": 5})
        self.assertTrue(data["value"] == 5)

    def test_min_length_string(self):
        v = Validator()
        v.set(value={"type": str, "minlength": 10})
        self.assertRaisesRegex(
            Exception,
            "err-704",
            v.validate, {"value": "test"})
        data = v.validate({"value": "test test test"})
        self.assertTrue(data["value"] == "test test test")

    def test_max_length_string(self):
        v = Validator()
        v.set(value={"type": str, "maxlength": 10})
        self.assertRaisesRegex(
            Exception,
            "err-705",
            v.validate, {"value": "test test test"})
        data = v.validate({"value": "test"})
        self.assertTrue(data["value"] == "test")

    def test_min_length_dict(self):
        v = Validator()
        v.set(value={"type": dict, "minlength": 20})
        self.assertRaisesRegex(
            Exception,
            "err-713",
            v.validate, {"value": {"test": "test"}})

    def test_max_length_dict(self):
        v = Validator()
        v.set(value={"type": dict, "maxlength": 10})
        self.assertRaisesRegex(
            Exception,
            "err-714",
            v.validate, {"value": {"test": "test"}})

    def test_min_length_list(self):
        v = Validator()
        v.set(value={"type": list, "minlength": 20})
        self.assertRaisesRegex(
            Exception,
            "err-713",
            v.validate, {"value": ["test", "test"]})

    def test_max_length_list(self):
        v = Validator()
        v.set(value={"type": list, "maxlength": 10})
        self.assertRaisesRegex(
            Exception,
            "err-714",
            v.validate, {"value": ["test", "test"]})

    def test_min_size_dict(self):
        v = Validator()
        v.set(value={"type": dict, "minsize": 10})
        self.assertRaisesRegex(
            Exception,
            "err-706",
            v.validate, {"value": {"test": "test"}})

    def test_max_size_dict(self):
        v = Validator()
        v.set(value={"type": dict, "maxsize": 1})
        self.assertRaisesRegex(
            Exception,
            "err-707",
            v.validate, {"value": {"test1": 1, "test2": 2}})

    def test_min_size_list(self):
        v = Validator()
        v.set(value={"type": list, "minsize": 10})
        self.assertRaisesRegex(
            Exception,
            "err-706",
            v.validate, {"value": ["test", "test"]})

    def test_max_size_list(self):
        v = Validator()
        v.set(value={"type": list, "maxsize": 1})
        self.assertRaisesRegex(
            Exception,
            "err-707",
            v.validate, {"value": ["test", "test"]})

    def test_regex_string(self):
        v = Validator()
        v.set(value={"type": str, "regex": "foo"})
        data = v.validate({"value": "foo"})
        self.assertTrue(data["value"] == "foo")
        self.assertRaisesRegex(
            Exception,
            "err-710",
            v.validate, {"value": "bar"})

    def test_regex_list(self):
        v = Validator()
        v.set(value={"type": list, "regex": "foo"})
        data = v.validate({"value": ["foo", "foo", "foo"]})
        self.assertTrue(data["value"][0] == "foo")
        self.assertRaisesRegex(
            Exception,
            "err-711",
            v.validate, {"value": ["foo", "bar"]})

    def test_regex_compiled(self):
        v = Validator()
        v.set(value={"type": str, "regex": re.compile("foo")})
        data = v.validate({"value": "foo"})
        self.assertTrue(data["value"] == "foo")
        self.assertRaisesRegex(
            Exception,
            "err-710",
            v.validate, {"value": "bar"})

    def test_regex_allow_empty(self):
        v = Validator()
        v.set(value={"type": str, "regex": "foo", "allow_empty": True})
        data = v.validate({"value": ""})
        self.assertTrue(data["value"] == "")

    def test_regex_allow_none(self):
        v = Validator()
        v.set(value={"type": str, "regex": "foo", "allow_none": True})
        data = v.validate({"value": None})
        self.assertTrue(data["value"] == None)

    def test_regex_cl(self):
        v = Validator()
        v.set(value={"type": str, "regex_cl": "foo"})
        data = v.validate({"value": " foo , foo , foo "})
        self.assertTrue(data["value"] == " foo , foo , foo ")
        self.assertRaisesRegex(
            Exception,
            "err-711",
            v.validate, {"value": " foo , bar , baz "})

    def test_isa_bytes(self):
        v = Validator()
        v.set(value={"type": str, "isa": "bytes"})
        for o in [
                ["1", 1],
                ["1kb", 1024],
                ["1mb", 1048576],
                ["1gb", 1073741824],
                ["1tb", 1099511627776],
                ["1pb", 1125899906842624],
                ["1eb", 1152921504606846976]]:
            data = v.validate({"value": o[0]})
            self.assertTrue(data["value"] == o[1])

    def test_isa_seconds(self):
        v = Validator()
        v.set(value={"type": str, "isa": "seconds"})
        for o in [
                ["1", 1],
                ["1m", 60],
                ["1h", 3600],
                ["1d", 86400]]:
            data = v.validate({"value": o[0]})
            self.assertTrue(data["value"] == o[1])

    def test_isa_unixtime(self):
        v = Validator()
        v.set(value={"type": str, "isa": "unixtime"})
        for o in [
                [1, 1],
                ["1", 1],
                ["2000-01-01", 946681200],
                ["2000-01-01 00:00:00", 946681200]]:
            data = v.validate({"value": o[0]})
            self.assertTrue(data["value"] == o[1])
        self.assertRaisesRegex(
            Exception,
            "err-710",
            v.validate, {"value": "0000-00-00 00:00:00"})

    def test_bool(self):
        v = Validator()
        v.set(value={"type": bool})
        for o in [
                [True, True],
                ["1", True],
                ["Yes", True],
                ["yes", True],
                ["True", True],
                ["true", True],
                [False, False],
                [0, False],
                ["No", False],
                ["no", False],
                ["False", False],
                ["false", False]]:
            data = v.validate({"value": o[0]})
            self.assertTrue(data["value"] == o[1])
        self.assertRaisesRegex(
            Exception,
            "err-712",
            v.validate, {"value": "string"})

    def test_optional(self):
        v = Validator()
        v.set(value={"type": str, "optional": True})
        data = v.validate({})
        self.assertTrue(data["value"] is None)

    def test_optional_ignore_missing(self):
        v = Validator()
        v.set(value={"type": str, "optional": True})
        data = v.validate({}, ignore_missing=True)
        self.assertTrue(len(data) == 0)

    def test_default(self):
        v = Validator()
        v.set(value={"type": str, "default": None})
        data = v.validate({})
        self.assertTrue(data["value"] is None)

    def test_function(self):
        def __test_function(value, *args, **kwargs):
            if (args[0] == "foo"
                    and args[1] == "bar"
                    and kwargs["foo"] == "bar"
                    and value == "foobar"):
                return True
            return False
        v = Validator()
        v.set(value={
            "type": str,
            "function": __test_function,
            "function_args": ("foo", "bar"),
            "function_kwargs": {"foo": "bar"}
        })
        data = v.validate({"value": "foobar"})
        self.assertTrue(data["value"] == "foobar")
        self.assertRaisesRegex(
            Exception,
            "err-710",
            v.validate, {"value": "wrong value"})

    def test_multiple(self):
        v = Validator()
        v.set({
            "param": {
                "type": (int,float),
                "multiple": True
            }
        })
        self.assertRaisesRegex(
            Exception,
            "err-711",
            v.validate, {"param": ["2", "x", "2.4"]})
        data = v.validate({"param": ["2", "2.4"]})
        self.assertTrue(isinstance(data["param"], list))
        self.assertTrue(data["param"][0] == 2)
        self.assertTrue(data["param"][1] == 2.4)

    def test_list_of_dicts_options(self):
        v = Validator()
        v.set({
            "user_id": {
                "type": int,
                "list_of_dicts_key": "id",
                "list_of_dicts_options": [
                    {"id": 1, "name": "Jessica"},
                    {"id": 2, "name": "Jonny"},
                    {"id": 3, "name": "Joanne"},
                    {"id": 4, "name": "Jannik"},
                    {"id": 5, "name": "Niklas"},
                    {"id": 6, "name": "Linus"},
                    {"id": 7, "name": "Mila"},
                ]
            }
        })
        v.validate({"user_id": 1})

    def test_list_of_dicts_multioptions(self):
        v = Validator()
        v.set({
            "user_id": {
                "list_of_dicts_key": "id",
                "list_of_dicts_multioptions": [
                    {"id": 1, "name": "Jessica"},
                    {"id": 2, "name": "Jonny"},
                    {"id": 3, "name": "Joanne"},
                    {"id": 4, "name": "Jannik"},
                    {"id": 5, "name": "Niklas"},
                    {"id": 6, "name": "Linus"},
                    {"id": 7, "name": "Mila"},
                ]
            }
        })
        v.validate({"user_id": [1, 2, 3]})

    def test_depends_on(self):
        v = Validator()
        v.set({
            "from_time": {
                "type": int,
                "optional": True,
            },
            "to_time": {
                "type": int,
                "optional": True,
                "depends_on": {
                    "from_time": {
                        "lower_or_greater": "greater",
                        "max_diff": 10
                    }
                }
            },
        })
        self.assertTrue(v.validate({"from_time": 5}))
        self.assertRaisesRegex(
            Exception,
            "err-702",
            v.validate, {"from_time": 10, "to_time": 5})
        self.assertRaisesRegex(
            Exception,
            "err-703",
            v.validate, {"from_time": 0, "to_time": 20})

    def test_no_mixin(self):
        v = Validator(no_mixin=[["data_a", "data_b"]])
        v.set({
            "data_a": {
                "type": int,
                "optional": True,
            },
            "data_b": {
                "type": int,
                "optional": True,
            },
            "data_c": {
                "type": int,
                "optional": True,
            },
        })
        self.assertRaisesRegex(
            Exception,
            "err-719",
            v.validate, {"data_a": 1, "data_b": 1, "data_c": 1})
        self.assertRaisesRegex(
            Exception,
            "err-719",
            v.validate, {"data_a": 1, "data_b": 1})
        self.assertTrue(v.validate({"data_a": 1, "data_c": 1}))

unittest.main(verbosity=2)
